/*
 * Copyright (C) 2023 European Spallation Source ERIC.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.openepics.names.service;

import java.text.MessageFormat;
import java.util.Date;
import java.util.List;
import java.util.UUID;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.commons.lang3.StringUtils;
import org.openepics.names.repository.ISubsystemRepository;
import org.openepics.names.repository.SubsystemRepository;
import org.openepics.names.repository.model.Subsystem;
import org.openepics.names.rest.beans.Status;
import org.openepics.names.rest.beans.Type;
import org.openepics.names.rest.beans.element.NameElementCommand;
import org.openepics.names.rest.beans.element.StructureElementCommand;
import org.openepics.names.util.EssNamingConvention;
import org.openepics.names.util.HolderSystemDeviceStructure;
import org.openepics.names.util.StructureChoice;
import org.openepics.names.util.StructureCommand;
import org.openepics.names.util.StructureElementUtil;
import org.openepics.names.util.StructureUtil;
import org.openepics.names.util.TextUtil;
import org.openepics.names.util.ValidateUtil;
import org.openepics.names.util.notification.NotificationUtil;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;

/**
 * This class provides structures services for subsystem.
 *
 * @author Lars Johansson
 */
@Service
public class SubsystemService {

    // note
    //     structurecommand
    //         cud - create update delete
    //         acr - approve cancel reject

    private static final Logger LOGGER = Logger.getLogger(SubsystemService.class.getName());

    private ISubsystemRepository iSubsystemRepository;
    private SubsystemRepository subsystemRepository;
    private NamesService namesService;

    @Autowired
    public SubsystemService(
            ISubsystemRepository iSubsystemRepository,
            SubsystemRepository subsystemRepository,
            NamesService namesService) {
        this.iSubsystemRepository = iSubsystemRepository;
        this.subsystemRepository = subsystemRepository;
        this.namesService = namesService;
    }

    @Transactional(propagation = Propagation.MANDATORY)
    public StructureElementNotification createStructure(StructureElementCommand structureElementCommand, Date when, String username, EssNamingConvention namingConvention, HolderSystemDeviceStructure holderStructure) {
        // validation outside method
        // transaction
        //     support a current transaction, throw an exception if none exists
        //     attributes
        //     create structure to pending, not latest, not deleted, with data
        //     return
        //         structure element for created structure
        //         notification
        //
        // attributes
        //     type, parent, name, mnemonic, description, comment

        // note
        //     namingConvention.equivalenceClassRepresentative return null for null input
        //     rules for mnemonic

        String mnemonic = structureElementCommand.getMnemonic();
        mnemonic = StringUtils.isEmpty(mnemonic) ? null : mnemonic;
        String equivalenceClassRepresentative = namingConvention.equivalenceClassRepresentative(mnemonic);

        // create
        Subsystem subsystem = new Subsystem(UUID.randomUUID(), structureElementCommand.getParent(),
                structureElementCommand.getName(), mnemonic, equivalenceClassRepresentative,
                structureElementCommand.getDescription(), Status.PENDING, Boolean.FALSE, Boolean.FALSE,
                when, username, structureElementCommand.getComment());
        subsystemRepository.createSubsystem(subsystem);

        return new StructureElementNotification(
                StructureElementUtil.getStructureElementRequested(subsystem),
                NotificationUtil.prepareNotification(Type.SUBSYSTEM, StructureCommand.CREATE, null, null, subsystem, holderStructure));
    }

    @Transactional(propagation = Propagation.MANDATORY)
    public StructureElementNotification updateStructure(StructureElementCommand structureElementCommand, Date when, String username, EssNamingConvention namingConvention, HolderSystemDeviceStructure holderStructure) {
        // validation outside method
        // transaction
        //     support a current transaction, throw an exception if none exists
        //     attributes
        //     create structure to pending, not latest, not deleted, with data
        //     find out previous
        //     return
        //         structure element for updated structure
        //         notification
        //
        // attributes
        //     uuid, type, parent, name, mnemonic, description, comment

        // note
        //     namingConvention.equivalenceClassRepresentative return null for null input
        //     rules for mnemonic

        String mnemonic = structureElementCommand.getMnemonic();
        mnemonic = StringUtils.isEmpty(mnemonic) ? null : mnemonic;
        String equivalenceClassRepresentative = namingConvention.equivalenceClassRepresentative(mnemonic);

        // create
        Subsystem subsystem = new Subsystem(structureElementCommand.getUuid(), structureElementCommand.getParent(),
                structureElementCommand.getName(), mnemonic, equivalenceClassRepresentative,
                structureElementCommand.getDescription(), Status.PENDING, Boolean.FALSE, Boolean.FALSE,
                when, username, structureElementCommand.getComment());
        subsystemRepository.createSubsystem(subsystem);

        // previous
        List<Subsystem> previouses = iSubsystemRepository.findPreviousByUuidAndId(subsystem.getUuid().toString(),  subsystem.getId());

        return new StructureElementNotification(
                StructureElementUtil.getStructureElementRequested(subsystem),
                NotificationUtil.prepareNotification(Type.SUBSYSTEM, StructureCommand.UPDATE, null,
                        previouses != null && !previouses.isEmpty() ? previouses.get(0) : null, subsystem, holderStructure));
    }

    @Transactional(propagation = Propagation.MANDATORY)
    public StructureElementNotification deleteStructure(StructureElementCommand structureElementCommand, Date when, String username, HolderSystemDeviceStructure holderStructure) {
        // validation outside method
        // transaction
        //     support a current transaction, throw an exception if none exists
        //     attributes
        //     find
        //     create structure to pending, not latest, deleted, with data
        //     return
        //         structure element for deleted structure
        //         notification
        //
        // attributes
        //     uuid, comment

        // find
        List<Subsystem> subsystems = subsystemRepository.readSubsystems(Status.APPROVED, Boolean.FALSE, structureElementCommand.getUuid().toString(), null, null, null, null, null, null);
        if (ValidateUtil.isNullOrNotSize(subsystems, 1)) {
            return null;
        }
        Subsystem toBeDeleted = subsystems.get(0);

        // create
        Subsystem subsystem = new Subsystem(toBeDeleted.getUuid(), toBeDeleted.getParentUuid(),
                toBeDeleted.getName(), toBeDeleted.getMnemonic(), toBeDeleted.getMnemonicEquivalence(),
                toBeDeleted.getDescription(), Status.PENDING, Boolean.FALSE, Boolean.TRUE,
                when, username, structureElementCommand.getComment());
        subsystemRepository.createSubsystem(subsystem);

        return new StructureElementNotification(
                StructureElementUtil.getStructureElementRequested(subsystem),
                NotificationUtil.prepareNotification(Type.SUBSYSTEM, StructureCommand.DELETE, null, null, subsystem, holderStructure));
    }

    @Transactional(propagation = Propagation.MANDATORY)
    public StructureElementNotification approveStructure(StructureElementCommand structureElementCommand, Date when, String username, HolderSystemDeviceStructure holderStructure) {
        // validation outside method
        // transaction
        //     support a current transaction, throw an exception if none exists
        //     attributes
        //     find
        //     update not latest
        //     find
        //     approve - update structure to status APPROVED, latest to true
        //     additional
        //         find out previous to find out what is approved - create update delete
        //         approve create
        //             automatically create name when system structure is approved
        //             condition on name and structure entry
        //                 system structure should exist (uuid), one entry that is not deleted
        //                 name should not exist (system structure mnemonic)
        //             within current transaction
        //         approve update
        //             update related names
        //             for different mnemonic
        //             not for device group
        //         approve delete
        //             keep related names - legacy names
        //             delete sub structures
        //     return
        //         structure element for approved structure
        //         notification
        //
        // attributes
        //     uuid, comment

        String uuid = structureElementCommand.getUuid().toString();

        // find
        List<Subsystem> subsystems = subsystemRepository.readSubsystems(Status.APPROVED, Boolean.FALSE, uuid, null, null, null, null, null, null);
        Subsystem subsystem = null;
        if (ValidateUtil.isSize(subsystems, 1)) {
            subsystem = subsystems.get(0);

            // update not latest
            subsystem.setLatest(Boolean.FALSE);
            subsystemRepository.updateSubsystem(subsystem);
        }

        // find
        subsystems = subsystemRepository.readSubsystems(Status.PENDING, null, uuid, null, null, null, null, null, null);
        if (ValidateUtil.isNullOrNotSize(subsystems, 1)) {
            return null;
        }
        subsystem = subsystems.get(0);

        // approve
        subsystem.setAttributesStatusProcessed(Status.APPROVED, when, username, structureElementCommand.getComment());
        subsystem.setLatest(Boolean.TRUE);
        subsystemRepository.updateSubsystem(subsystem);

        // previous
        List<Subsystem> previouses = iSubsystemRepository.findPreviousByUuidAndId(uuid, subsystem.getId());
        StructureCommand structureCommandCUD = StructureUtil.getStructureCommandCUD(previouses, subsystem);
        Subsystem previous = previouses != null && !previouses.isEmpty() ? previouses.get(0) : null;

        // additional
        if (StructureCommand.CREATE.equals(structureCommandCUD)) {
            boolean hasMnemonic = !StringUtils.isEmpty(subsystem.getMnemonic());
            boolean existsName = hasMnemonic && namesService.existsName(StructureUtil.getMnemonicPath(subsystem, holderStructure));
            if (hasMnemonic && !existsName) {
                NameElementCommand nameElement = new NameElementCommand(null, subsystem.getUuid(), null, null, StructuresService.SYSTEM_STRUCTURE_ONLY, null);
                namesService.createName(nameElement, when, username, holderStructure);
            }
        } else if (StructureCommand.UPDATE.equals(structureCommandCUD)) {
            namesService.updateNames(previous, subsystem, username);
        } else if (StructureCommand.DELETE.equals(structureCommandCUD)) {
            // not delete names - legacy names
        }

        LOGGER.log(Level.FINE, () -> MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.APPROVE_STRUCTURE, TextUtil.COMMAND, structureCommandCUD));

        return new StructureElementNotification(
                StructureElementUtil.getStructureElementProcessed(subsystem, holderStructure, StructureChoice.STRUCTURE),
                NotificationUtil.prepareNotification(Type.SUBSYSTEM, structureCommandCUD, StructureCommand.APPROVE, previous, subsystem, holderStructure));
    }

    @Transactional(propagation = Propagation.MANDATORY)
    public StructureElementNotification cancelStructure(StructureElementCommand structureElementCommand, Date when, String username, HolderSystemDeviceStructure holderStructure) {
        // validation outside method
        // transaction
        //     support a current transaction, throw an exception if none exists
        //     attributes
        //     find
        //     cancel - update structure to status CANCELLED
        //     find out previous
        //     return
        //         structure element for created structure
        //         notification
        //
        // attributes
        //     uuid, comment

        String uuid = structureElementCommand.getUuid().toString();

        // find
        List<Subsystem> subsystems = subsystemRepository.readSubsystems(Status.PENDING, null, uuid, null, null, null, null, null, null);
        if (ValidateUtil.isNullOrNotSize(subsystems, 1)) {
            return null;
        }
        Subsystem subsystem = subsystems.get(0);

        // cancel
        subsystem.setAttributesStatusProcessed(Status.CANCELLED, when, username, structureElementCommand.getComment());
        subsystemRepository.updateSubsystem(subsystem);

        // previous
        List<Subsystem> previouses = iSubsystemRepository.findPreviousByUuidAndId(uuid, subsystem.getId());
        StructureCommand structureCommandCUD = NotificationUtil.getStructureCommandCUD(previouses, subsystem);

        LOGGER.log(Level.FINE, () -> MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.CANCEL_STRUCTURE, TextUtil.COMMAND, structureCommandCUD));

        return new StructureElementNotification(
                StructureElementUtil.getStructureElementProcessed(subsystem, holderStructure, StructureChoice.STRUCTURE),
                NotificationUtil.prepareNotification(Type.SUBSYSTEM, structureCommandCUD, StructureCommand.CANCEL,
                        previouses != null && !previouses.isEmpty() ? previouses.get(0) : null, subsystem, holderStructure));
    }

    @Transactional(propagation = Propagation.MANDATORY)
    public StructureElementNotification rejectStructure(StructureElementCommand structureElementCommand, Date when, String username, HolderSystemDeviceStructure holderStructure) {
        // validation outside method
        // transaction
        //     support a current transaction, throw an exception if none exists
        //     attributes
        //     find
        //     reject - update structure to status REJECTED
        //     find out previous
        //     return
        //         structure element for created structure
        //         notification
        //
        // attributes
        //     uuid, comment

        String uuid = structureElementCommand.getUuid().toString();

        // find
        List<Subsystem> subsystems = subsystemRepository.readSubsystems(Status.PENDING, null, uuid, null, null, null, null, null, null);
        if (ValidateUtil.isNullOrNotSize(subsystems, 1)) {
            return null;
        }
        Subsystem subsystem = subsystems.get(0);

        // reject
        subsystem.setAttributesStatusProcessed(Status.REJECTED, when, username, structureElementCommand.getComment());
        subsystemRepository.updateSubsystem(subsystem);

        // previous
        List<Subsystem> previouses = iSubsystemRepository.findPreviousByUuidAndId(uuid, subsystem.getId());
        StructureCommand structureCommandCUD = NotificationUtil.getStructureCommandCUD(previouses, subsystem);

        LOGGER.log(Level.FINE, () -> MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.REJECT_STRUCTURE, TextUtil.COMMAND, structureCommandCUD));

        return new StructureElementNotification(
                StructureElementUtil.getStructureElementProcessed(subsystem, holderStructure, StructureChoice.STRUCTURE),
                NotificationUtil.prepareNotification(Type.SUBSYSTEM, structureCommandCUD, StructureCommand.REJECT,
                        previouses != null && !previouses.isEmpty() ? previouses.get(0) : null, subsystem, holderStructure));
    }

}
