/*
 * Copyright (C) 2022 European Spallation Source ERIC.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package org.openepics.names.util;

import java.util.List;
import java.util.UUID;

import org.apache.commons.lang3.StringUtils;
import org.openepics.names.exception.DataConflictException;
import org.openepics.names.repository.model.DeviceGroup;
import org.openepics.names.repository.model.DeviceType;
import org.openepics.names.repository.model.Discipline;
import org.openepics.names.repository.model.Subsystem;
import org.openepics.names.repository.model.System;
import org.openepics.names.repository.model.SystemGroup;
import org.openepics.names.rest.beans.FieldStructure;
import org.openepics.names.rest.beans.Status;
import org.openepics.names.rest.beans.Type;
import org.openepics.names.rest.beans.element.StructureElementCommand;

import com.google.common.collect.Lists;

/**
 * Utility class to assist in handling of validation.
 *
 * @author Lars Johansson
 */
public class ValidateStructureElementUtil {

    // note
    //     handling of system structure, device structure
    //         parent system structure uuid (system group, system, subsystem)
    //         parent device structure uuid (device type)
    //
    //     StructureElement
    //         type, uuid, parent,
    //         name, mnemonic, mnemonic path, level,
    //         description, status, latest, deleted, when, who, comment
    //
    //     StructureElementCommand
    //	       create  -       type, parent, name, mnemonic, description, comment
    //	       update  - uuid, type, parent, name, mnemonic, description, comment
    //	       delete  - uuid, type,                                      comment
    //	       approve - uuid, type,                                      comment
    //	       cancel  - uuid, type,                                      comment
    //	       reject  - uuid, type,                                      comment

    /**
     * This class is not to be instantiated.
     */
    private ValidateStructureElementUtil() {
        throw new IllegalStateException("Utility class");
    }

    /**
     * Validate parameters for read structures.
     *
     * @param type type
     * @param statuses statuses
     * @param deleted deleted
     * @param uuid uuid
     * @param parent parent uuid
     * @param name name
     * @param mnemonic mnemonic
     * @param mnemonicEquivalence mnemonic equivalence
     * @param mnemonicPath mnemonic path
     * @param description description
     * @param includeHistory include history
     * @param orderBy order by
     * @param isAsc is ascending
     * @param offset offset
     * @param limit limit
     */
    public static void validateStructuresInputRead(Type type, Status[] statuses, Boolean deleted,
            String uuid, String parent, String name, String mnemonic, String mnemonicEquivalence, String mnemonicPath, String description,
            Boolean includeHistory, FieldStructure orderBy, Boolean isAsc, Integer offset, Integer limit) {

        // validate input
        //     type
        //     uuid

        ValidateUtil.validateInputType(type);
        if (!StringUtils.isEmpty(uuid)) {
            ValidateUtil.validateInputUuid(uuid);
        }
    }

    /**
     * Validate structure element command parameters (input) for create.
     *
     * @param structureElementCommand structure element command
     */
    public static void validateStructureElementInputCreate(StructureElementCommand structureElementCommand, EssNamingConvention namingConvention) {
        validateStructureElementInput(structureElementCommand, namingConvention, StructureCommand.CREATE);
    }

    /**
     * Validate structure element command parameters (input) for update.
     *
     * @param structureElementCommand structure element command
     */
    public static void validateStructureElementInputUpdate(StructureElementCommand structureElementCommand, EssNamingConvention namingConvention) {
        validateStructureElementInput(structureElementCommand, namingConvention, StructureCommand.UPDATE);
    }

    /**
     * Validate structure element command parameters (input) for delete.
     *
     * @param structureElementCommand structure element command
     */
    public static void validateStructureElementInputDelete(StructureElementCommand structureElementCommand, EssNamingConvention namingConvention) {
        validateStructureElementInput(structureElementCommand, namingConvention, StructureCommand.DELETE);
    }

    /**
     * Validate structure element command parameters (input) for approve.
     *
     * @param structureElementCommand structure element command
     */
    public static void validateStructureElementInputApprove(StructureElementCommand structureElementCommand, EssNamingConvention namingConvention) {
        validateStructureElementInput(structureElementCommand, namingConvention, StructureCommand.APPROVE);
    }

    /**
     * Validate structure element command parameters (input) for cancel.
     *
     * @param structureElementCommand structure element command
     */
    public static void validateStructureElementInputCancel(StructureElementCommand structureElementCommand, EssNamingConvention namingConvention) {
        validateStructureElementInput(structureElementCommand, namingConvention, StructureCommand.CANCEL);
    }

    /**
     * Validate structure element command parameters (input) for reject.
     *
     * @param structureElementCommand structure element command
     */
    public static void validateStructureElementInputReject(StructureElementCommand structureElementCommand, EssNamingConvention namingConvention) {
        validateStructureElementInput(structureElementCommand, namingConvention, StructureCommand.REJECT);
    }

    /**
     * Validate structure element command parameters (input).
     *
     * @param structureElementCommand structure element command
     * @param structureCommand structure command
     */
    private static void validateStructureElementInput(StructureElementCommand structureElementCommand, EssNamingConvention namingConvention, StructureCommand structureCommand) {
        // command - attributes
        //     create  -       type, parent, name, mnemonic, description, comment
        //	   update  - uuid, type, parent, name, mnemonic, description, comment
        //	   delete  - uuid, type,                                      comment
        //	   approve - uuid, type,                                      comment
        //	   cancel  - uuid, type,                                      comment
        //	   reject  - uuid, type,                                      comment
        //
        //     create, update - optional - parent, mnemonic
        //
        // logic
        //     update, delete, approve, cancel, reject
        //     create, update
        //     create, update, delete, approve, cancel, reject

        if (ValidateUtil.isAnyNull(structureElementCommand, namingConvention, structureCommand)) {
            return;
        }

        UUID uuid = structureElementCommand.getUuid();
        Type type = structureElementCommand.getType();
        UUID parent = structureElementCommand.getParent();
        String name = structureElementCommand.getName();
        String mnemonic = structureElementCommand.getMnemonic();
        String description = structureElementCommand.getDescription();
        String comment = structureElementCommand.getComment();

        if (ValidateUtil.isAnyEqual(structureCommand, StructureCommand.UPDATE, StructureCommand.DELETE, StructureCommand.APPROVE, StructureCommand.CANCEL, StructureCommand.REJECT)) {
            ValidateUtil.validateInputUuid(uuid != null ? uuid.toString() : null);
        }

        ValidateUtil.validateInputType(type);

        if (ValidateUtil.isAnyEqual(structureCommand, StructureCommand.CREATE, StructureCommand.UPDATE)) {
            if (ValidateUtil.isAnyEqual(type, Type.SYSTEM, Type.SUBSYSTEM, Type.DEVICEGROUP, Type.DEVICETYPE)) {
                ExceptionUtil.validateConditionInputNotAvailableException(parent != null,
                        TextUtil.PARENT_IS_NOT_AVAILABLE, structureElementCommand.toString(), TextUtil.PARENT);
                ValidateUtil.validateInputUuid(parent.toString());
            }

            ValidateUtil.validateInputName(name);

            // validate mnemonic
            //     validate mnemonic input (value itself, not in relation to other values)
            //     validateMnemonic takes isMnemonicRequired into account
            MnemonicValidation mnemonicValidation = namingConvention.validateMnemonic(type, mnemonic);
            ExceptionUtil.validateConditionInputNotValidException(MnemonicValidation.VALID.equals(mnemonicValidation),
                    TextUtil.MNEMONIC_IS_NOT_VALID, structureElementCommand.toString(), TextUtil.MNEMONIC);

            ValidateUtil.validateInputDescription(description);
        }

        ValidateUtil.validateInputComment(comment);
    }

    // ----------------------------------------------------------------------------------------------------

    /**
     * Validate structure element command data for create.
     *
     * @param structureElementCommand structure element command
     * @param namingConvention naming convention
     * @param holderRepositories holder repositories
     * @param holderStructure holder of system and device structure content
     */
    public static void validateStructureElementDataCreate(StructureElementCommand structureElementCommand, EssNamingConvention namingConvention, HolderRepositories holderRepositories, HolderSystemDeviceStructure holderStructure) {
        validateStructureElementDataInItself(structureElementCommand, holderRepositories, StructureCommand.CREATE);
        validateStructureElementDataRelativeOtherData(structureElementCommand, namingConvention, holderRepositories, holderStructure, StructureCommand.CREATE);
    }

    /**
     * Validate structure element command data for update.
     *
     * @param structureElementCommand structure element command
     * @param namingConvention naming convention
     * @param holderRepositories holder repositories
     * @param holderStructure holder of system and device structure content
     */
    public static void validateStructureElementDataUpdate(StructureElementCommand structureElementCommand, EssNamingConvention namingConvention, HolderRepositories holderRepositories, HolderSystemDeviceStructure holderStructure) {
        validateStructureElementDataInItself(structureElementCommand, holderRepositories, StructureCommand.UPDATE);
        validateStructureElementDataRelativeOtherData(structureElementCommand, namingConvention, holderRepositories, holderStructure, StructureCommand.UPDATE);
    }

    /**
     * Validate structure element command data for delete.
     *
     * @param structureElementCommand structure element command
     * @param namingConvention naming convention
     * @param holderRepositories holder repositories
     * @param holderStructure holder of system and device structure content
     */
    public static void validateStructureElementDataDelete(StructureElementCommand structureElementCommand, EssNamingConvention namingConvention, HolderRepositories holderRepositories, HolderSystemDeviceStructure holderStructure) {
        validateStructureElementDataInItself(structureElementCommand, holderRepositories, StructureCommand.DELETE);
        validateStructureElementDataRelativeOtherData(structureElementCommand, namingConvention, holderRepositories, holderStructure, StructureCommand.DELETE);
    }

    /**
     * Validate structure element command data for approve.
     *
     * @param structureElementCommand structure element command
     * @param namingConvention naming convention
     * @param holderRepositories holder repositories
     * @param holderStructure holder of system and device structure content
     */
    public static void validateStructureElementDataApprove(StructureElementCommand structureElementCommand, EssNamingConvention namingConvention, HolderRepositories holderRepositories, HolderSystemDeviceStructure holderStructure) {
        validateStructureElementDataInItself(structureElementCommand, holderRepositories, StructureCommand.APPROVE);
        validateStructureElementDataRelativeOtherData(structureElementCommand, namingConvention, holderRepositories, holderStructure, StructureCommand.APPROVE);
    }

    /**
     * Validate structure element command data for cancel.
     *
     * @param structureElementCommand structure element command
     * @param namingConvention naming convention
     * @param holderRepositories holder repositories
     * @param holderStructure holder of system and device structure content
     */
    public static void validateStructureElementDataCancel(StructureElementCommand structureElementCommand, EssNamingConvention namingConvention, HolderRepositories holderRepositories, HolderSystemDeviceStructure holderStructure) {
        validateStructureElementDataInItself(structureElementCommand, holderRepositories, StructureCommand.CANCEL);
        validateStructureElementDataRelativeOtherData(structureElementCommand, namingConvention, holderRepositories, holderStructure, StructureCommand.CANCEL);
    }

    /**
     * Validate structure element command data for reject.
     *
     * @param structureElementCommand structure element command
     * @param namingConvention naming convention
     * @param holderRepositories holder repositories
     * @param holderStructure holder of system and device structure content
     */
    public static void validateStructureElementDataReject(StructureElementCommand structureElementCommand, EssNamingConvention namingConvention, HolderRepositories holderRepositories, HolderSystemDeviceStructure holderStructure) {
        validateStructureElementDataInItself(structureElementCommand, holderRepositories, StructureCommand.REJECT);
        validateStructureElementDataRelativeOtherData(structureElementCommand, namingConvention, holderRepositories, holderStructure, StructureCommand.REJECT);
    }

    /**
     * Validate structure element command data in itself.
     *
     * @param structureElementCommand structure element command
     * @param holderRepositories holder respositories
     * @param structureCommand structure command
     */
    public static void validateStructureElementDataInItself(StructureElementCommand structureElementCommand, HolderRepositories holderRepositories, StructureCommand structureCommand) {
        // command - attributes
        //     create  -       type, parent, name, mnemonic, description, comment
        //	   update  - uuid, type, parent, name, mnemonic, description, comment
        //	   delete  - uuid, type,                                      comment
        //	   approve - uuid, type,                                      comment
        //	   cancel  - uuid, type,                                      comment
        //	   reject  - uuid, type,                                      comment
        //
        //     create, update - optional - parent, mnemonic
        //
        // check structure element data in itself
        //     update, delete
        //         definitely (not possibly)
        //             no pending entry waiting to be approved, cancelled, rejected (pending with higher id than currently approved)
        //         uuid - approved, latest, not deleted - list size 1
        //     approve, reject, cancel
        //         uuid - pending, latest, not deleted  - list size 1

        if (ValidateUtil.isAnyNull(structureElementCommand, holderRepositories, structureCommand)) {
            return;
        }

        UUID uuid = structureElementCommand.getUuid();
        Type type = structureElementCommand.getType();
        UUID parent = structureElementCommand.getParent();

        String message = type.toString().toLowerCase() + TextUtil.SPACE + TextUtil.IS_NOT_CORRECT;
        String details = structureElementCommand.toString();
        String field   = TextUtil.UUID;

        // validate data in itself
        //     systemgroup, discipline
        //         not have parent
        //     update, delete
        //         entry not pending
        //         entry approved and not deleted
        //     approve, cancel, reject
        //         entry pending

        if (ValidateUtil.isAnyEqual(type, Type.SYSTEMGROUP, Type.DISCIPLINE) && parent != null) {
            throw ExceptionUtil.createDataNotCorrectException(TextUtil.PARENT_IS_NOT_CORRECT, details, field);
        } else if (ValidateUtil.isAnyEqual(structureCommand, StructureCommand.UPDATE, StructureCommand.DELETE)) {
            validateStructuresFound(uuid, type,
                    holderRepositories, message, details, field);
            validateStructuresValidSize0(uuid, type, Status.PENDING, null,
                    holderRepositories, message, details, field);
            validateStructuresValidSize1(uuid, type, Status.APPROVED, Boolean.FALSE,
                    holderRepositories, message, details, field);
        } else if (ValidateUtil.isAnyEqual(structureCommand, StructureCommand.APPROVE, StructureCommand.CANCEL, StructureCommand.REJECT)) {
            validateStructuresFound(uuid, type,
                    holderRepositories, message, details, field);
            validateStructuresValidSize1(uuid, type, Status.PENDING, null,
                    holderRepositories, message, details, field);
        }
    }

    private static void validateStructuresFound(UUID uuid, Type type,
            HolderRepositories holderRepositories, String message, String details, String field) {
        ExceptionUtil.validateConditionDataNotFoundException(
                getStructuresSize(uuid, type, null, null, holderRepositories) > 0, message, details, field);
    }
    private static void validateStructuresValidSize0(UUID uuid, Type type, Status status, Boolean deleted,
            HolderRepositories holderRepositories, String message, String details, String field) {
        ExceptionUtil.validateConditionDataNotValidException(
                getStructuresSize(uuid, type, status, deleted, holderRepositories) == 0, message, details, field);
    }
    private static void validateStructuresValidSize1(UUID uuid, Type type, Status status, Boolean deleted,
            HolderRepositories holderRepositories, String message, String details, String field) {
        ExceptionUtil.validateConditionDataNotValidException(
                getStructuresSize(uuid, type, status, deleted, holderRepositories) == 1, message, details, field);
    }
    private static int getStructuresSize(UUID uuid, Type type, Status status, Boolean deleted,
            HolderRepositories holderRepositories) {
        if (Type.SYSTEMGROUP.equals(type)) {
            return holderRepositories.systemGroupRepository().readSystemGroups(status, deleted, uuid.toString(), null, null, null, null, null).size();
        } else if (Type.SYSTEM.equals(type)) {
            return holderRepositories.systemRepository().readSystems(status, deleted, uuid.toString(), null, null, null, null, null, null).size();
        } else if (Type.SUBSYSTEM.equals(type)) {
            return holderRepositories.subsystemRepository().readSubsystems(status, deleted, uuid.toString(), null, null, null, null, null, null).size();
        } else if (Type.DISCIPLINE.equals(type)) {
            return holderRepositories.disciplineRepository().readDisciplines(status, deleted, uuid.toString(), null, null, null, null, null).size();
        } else if (Type.DEVICEGROUP.equals(type)) {
            return holderRepositories.deviceGroupRepository().readDeviceGroups(status, deleted, uuid.toString(), null, null, null, null, null, null).size();
        } else if (Type.DEVICETYPE.equals(type)) {
            return holderRepositories.deviceTypeRepository().readDeviceTypes(status, deleted, uuid.toString(), null, null, null, null, null, null).size();
        }
        return 0;
    }

    /**
     * Validate structure element command data relative other data.
     *
     * @param structureElementCommand structure element command
     * @param namingConvention naming convention
     * @param holderRepositories holder repositories
     * @param holderStructure holder of system and device structure content
     * @param structureCommand structure command
     */
    public static void validateStructureElementDataRelativeOtherData(StructureElementCommand structureElementCommand, EssNamingConvention namingConvention, HolderRepositories holderRepositories, HolderSystemDeviceStructure holderStructure, StructureCommand structureCommand) {
        // command - attributes
        //     create  -       type, parent, name, mnemonic, description, comment
        //	   update  - uuid, type, parent, name, mnemonic, description, comment
        //	   delete  - uuid, type,                                      comment
        //	   approve - uuid, type,                                      comment
        //	   cancel  - uuid, type,                                      comment
        //	   reject  - uuid, type,                                      comment
        //
        //     create, update - optional - parent, mnemonic
        //
        // check structure element data in relation to other data
        //     create, update
        //         parent               - (if applicable) approved, latest, not deleted
        //         mnemonic             - (if not same)   not exists
        //         mnemonic equivalence - (if not same)   equivalence not exists
        //     approve
        //         uuid - pending, not latest, deleted
        //         no or less checks if entry to be deleted                - approve (delete)
        //         more or same checks as above if entry is not be deleted - approve (create), approve (update)
        //         need checks as content may have changed from time of create, update, delete to time of approve
        //     checks on mnemonic, mnemonic equivalence are to ensure can coexist, can move
        //     possibly
        //         additional checks if names are affected
        //         comment not same as previous comment

        if (ValidateUtil.isAnyNull(structureElementCommand, namingConvention, holderRepositories, holderStructure, structureCommand)) {
            return;
        }

        String message = null;
        String details = structureElementCommand.toString();
        String field   = TextUtil.MNEMONIC;

        UUID uuid = structureElementCommand.getUuid();
        Type type = structureElementCommand.getType();
        UUID parent = structureElementCommand.getParent();
        String mnemonic = structureElementCommand.getMnemonic();

        // validate data relative other data
        //     create, update
        //         entry (uuid,) status, mnemonic
        //         entry (uuid,) status, mnemonic equivalence
        //     approve
        //         entry uuid, status

        if (ValidateUtil.isAnyEqual(structureCommand, StructureCommand.CREATE, StructureCommand.UPDATE)) {
            if (Type.SYSTEMGROUP.equals(type)) {
                // note rules for mnemonic for system group
                if (!StringUtils.isEmpty(mnemonic)) {
                    // status, mnemonic
                    message = TextUtil.SYSTEMGROUP_IS_NOT_CORRECT;
                    validateStructuresStatusMnemonic(uuid, type, parent, mnemonic,
                            namingConvention, holderRepositories, structureCommand, message, details, field);

                    // status, mnemonic equivalence
                    validateStructuresStatusMnemonicequivalence(uuid, type, parent, mnemonic,
                            namingConvention, holderRepositories, structureCommand, message, details, field);
                }
            } else if (Type.SYSTEM.equals(type)) {
                // status, parent
                message = TextUtil.SYSTEMGROUP_IS_NOT_CORRECT;
                validateStructuresParent(type, parent, Status.APPROVED, false,
                        holderRepositories, message, details, field);

                // status, mnemonic
                message = TextUtil.SYSTEM_IS_NOT_CORRECT;
                validateStructuresStatusMnemonic(uuid, type, parent, mnemonic,
                        namingConvention, holderRepositories, structureCommand, message, details, field);

                // status, mnemonic equivalence
                validateStructuresStatusMnemonicequivalence(uuid, type, parent, mnemonic,
                        namingConvention, holderRepositories, structureCommand, message, details, field);
            } else if (Type.SUBSYSTEM.equals(type)) {
                // status, parent
                message = TextUtil.SYSTEM_IS_NOT_CORRECT;
                validateStructuresParent(type, parent, Status.APPROVED, false,
                        holderRepositories, message, details, field);

                // status, mnemonic
                message = TextUtil.SUBSYSTEM_IS_NOT_CORRECT;
                validateStructuresStatusMnemonic(uuid, type, parent, mnemonic,
                        namingConvention, holderRepositories, structureCommand, message, details, field);

                // status, mnemonic equivalence
                validateStructuresStatusMnemonicequivalence(uuid, type, parent, mnemonic,
                        namingConvention, holderRepositories, structureCommand, message, details, field);
            } else if (Type.DISCIPLINE.equals(type)) {
                message = TextUtil.DISCIPLINE_IS_NOT_CORRECT;

                // status, mnemonic
                validateStructuresStatusMnemonic(uuid, type, parent, mnemonic,
                        namingConvention, holderRepositories, structureCommand, message, details, field);

                // status, mnemonic equivalence
                validateStructuresStatusMnemonicequivalence(uuid, type, parent, mnemonic,
                        namingConvention, holderRepositories, structureCommand, message, details, field);
            } else if (Type.DEVICEGROUP.equals(type)) {
                // status, parent
                message = TextUtil.DISCIPLINE_IS_NOT_CORRECT;
                List<Discipline> disciplines = holderRepositories.disciplineRepository().readDisciplines(Status.APPROVED, false, parent.toString(), null, null, null, null, null);
                ExceptionUtil.validateConditionDataNotFoundException(ValidateUtil.isSize(disciplines, 1),
                        message, details, TextUtil.PARENT);

                // note rules for mnemonic for device group

                // mnemonic
                message = TextUtil.DEVICEGROUP_IS_NOT_CORRECT;
                ExceptionUtil.validateConditionDataConflictException(StringUtils.isEmpty(mnemonic), message, details, field);
            } else if (Type.DEVICETYPE.equals(type)) {
                // status, parent
                message = TextUtil.DEVICEGROUP_IS_NOT_CORRECT;
                validateStructuresParent(type, parent, Status.APPROVED, false,
                        holderRepositories, message, details, field);

                // status, mnemonic
                message = TextUtil.DEVICETYPE_IS_NOT_CORRECT;
                validateStructuresStatusMnemonic(uuid, type, parent, mnemonic,
                        namingConvention, holderRepositories, structureCommand, message, details, field);

                // status, mnemonic equivalence
                validateStructuresStatusMnemonicequivalence(uuid, type, parent, mnemonic,
                        namingConvention, holderRepositories, structureCommand, message, details, field);
            }
        } else if (StructureCommand.APPROVE.equals(structureCommand)) {
            // possibly additional checks to ensure (status, mnemonic) and (status, mnemonic equivalence)
            // however that would be duplicate checks as already checked at time of create, update

            if (Type.SYSTEMGROUP.equals(type)) {
                // status, uuid
                message = TextUtil.SYSTEMGROUP_IS_NOT_CORRECT;
                validateStructuresFound(uuid, type,
                        holderRepositories, message, details, field);
                validateStructuresValidSize1(uuid, type, Status.PENDING, null,
                        holderRepositories, message, details, field);
            } else if (Type.SYSTEM.equals(type)) {
                // status, uuid
                message = TextUtil.SYSTEM_IS_NOT_CORRECT;
                validateStructuresFound(uuid, type,
                        holderRepositories, message, details, field);
                validateStructuresValidSize1(uuid, type, Status.PENDING, null,
                        holderRepositories, message, details, field);

                List<System> systems = holderRepositories.systemRepository().readSystems(Status.PENDING, null, uuid.toString(), null, null, null, null, null, null);

                // status, parent
                message = TextUtil.SYSTEMGROUP_IS_NOT_CORRECT;
                validateStructuresFound(systems.get(0).getParentUuid(), Type.SYSTEMGROUP,
                        holderRepositories, message, details, field);
                validateStructuresValidSize1(systems.get(0).getParentUuid(), Type.SYSTEMGROUP, Status.APPROVED, false,
                        holderRepositories, message, details, field);
            } else if (Type.SUBSYSTEM.equals(type)) {
                // status, uuid
                message = TextUtil.SUBSYSTEM_IS_NOT_CORRECT;
                validateStructuresFound(uuid, type,
                        holderRepositories, message, details, field);
                validateStructuresValidSize1(uuid, type, Status.PENDING, null,
                        holderRepositories, message, details, field);

                List<Subsystem> subsystems = holderRepositories.subsystemRepository().readSubsystems(Status.PENDING, null, uuid.toString(), null, null, null, null, null, null);

                // status, parent
                message = TextUtil.SYSTEM_IS_NOT_CORRECT;
                validateStructuresFound(subsystems.get(0).getParentUuid(), Type.SYSTEM,
                        holderRepositories, message, details, field);
                validateStructuresValidSize1(subsystems.get(0).getParentUuid(), Type.SYSTEM, Status.APPROVED, false,
                        holderRepositories, message, details, field);
            } else if (Type.DISCIPLINE.equals(type)) {
                // status, uuid
                message = TextUtil.DISCIPLINE_IS_NOT_CORRECT;
                validateStructuresFound(uuid, type,
                        holderRepositories, message, details, field);
                validateStructuresValidSize1(uuid, type, Status.PENDING, null,
                        holderRepositories, message, details, field);
            } else if (Type.DEVICEGROUP.equals(type)) {
                // status, uuid
                message = TextUtil.DEVICEGROUP_IS_NOT_CORRECT;
                validateStructuresFound(uuid, type,
                        holderRepositories, message, details, field);
                validateStructuresValidSize1(uuid, type, Status.PENDING, null,
                        holderRepositories, message, details, field);

                List<DeviceGroup> deviceGroups = holderRepositories.deviceGroupRepository().readDeviceGroups(Status.PENDING, null, uuid.toString(), null, null, null, null, null, null);

                // status, parent
                message = TextUtil.DISCIPLINE_IS_NOT_CORRECT;
                validateStructuresFound(deviceGroups.get(0).getParentUuid(), Type.DISCIPLINE,
                        holderRepositories, message, details, field);
                validateStructuresValidSize1(deviceGroups.get(0).getParentUuid(), Type.DISCIPLINE, Status.APPROVED, false,
                        holderRepositories, message, details, field);
            } else if (Type.DEVICETYPE.equals(type)) {
                // status, uuid
                message = TextUtil.DEVICETYPE_IS_NOT_CORRECT;
                validateStructuresFound(uuid, type,
                        holderRepositories, message, details, field);
                validateStructuresValidSize1(uuid, type, Status.PENDING, null,
                        holderRepositories, message, details, field);

                List<DeviceType> deviceTypes = holderRepositories.deviceTypeRepository().readDeviceTypes(Status.PENDING, null, uuid.toString(), null, null, null, null, null, null);

                // status, parent
                message = TextUtil.DISCIPLINE_IS_NOT_CORRECT;
                validateStructuresFound(deviceTypes.get(0).getParentUuid(), Type.DEVICEGROUP,
                        holderRepositories, message, details, field);
                validateStructuresValidSize1(deviceTypes.get(0).getParentUuid(), Type.DEVICEGROUP, Status.APPROVED, false,
                        holderRepositories, message, details, field);
            }
        }
    }

    private static void validateStructuresParent(Type type, UUID parent, Status status, Boolean deleted,
            HolderRepositories holderRepositories, String message, String details, String field){
        // about finding out if structures are in place in order for structure operation to be allowed

        ExceptionUtil.validateConditionDataNotFoundException(
                getStructuresParentSize(type, parent, status, deleted, holderRepositories) == 1, message, details, field);
    }
    private static void validateStructuresStatusMnemonic(UUID uuid, Type type, UUID parent, String mnemonic,
            EssNamingConvention namingConvention, HolderRepositories holderRepositories, StructureCommand structureCommand,
            String message, String details, String field){
        // about finding out if structures are in place in order for structure operation to be allowed

        // parent may be null
        List<UUID> listUuid = null;

        listUuid = getStructures(type, parent, mnemonic, Status.PENDING, false, false, namingConvention, holderRepositories);
        validateConditionIfStructureCommand(StructureCommand.CREATE, structureCommand, listUuid.isEmpty(), message, details, field);
        validateConditionIfStructureCommand(StructureCommand.UPDATE, structureCommand, ValidateUtil.isEmptyOrEquals(listUuid, uuid), message, details, field);

        listUuid = getStructures(type, parent, mnemonic, Status.APPROVED, false, false, namingConvention, holderRepositories);
        validateConditionIfStructureCommand(StructureCommand.CREATE, structureCommand, listUuid.isEmpty(), message, details, field);
        validateConditionIfStructureCommand(StructureCommand.UPDATE, structureCommand, ValidateUtil.isEmptyOrEquals(listUuid, uuid), message, details, field);
    }
    private static void validateStructuresStatusMnemonicequivalence(UUID uuid, Type type, UUID parent, String mnemonic,
            EssNamingConvention namingConvention, HolderRepositories holderRepositories, StructureCommand structureCommand,
            String message, String details, String field){
        // about finding out if structures are in place in order for structure operation to be allowed

        // parent may be null
        List<UUID> listUuid = null;

        listUuid = getStructures(type, parent, mnemonic, Status.PENDING, false, true, namingConvention, holderRepositories);
        validateConditionIfStructureCommand(StructureCommand.CREATE, structureCommand, listUuid.isEmpty(), message, details, field);
        validateConditionIfStructureCommand(StructureCommand.UPDATE, structureCommand, ValidateUtil.isEmptyOrEquals(listUuid, uuid), message, details, field);

        listUuid = getStructures(type, parent, mnemonic, Status.APPROVED, false, true, namingConvention, holderRepositories);
        validateConditionIfStructureCommand(StructureCommand.CREATE, structureCommand, listUuid.isEmpty(), message, details, field);
        validateConditionIfStructureCommand(StructureCommand.UPDATE, structureCommand, ValidateUtil.isEmptyOrEquals(listUuid, uuid), message, details, field);
    }
    private static int getStructuresParentSize(Type type, UUID parent, Status status, Boolean deleted,
            HolderRepositories holderRepositories) {
        String queryValue = parent.toString();
        if (Type.SYSTEM.equals(type)) {
            return holderRepositories.systemGroupRepository().readSystemGroups(status, deleted, queryValue, null, null, null, null, null).size();
        } else if (Type.SUBSYSTEM.equals(type)) {
            return holderRepositories.systemRepository().readSystems(status, deleted, queryValue, null, null, null, null, null, null).size();
        } else if (Type.DEVICEGROUP.equals(type)) {
            return holderRepositories.disciplineRepository().readDisciplines(status, deleted, queryValue, null, null, null, null, null).size();
        } else if (Type.DEVICETYPE.equals(type)) {
            return holderRepositories.deviceGroupRepository().readDeviceGroups(status, deleted, queryValue, null, null, null, null, null, null).size();
        }
        return 0;
    }
    private static List<UUID> getStructures(Type type, UUID parent, String mnemonic, Status status, Boolean deleted, Boolean equivalence,
            EssNamingConvention namingConvention, HolderRepositories holderRepositories) {
        // get structures to be used in validation
        // may be difficult to understand, need context of how and where it is used
        //     before ok to create or approve, structures need to exist or not exist, this method helps to find out

        String queryUuid = null;
        String queryParent = null;
        String queryMnemonic = null;
        String queryMnemonicEquivalence = null;

        boolean hasMnemonic = !StringUtils.isEmpty(mnemonic);
        if (hasMnemonic) {
            if (equivalence) {
                queryMnemonicEquivalence = namingConvention.equivalenceClassRepresentative(mnemonic);
            } else {
                queryMnemonic = mnemonic;
            }
        }

        // uuid, parent, mnemonic
        boolean queryParentTable = false;
        if (ValidateUtil.isAnyEqual(type, Type.SYSTEMGROUP, Type.DISCIPLINE)) {
            // no parent - query table for type
            if (hasMnemonic) {
            } else {
                return Lists.newArrayList();
            }
        } else {
            // parent
            if (hasMnemonic) {
                // query table for type
                queryParent = parent.toString();
            } else {
                // query table for parent to type
                queryParentTable = true;
                queryUuid = parent.toString();
                queryParent = null;
                queryMnemonic = null;
                queryMnemonicEquivalence = null;
            }
        }

        // query table and return list with uuid
        if (Type.SYSTEMGROUP.equals(type)) {
            return StructureUtil.listSystemGroup2Uuid(
                    holderRepositories.systemGroupRepository().readSystemGroups(status, deleted, queryUuid, null, queryMnemonic, queryMnemonicEquivalence, null, null));
        } else if (Type.SYSTEM.equals(type)) {
            if (queryParentTable) {
                return StructureUtil.listSystemGroup2Uuid(holderRepositories.systemGroupRepository().readSystemGroups(status, deleted, queryUuid, null, queryMnemonic, queryMnemonicEquivalence, null, null));
            } else {
                return StructureUtil.listSystem2Uuid(holderRepositories.systemRepository().readSystems(status, deleted, queryUuid, queryParent, null, queryMnemonic, queryMnemonicEquivalence, null, null));
            }
        } else if (Type.SUBSYSTEM.equals(type)) {
            if (queryParentTable) {
                return StructureUtil.listSystem2Uuid(holderRepositories.systemRepository().readSystems(status, deleted, queryUuid, queryParent, null, queryMnemonic, queryMnemonicEquivalence, null, null));
            } else {
                return StructureUtil.listSubsystem2Uuid(holderRepositories.subsystemRepository().readSubsystems(status, deleted, queryUuid, queryParent, null, queryMnemonic, queryMnemonicEquivalence, null, null));
            }
        } else if (Type.DISCIPLINE.equals(type)) {
            return StructureUtil.listDiscipline2Uuid(holderRepositories.disciplineRepository().readDisciplines(status, deleted, queryUuid, null, queryMnemonic, queryMnemonicEquivalence, null, null));
        } else if (Type.DEVICEGROUP.equals(type)) {
            if (queryParentTable) {
                return StructureUtil.listDiscipline2Uuid(holderRepositories.disciplineRepository().readDisciplines(status, deleted, queryUuid, null, queryMnemonic, queryMnemonicEquivalence, null, null));
            } else {
                return StructureUtil.listDeviceGroup2Uuid(holderRepositories.deviceGroupRepository().readDeviceGroups(status, deleted, queryUuid, queryParent, null, queryMnemonic, queryMnemonicEquivalence, null, null));            }
        } else if (Type.DEVICETYPE.equals(type)) {
            if (queryParentTable) {
                return StructureUtil.listDeviceGroup2Uuid(holderRepositories.deviceGroupRepository().readDeviceGroups(status, deleted, queryUuid, queryParent, null, queryMnemonic, queryMnemonicEquivalence, null, null));
            } else {
                return StructureUtil.listDeviceType2Uuid(holderRepositories.deviceTypeRepository().readDeviceTypes(status, deleted, queryUuid, queryParent, null, queryMnemonic, queryMnemonicEquivalence, null, null));
            }
        }
        return Lists.newArrayList();
    }

    // ----------------------------------------------------------------------------------------------------

    /**
     * Validate structure data.
     * This method corresponds to structure element data for create, albeit in a different way.
     *
     * @param type type
     * @param mnemonicPath mnemonic path
     * @param namingConvention naming convention
     * @param holderIRepositories holder repositories
     * @param holderStructure holder of system and device structure content
     */
    public static void validateStructureDataCreate(Type type, String mnemonicPath, EssNamingConvention namingConvention, HolderIRepositories holderIRepositories, HolderSystemDeviceStructure holderStructure) {
        if (ValidateUtil.isAnyNull(type, mnemonicPath, namingConvention, holderIRepositories, holderStructure)) {
            return;
        }

        String details = mnemonicPath;
        String field = TextUtil.MNEMONIC_PATH;

        String[] path = NamingConventionUtil.string2MnemonicPath(mnemonicPath);
        String mnemonicpathEquivalence = namingConvention.equivalenceClassRepresentative(mnemonicPath);

        // validate path, including null check
        ExceptionUtil.validateConditionDataNotCorrectException(path != null && path.length >= 1 && path.length <= 2,
                TextUtil.STRUCTURE_MNEMONIC_PATH_IS_NOT_CORRECT, details, field);

        if (Type.SYSTEMGROUP.equals(type)) {
            validateStructuresMnemonicpathNotValid(path.length == 1, type, details, field);

            // system group may have empty path but there will be mnemonic path in this context

            // mnemonic path
            SystemGroup sg = holderIRepositories.systemGroupRepository().findLatestNotDeletedByMnemonic(path[0]);
            validateStructuresMnemonicpathDuplicate(sg == null, type, details, field);

            // mnemonic path equivalence
            List<SystemGroup> systemGroups = holderIRepositories.systemGroupRepository().findLatestNotDeleted();
            for (SystemGroup systemGroup : systemGroups) {
                validateStructuresMnemonicpathequivalenceDuplicate(
                        !StringUtils.equals(systemGroup.getMnemonicEquivalence(), namingConvention.equivalenceClassRepresentative(path[0])),
                        type, details, field);
            }
        } else if (Type.SYSTEM.equals(type)) {
            validateStructuresMnemonicpathNotValid(path.length == 1, type, details, field);

            // path with 1 element  - system group or  system ---> check both individually
            // check mnemonic, mnemonic equivalence

            // mnemonic path system group
            List<String> mnemonicPaths = StructureUtil.getMnemonicPathsSystemGroup(holderStructure, false, namingConvention);
            for (String existingPath : mnemonicPaths) {
                validateStructuresMnemonicpathDuplicate(
                        !StringUtils.equals(mnemonicPath, existingPath),
                        type, details, field);
            }

            // mnemonic path equivalence system group
            mnemonicPaths = StructureUtil.getMnemonicPathsSystemGroup(holderStructure, true, namingConvention);
            for (String existingPath : mnemonicPaths) {
                validateStructuresMnemonicpathequivalenceDuplicate(
                        !StringUtils.equals(mnemonicpathEquivalence, namingConvention.equivalenceClassRepresentative(existingPath)),
                        type, details, field);
            }

            // mnemonic path
            mnemonicPaths = StructureUtil.getMnemonicPathsSystem(holderStructure, false, namingConvention);
            for (String existingPath : mnemonicPaths) {
                validateStructuresMnemonicpathDuplicate(
                        !StringUtils.equals(mnemonicPath, existingPath),
                        type, details, field);
            }

            // mnemonic path equivalence
            mnemonicPaths = StructureUtil.getMnemonicPathsSystem(holderStructure, true, namingConvention);
            for (String existingPath : mnemonicPaths) {
                validateStructuresMnemonicpathequivalenceDuplicate(
                        !StringUtils.equals(mnemonicpathEquivalence, namingConvention.equivalenceClassRepresentative(existingPath)),
                        type, details, field);
            }
        } else if (Type.SUBSYSTEM.equals(type)) {
            validateStructuresMnemonicpathNotValid(path.length == 2, type, details, field);

            System sys = holderIRepositories.systemRepository().findLatestNotDeletedByMnemonic(path[0]);
            validateStructuresMnemonicpathNotFound(sys != null, type, details, field);
            SystemGroup sg = holderIRepositories.systemGroupRepository().findLatestNotDeletedByUuid(sys.getParentUuid().toString());

            // mnemonic path
            validateStructuresMnemonicpathDuplicate(!StringUtils.equals(path[0], path[1]), type, details, field);
            if (!StringUtils.isEmpty(sg.getMnemonic())) {
                validateStructuresMnemonicpathDuplicate(
                        !StringUtils.equals(sg.getMnemonic(), path[1]),
                        type, details, field);
            }

            List<String> mnemonicPaths = StructureUtil.getMnemonicPathsSubsystem(holderStructure, false, namingConvention);
            for (String existingPath : mnemonicPaths) {
                validateStructuresMnemonicpathDuplicate(
                        !StringUtils.equals(mnemonicPath, existingPath),
                        type, details, field);
            }

            // mnemonic path equivalence
            validateStructuresMnemonicpathDuplicate(
                    !StringUtils.equals(namingConvention.equivalenceClassRepresentative(path[0]), namingConvention.equivalenceClassRepresentative(path[1])),
                    type, details, field);
            if (!StringUtils.isEmpty(sg.getMnemonic())) {
                validateStructuresMnemonicpathDuplicate(
                        !StringUtils.equals(namingConvention.equivalenceClassRepresentative(sg.getMnemonic()), namingConvention.equivalenceClassRepresentative(path[1])),
                        type, details, field);
            }

            mnemonicPaths = StructureUtil.getMnemonicPathsSubsystem(holderStructure, true, namingConvention);
            for (String existingPath : mnemonicPaths) {
                validateStructuresMnemonicpathequivalenceDuplicate(
                        !StringUtils.equals(mnemonicpathEquivalence, namingConvention.equivalenceClassRepresentative(existingPath)),
                        type, details, field);
            }
        } else if (Type.DISCIPLINE.equals(type)) {
            validateStructuresMnemonicpathNotValid(path.length == 1, type, details, field);

            // mnemonic path
            Discipline di = holderIRepositories.disciplineRepository().findLatestNotDeletedByMnemonic(path[0]);
            validateStructuresMnemonicpathDuplicate(di == null, type, details, field);

            // mnemonic path equivalence
            List<Discipline> disciplines = holderIRepositories.disciplineRepository().findLatestNotDeleted();
            for (Discipline discipline : disciplines) {
                validateStructuresMnemonicpathequivalenceDuplicate(
                        !StringUtils.equals(discipline.getMnemonicEquivalence(), namingConvention.equivalenceClassRepresentative(path[0])),
                        type, details, field);
            }
        } else if (Type.DEVICEGROUP.equals(type)) {
            throw ExceptionUtil.createDataNotValidException(TextUtil.DEVICEGROUP_IS_NOT_VALID, details, field);
        } else if (Type.DEVICETYPE.equals(type)) {
            // since device group is between discipline and device type and device group has no mnemonic,
            //     it can not be traced to which device group that this device type belongs,
            //     therefore it can not be known to which mnemonic line it belongs
            //     by looking at mnemonic
            // instead similar approach as for other Type, i.e. compare with list of mnemonic paths
            //     rest of checks in validate create

            validateStructuresMnemonicpathNotValid(path.length == 2, type, details, field);

            // discipline
            Discipline discipline = holderIRepositories.disciplineRepository().findLatestNotDeletedByMnemonic(path[0]);
            ExceptionUtil.validateConditionDataNotFoundException(discipline != null,
                    TextUtil.DISCIPLINE_IS_NOT_FOUND, details, field);

            // mnemonic path
            validateStructuresMnemonicpathDuplicate(!StringUtils.equals(path[0], path[1]), type, details, field);

            List<String> mnemonicPaths = StructureUtil.getMnemonicPathsDeviceType(holderStructure, false, namingConvention);
            for (String existingPath : mnemonicPaths) {
                validateStructuresMnemonicpathDuplicate(
                        !StringUtils.equals(mnemonicPath, existingPath),
                        type, details, field);
            }

            // mnemonic path equivalence
            validateStructuresMnemonicpathDuplicate(
                    !StringUtils.equals(namingConvention.equivalenceClassRepresentative(path[0]), namingConvention.equivalenceClassRepresentative(path[1])),
                    type, details, field);

            mnemonicPaths = StructureUtil.getMnemonicPathsDeviceType(holderStructure, true, namingConvention);
            for (String existingPath : mnemonicPaths) {
                validateStructuresMnemonicpathequivalenceDuplicate(
                        !StringUtils.equals(mnemonicpathEquivalence, namingConvention.equivalenceClassRepresentative(existingPath)),
                        type, details, field);
            }
        }
    }

    private static void validateStructuresMnemonicpathNotFound(boolean condition, Type type, String details, String field) {
        ExceptionUtil.validateConditionDataNotFoundException(condition,
                type.toString().toLowerCase() + TextUtil.SPACE + TextUtil.MNEMONIC_PATH_IS_NOT_FOUND,
                details,
                field);
    }
    private static void validateStructuresMnemonicpathNotValid(boolean condition, Type type, String details, String field) {
        ExceptionUtil.validateConditionDataNotValidException(condition,
                type.toString().toLowerCase() + TextUtil.SPACE + TextUtil.MNEMONIC_PATH_IS_NOT_VALID,
                details,
                field);
    }
    private static void validateStructuresMnemonicpathDuplicate(boolean condition, Type type, String details, String field) {
        ExceptionUtil.validateConditionDataConflictException(condition,
                type.toString().toLowerCase() + TextUtil.SPACE + TextUtil.MNEMONIC_PATH_HAS_DUPLICATE,
                details,
                field);
    }
    private static void validateStructuresMnemonicpathequivalenceDuplicate(boolean condition, Type type, String details, String field) {
        ExceptionUtil.validateConditionDataConflictException(condition,
                type.toString().toLowerCase() + TextUtil.SPACE + TextUtil.MNEMONIC_PATH_EQUIVALENCE_HAS_DUPLICATE,
                details,
                field);
    }

    // ----------------------------------------------------------------------------------------------------

    /**
     * Validate condition if precondition is fulfilled and throw
     * data conflict exception with reason if validation fails.
     *
     * @param precondition precondition
     * @param condition condition
     * @param message message
     * @param details details
     * @param field field
     *
     * @see DataConflictException
     */
    private static void validateConditionIfPrecondition(boolean precondition,
            boolean condition, String message, String details, String field) {
        if (precondition) {
            ExceptionUtil.validateConditionDataConflictException(condition, message, details, field);
        }
    }

    /**
     * Validate condition if precondition (StructureCommand) is fulfilled and
     * throw data conflict exception with reason if validation fails.
     *
     * @param expected expected structure command
     * @param actual actual structure command
     * @param condition condition
     * @param message message
     * @param details details
     * @param field field
     *
     * @see DataConflictException
     */
    private static void validateConditionIfStructureCommand(StructureCommand expected, StructureCommand actual,
            boolean condition, String message, String details, String field) {
        validateConditionIfPrecondition(expected != null && expected.equals(actual),
                condition, message, details, field);
    }

}
