/*
 * Copyright (C) 2022 European Spallation Source ERIC.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.openepics.names.docker;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.junit.jupiter.api.Assertions.fail;

import java.io.IOException;
import java.net.HttpURLConnection;
import java.util.Map;
import java.util.TreeMap;
import java.util.UUID;
import java.util.Map.Entry;

import org.openepics.names.docker.ITUtil.AuthorizationChoice;
import org.openepics.names.docker.ITUtil.EndpointChoice;
import org.openepics.names.rest.beans.Status;
import org.openepics.names.rest.beans.element.NameElement;
import org.openepics.names.rest.beans.element.NameElementCommand;
import org.openepics.names.rest.beans.element.NameElementCommandConfirm;
import org.openepics.names.rest.beans.element.NameElementCommandCreate;
import org.openepics.names.rest.beans.element.NameElementCommandUpdate;
import org.openepics.names.rest.beans.response.ResponseBoolean;
import org.openepics.names.rest.beans.response.ResponseBooleanList;
import org.openepics.names.rest.beans.response.ResponsePageNameElements;
import org.openepics.names.util.NameCommand;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;

/**
 * Utility class to help (Docker) integration tests for Naming and PostgreSQL.
 *
 * @author Lars Johansson
 */
public class ITUtilNameElement {

    static final ObjectMapper mapper = new ObjectMapper();

    /**
     * This class is not to be instantiated.
     */
    private ITUtilNameElement() {
        throw new IllegalStateException("Utility class");
    }

    // ----------------------------------------------------------------------------------------------------

    /**
     * Return string for name element command array.
     *
     * @param value name element command array
     * @return string for name element command array
     */
    static String object2Json(NameElementCommandCreate[] value) {
        try {
            return mapper.writeValueAsString(value);
        } catch (JsonProcessingException e) {
            fail();
        }
        return null;
    }
    /**
     * Return string for name element command array.
     *
     * @param value name element command array
     * @return string for name element command array
     */
    static String object2Json(NameElementCommandUpdate[] value) {
        try {
            return mapper.writeValueAsString(value);
        } catch (JsonProcessingException e) {
            fail();
        }
        return null;
    }
    /**
     * Return string for name element command array.
     *
     * @param value name element command array
     * @return string for name element command array
     */
    static String object2Json(NameElementCommandConfirm[] value) {
        try {
            return mapper.writeValueAsString(value);
        } catch (JsonProcessingException e) {
            fail();
        }
        return null;
    }

    // ----------------------------------------------------------------------------------------------------

    /**
     * Assert that array of name elements has expected length.
     *
     * @param actual array of name elements
     * @param expectedLength expected length of array
     */
    static void assertContentLength(NameElement[] actual, int expectedLength) {
        assertNotNull(actual);
        assertEquals(expectedLength, actual.length);
    }

    /**
     * Assert that array of name elements has length one and name element is available.
     *
     * @param actual array of name elements
     * @return name element
     */
    static NameElement assertContentLengthOne(NameElement[] actual) {
        assertContentLength(actual, 1);
        assertNotNull(actual[0]);
        return actual[0];
    }

    /**
     * Assert that response page with list of name elements has expected length.
     *
     * @param actual response page with list of name elements
     * @param expectedLength expected length of list
     */
    static void assertContentLength(ResponsePageNameElements actual, int expectedLength) {
        assertNotNull(actual);
        assertEquals(expectedLength, actual.getList().size());
        assertEquals(expectedLength, actual.getListSize());
    }

    /**
     * Assert that response page with list of name elements has length one and name element is available.
     *
     * @param actual response page with list of name elements
     * @return name element
     */
    static NameElement assertContentLengthOne(ResponsePageNameElements actual) {
        assertContentLength(actual, 1);
        assertNotNull(actual.getList().get(0));
        return actual.getList().get(0);
    }

    // ----------------------------------------------------------------------------------------------------

    /**
     * Assert that arrays are equal with same length and same content in each array position.
     *
     * @param actual actual array of name elements
     * @param expected expected arbitrary number of name elements
     */
    static void assertEqualsNameElements(NameElement[] actual, NameElement... expected) {
        if (expected != null) {
            assertNotNull(actual);
            assertEquals(expected.length, actual.length);
            for (int i=0; i<expected.length; i++) {
                assertTrue(expected[i].equals(actual[i]));
            }
        } else {
            assertNull(actual);
        }
    }

    /**
     * Assert content of name element.
     *
     * @param actual name element
     * @param expectedUuid expected uuid
     * @param expectedParentSystemStructure expected parent system structure uuid (system group, system, subsystem)
     * @param expectedParentDeviceStructure parent device structure uuid (device type)
     * @param expectedSystemStructure expected system structure
     * @param expectedDeviceStructure expected device structure
     * @param expectedIndex expected index
     * @param expectedName expected name
     * @param expectedDescription expected description
     * @param expectedStatus expected status
     * @param expectedIsLatest expected latest
     * @param expectedIsDeleted expected deleted
     * @param expectedWho expected who
     * @param expectedComment expected comment
     */
    static void assertContent(NameElement actual,
            UUID expectedUuid, UUID expectedParentSystemStructure, UUID expectedParentDeviceStructure,
            String expectedSystemStructure, String expectedDeviceStructure,
            String expectedIndex, String expectedName,
            String expectedDescription, Status expectedStatus, Boolean expectedIsLatest, Boolean expectedIsDeleted,
            String expectedWho, String expectedComment) {
        assertNotNull(actual);
        assertEquals(expectedUuid, actual.getUuid());
        assertEquals(expectedParentSystemStructure, actual.getParentSystemStructure());
        assertEquals(expectedParentDeviceStructure, actual.getParentDeviceStructure());
        assertEquals(expectedSystemStructure, actual.getSystemStructure());
        assertEquals(expectedDeviceStructure, actual.getDeviceStructure());
        assertEquals(expectedIndex, actual.getIndex());
        assertEquals(expectedName, actual.getName());
        assertEquals(expectedDescription, actual.getDescription());
        assertEquals(expectedStatus, actual.getStatus());
        assertEquals(expectedIsLatest, actual.isLatest());
        assertEquals(expectedIsDeleted, actual.isDeleted());
        assertNotNull(actual.getWhen());
        assertEquals(expectedWho, actual.getWho());
        assertEquals(expectedComment, actual.getComment());
    }

    /**
     * Assert content of name element.
     *
     * @param actual name element
     * @param expected name element command
     */
    static void assertContent(NameElement actual, NameElementCommand expected) {
        assertNotNull(actual);
        if (expected != null) {
            assertEquals(expected.getUuid(), actual.getUuid());
            assertEquals(expected.getParentSystemStructure(), actual.getParentSystemStructure());
            assertEquals(expected.getParentDeviceStructure(), actual.getParentDeviceStructure());
            assertEquals(expected.getIndex(), actual.getIndex());
            assertEquals(expected.getDescription(), actual.getDescription());
            assertEquals(expected.getComment(), actual.getComment());
        }
    }

    // ----------------------------------------------------------------------------------------------------

    /**
     * Utility method to read name elements.
     *
     * @param queryString query string
     * @return response page name elements
     */
    public static ResponsePageNameElements assertRead(String queryString) {
        return assertRead(queryString, -1, -1);
    }
    /**
     * Utility method to read name elements and assert number of items in list.
     *
     * @param queryString query string
     * @param expectedEqual (if non-negative number) equal to this number of items
     * @return response page name elements
     */
    public static ResponsePageNameElements assertRead(String queryString, int expectedEqual) {
        return assertRead(queryString, expectedEqual, expectedEqual);
    }
    /**
     * Utility method to read name elements and assert number of items in list.
     *
     * @param queryString query string
     * @param expectedGreaterThanOrEqual (if non-negative number) greater than or equal to this number of items
     * @param expectedLessThanOrEqual (if non-negative number) less than or equal to this number of items
     * @return response page name elements
     */
    public static ResponsePageNameElements assertRead(String queryString, int expectedGreaterThanOrEqual, int expectedLessThanOrEqual) {
        try {
            String[] response = null;
            ResponsePageNameElements responsePageNameElements = null;

            response = ITUtil.doGetJson(ITUtil.HTTP_IP_PORT_NAMING_API_V1_NAMES + queryString);
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_OK);
            responsePageNameElements = mapper.readValue(response[1], ResponsePageNameElements.class);

            // expected number of items in list
            //     (if non-negative number)
            //     expectedGreaterThanOrEqual <= nbr of items <= expectedLessThanOrEqual
            if (expectedGreaterThanOrEqual >= 0) {
                assertTrue(responsePageNameElements.getListSize() >= expectedGreaterThanOrEqual);
            }
            if (expectedLessThanOrEqual >= 0) {
                assertTrue(responsePageNameElements.getListSize() <= expectedLessThanOrEqual);
            }

            // expected value for latest for items in list
            assertLatest(responsePageNameElements);

            return responsePageNameElements;
        } catch (IOException e) {
            fail();
        } catch (Exception e) {
            fail();
        }
        return null;
    }

    /**
     * Assert that response page with list of name elements has maximum one entry that is latest
     * for any given uuid, which may not have status {@link Status#PENDING}.
     *
     * @param actual response page with list of name elements
     */
    public static void assertLatest(ResponsePageNameElements actual) {
        // latest
        //     status pending not with latest
        //     maximum one item for any given uuid in list may be latest

        assertNotNull(actual);
        Map<UUID, Long> mapUuidCountLatest = new TreeMap<>();
        for (NameElement nameElement : actual.getList()) {
            if (nameElement.isLatest()) {
                ITUtil.addOne(nameElement.getUuid(), mapUuidCountLatest);
                assertNotEquals(Status.PENDING, nameElement.getStatus());
            }
        }
        for (Entry<UUID, Long> entry : mapUuidCountLatest.entrySet()) {
            assertTrue(entry.getValue() <= 1);
        }
    }

    /**
     * Utility method to get name equivalence of a mnemonic or name.
     *
     * @param name name
     * @param expected expected response
     */
    public static void assertEquivalence(String name, String expected) {
        try {
            String[] response = ITUtil.doGetJson(ITUtil.HTTP_IP_PORT_NAMING_API_V1_NAMES + "/equivalence/" + name);
            ITUtil.assertResponseLength2CodeOKContent(response, expected);
        } catch (IOException e) {
            fail();
        } catch (Exception e) {
            fail();
        }
    }

    /**
     * Utility method to check if name exists.
     *
     * @param name name
     * @param expected expected response
     */
    public static void assertExists(String name, Boolean expected) {
        assertExists(name, expected, null);
    }
    /**
     * Utility method to check if name exists.
     *
     * @param name name
     * @param expected expected response
     * @param expectedMessageEmpty expected message empty
     */
    public static void assertExists(String name, Boolean expected, Boolean expectedMessageEmpty) {
        try {
            String[] response = null;

            response = ITUtil.doGetJson(ITUtil.HTTP_IP_PORT_NAMING_API_V1_NAMES + "/exists/" + name);
            ITUtil.assertResponseLength2CodeOK(response);
            if (expectedMessageEmpty != null) {
                ITUtil.assertEqualsResponseBoolean(mapper.readValue(response[1], ResponseBoolean.class), expected, expectedMessageEmpty);
            } else {
                ITUtil.assertEqualsResponseBoolean(mapper.readValue(response[1], ResponseBoolean.class), expected);
            }
        } catch (IOException e) {
            fail();
        } catch (Exception e) {
            fail();
        }
    }

    /**
     * Utility method to check if name is legacy name.
     *
     * @param name name
     * @param expected expected response
     */
    public static void assertIsLegacy(String name, Boolean expected) {
        try {
            String[] response = null;

            response = ITUtil.doGetJson(ITUtil.HTTP_IP_PORT_NAMING_API_V1_NAMES + "/isLegacy/" + name);
            ITUtil.assertResponseLength2CodeOK(response);
            ITUtil.assertEqualsResponseBoolean(mapper.readValue(response[1], ResponseBoolean.class), expected);
        } catch (IOException e) {
            fail();
        } catch (Exception e) {
            fail();
        }
    }

    /**
     * Utility method to check if name is valid to create.
     *
     * @param name name
     * @param expected expected response
     */
    public static void assertIsValidToCreate(String name, Boolean expected) {
        try {
            String[] response = null;

            response = ITUtil.doGetJson(ITUtil.HTTP_IP_PORT_NAMING_API_V1_NAMES + "/isValidToCreate/" + name);
            ITUtil.assertResponseLength2CodeOK(response);
            ITUtil.assertEqualsResponseBoolean(mapper.readValue(response[1], ResponseBoolean.class), expected);
        } catch (IOException e) {
            fail();
        } catch (Exception e) {
            fail();
        }
    }

    // ----------------------------------------------------------------------------------------------------

    /**
     * Utility method to validate name element command for create and assert expected response.
     *
     * @param nameElementCommand name element command
     * @param nameCommand name command
     * @param index index
     * @param expected expected response
     */
    public static void assertValidate(NameElementCommandCreate nameElementCommand, NameCommand nameCommand, String index, Boolean expected) {
        nameElementCommand.setIndex(index);
        assertValidate(AuthorizationChoice.NONE, object2Json(new NameElementCommandCreate[] {nameElementCommand}), nameCommand, HttpURLConnection.HTTP_OK, expected);
    }
    /**
     * Utility method to validate name element command for create and assert expected response.
     *
     * @param nameElementCommand name element command
     * @param nameCommand name command
     * @param expected expected response
     */
    public static void assertValidate(NameElementCommandCreate nameElementCommand, NameCommand nameCommand, Boolean expected) {
        assertValidate(AuthorizationChoice.NONE, object2Json(new NameElementCommandCreate[] {nameElementCommand}), nameCommand, HttpURLConnection.HTTP_OK, expected);
    }
    /**
     * Utility method to validate name element command for update and assert expected response.
     *
     * @param nameElementCommand name element command
     * @param nameCommand name command
     * @param expected expected response
     */
    public static void assertValidate(NameElementCommandUpdate nameElementCommand, NameCommand nameCommand, Boolean expected) {
        assertValidate(AuthorizationChoice.NONE, object2Json(new NameElementCommandUpdate[] {nameElementCommand}), nameCommand, HttpURLConnection.HTTP_OK, expected);
    }
    /**
     * Utility method to validate name element command for confirm and assert expected response.
     *
     * @param nameElementCommand name element command
     * @param nameCommand name command
     * @param expected expected response
     */
    public static void assertValidate(NameElementCommandConfirm nameElementCommand, NameCommand nameCommand, Boolean expected) {
        assertValidate(AuthorizationChoice.NONE, object2Json(new NameElementCommandConfirm[] {nameElementCommand}), nameCommand, HttpURLConnection.HTTP_OK, expected);
    }
    /**
     * Utility method to validate name element commands for create and assert expected response.
     *
     * @param nameElementCommands name element commands
     * @param nameCommand name command
     * @param expected expected response
     */
    public static void assertValidate(NameElementCommandCreate[] nameElementCommands, NameCommand nameCommand, Boolean expected) {
        assertValidate(AuthorizationChoice.NONE, object2Json(nameElementCommands), nameCommand, HttpURLConnection.HTTP_OK, expected);
    }
    /**
     * Utility method to validate name element commands for update and assert expected response.
     *
     * @param nameElementCommands name element commands
     * @param nameCommand name command
     * @param expected expected response
     */
    public static void assertValidate(NameElementCommandUpdate[] nameElementCommands, NameCommand nameCommand, Boolean expected) {
        assertValidate(AuthorizationChoice.NONE, object2Json(nameElementCommands), nameCommand, HttpURLConnection.HTTP_OK, expected);
    }
    /**
     * Utility method to validate name element commands for confirm and assert expected response.
     *
     * @param nameElementCommands name element commands
     * @param nameCommand name command
     * @param expected expected response
     */
    public static void assertValidate(NameElementCommandConfirm[] nameElementCommands, NameCommand nameCommand, Boolean expected) {
        assertValidate(AuthorizationChoice.NONE, object2Json(nameElementCommands), nameCommand, HttpURLConnection.HTTP_OK, expected);
    }
    /**
     * Utility method to validate json and assert expected response.
     *
     * @param json json
     * @param nameCommand name command
     * @param responseCode response code
     */
    public static void assertValidate(String json, NameCommand nameCommand, int responseCode) {
        assertValidate(AuthorizationChoice.NONE, json, nameCommand, responseCode, null);
    }
    /**
     * Utility method to validate json and assert expected response.
     *
     * @param authorizationChoice authorization choice (none, user, admin)
     * @param json json
     * @param nameCommand name command
     * @param responseCode response code
     * @param expected expected response
     */
    public static void assertValidate(AuthorizationChoice authorizationChoice, String json, NameCommand nameCommand, int responseCode, Boolean expected) {
        String path = getValidatePath(nameCommand);

        try {
            String[] response = null;

            response = ITUtil.runShellCommand(ITUtil.curlGetPathJson(authorizationChoice, EndpointChoice.NAMES, path, json));
            ITUtil.assertResponseLength2Code(response, responseCode);
            if (expected != null) {
                ITUtil.assertEqualsResponseBoolean(mapper.readValue(response[1], ResponseBooleanList.class), expected);
            }
        } catch (IOException e) {
            fail();
        } catch (InterruptedException e) {
            fail();
        } catch (Exception e) {
            fail();
        }
    }

    /**
     * Utility method to return validate path for name command.
     *
     * @param nameCommand name command
     * @return validate path
     */
    private static String getValidatePath(NameCommand nameCommand) {
        String validatePath = "";
        switch (nameCommand) {
            case CREATE:
                validatePath = "/validateCreate";
                break;
            case UPDATE:
                validatePath = "/validateUpdate";
                break;
            case DELETE:
                validatePath = "/validateDelete";
                break;
            default:
                break;
        }
        return validatePath;
    }

    // ----------------------------------------------------------------------------------------------------

    /**
     * Utility method to create a name element and assert result.
     *
     * @param nameElementCommand name element command
     * @return created name element
     */
    public static NameElement assertCreate(NameElementCommandCreate nameElementCommand) {
        NameElement[] createdNameElements = assertCreate(AuthorizationChoice.NONE, object2Json(new NameElementCommandCreate[] {nameElementCommand}), HttpURLConnection.HTTP_CREATED);
        return createdNameElements != null
                ? createdNameElements[0]
                : null;
    }
    /**
     * Utility method to create a name element and assert result and response code.
     *
     * @param nameElementCommand name element command
     * @param responseCode response code
     * @return created name element
     */
    public static NameElement assertCreate(NameElementCommandCreate nameElementCommand, int responseCode) {
        NameElement[] createdNameElements = assertCreate(AuthorizationChoice.NONE, object2Json(new NameElementCommandCreate[] {nameElementCommand}), responseCode);
        return createdNameElements != null
                ? createdNameElements[0]
                : null;
    }
    /**
     * Utility method to create name elements and assert result.
     *
     * @param nameElementCommands name element commands
     * @return created name element
     */
    public static NameElement[] assertCreate(NameElementCommandCreate[] nameElementCommands) {
        return assertCreate(AuthorizationChoice.NONE, object2Json(nameElementCommands), HttpURLConnection.HTTP_CREATED);
    }
    /**
     * Utility method to create name elements and assert result and response code.
     *
     * @param json json
     * @param responseCode response code
     * @return created name element
     */
    public static NameElement[] assertCreate(String json, int responseCode) {
        return assertCreate(AuthorizationChoice.NONE, json, responseCode);
    }
    /**
     * Utility method to create json (name elements) and assert result and response code.
     *
     * @param authorizationChoice authorization choice (none, user, admin)
     * @param json json
     * @param responseCode response code
     * @return created name element
     */
    public static NameElement[] assertCreate(AuthorizationChoice authorizationChoice, String json, int responseCode) {
        NameElement[] createdNameElements = null;

        try {
            String[] response = null;

            response = ITUtil.runShellCommand(ITUtil.curlPostPathJson(authorizationChoice, EndpointChoice.NAMES, "", json));
            ITUtil.assertResponseLength2Code(response, responseCode);

            if (HttpURLConnection.HTTP_CREATED == responseCode) {
                createdNameElements = mapper.readValue(response[1], NameElement[].class);

                for (NameElement createdNameElement : createdNameElements) {
                    assertNotNull(createdNameElement.getUuid());
                    assertEquals(Boolean.TRUE, createdNameElement.isLatest());
                    assertEquals(Boolean.FALSE, createdNameElement.isDeleted());
                    assertNotNull(createdNameElement.getWhen());
                }

                // not set uuid for input parameter(s)
                //     convenience, easy when array length 1, otherwise not easy
            }
        } catch (IOException e) {
            fail();
        } catch (InterruptedException e) {
            fail();
        } catch (Exception e) {
            fail();
        }

        return createdNameElements;
    }

    // ----------------------------------------------------------------------------------------------------

    /**
     * Utility method to update a name element and assert result.
     *
     * @param nameElementCommand name element command
     * @return updated name element
     */
    public static NameElement assertUpdate(NameElementCommandUpdate nameElementCommand) {
        NameElement[] updatedNameElements = assertUpdate(AuthorizationChoice.NONE, object2Json(new NameElementCommandUpdate[] {nameElementCommand}), HttpURLConnection.HTTP_OK);
        return updatedNameElements != null
                ? updatedNameElements[0]
                : null;
    }
    /**
     * Utility method to update a name element and assert result and response code.
     *
     * @param nameElementCommand name element command
     * @param responseCode response code
     * @return updated name element
     */
    public static NameElement assertUpdate(NameElementCommandUpdate nameElementCommand, int responseCode) {
        NameElement[] updatedNameElements = assertUpdate(AuthorizationChoice.NONE, object2Json(new NameElementCommandUpdate[] {nameElementCommand}), responseCode);
        return updatedNameElements != null
                ? updatedNameElements[0]
                : null;
    }
    /**
     * Utility method to update name elements and assert result.
     *
     * @param nameElementCommands name element commands
     * @return updated name element
     */
    public static NameElement[] assertUpdate(NameElementCommandUpdate[] nameElementCommands) {
        return assertUpdate(AuthorizationChoice.NONE, object2Json(nameElementCommands), HttpURLConnection.HTTP_OK);
    }
    /**
     * Utility method to update json (name elements) and assert result and response code.
     *
     * @param authorizationChoice authorization choice (none, user, admin)
     * @param json json
     * @param responseCode response code
     * @return updated name element
     */
    public static NameElement[] assertUpdate(AuthorizationChoice authorizationChoice, String json, int responseCode) {
        NameElement[] updatedNameElements = null;

        try {
            String[] response = null;

            response = ITUtil.runShellCommand(ITUtil.curlPutPathJson(authorizationChoice, EndpointChoice.NAMES, "", json));
            ITUtil.assertResponseLength2Code(response, responseCode);

            if (HttpURLConnection.HTTP_OK == responseCode) {
                updatedNameElements = mapper.readValue(response[1], NameElement[].class);

                for (NameElement updatedNameElement : updatedNameElements) {
                    assertNotNull(updatedNameElement.getUuid());
                    assertEquals(Boolean.TRUE, updatedNameElement.isLatest());
                    assertEquals(Boolean.FALSE, updatedNameElement.isDeleted());
                    assertNotNull(updatedNameElement.getWhen());
                }
            }
        } catch (IOException e) {
            fail();
        } catch (InterruptedException e) {
            fail();
        } catch (Exception e) {
            fail();
        }

        return updatedNameElements;
    }

    // ----------------------------------------------------------------------------------------------------

    /**
     * Utility method to delete a name element and assert result.
     *
     * @param nameElementCommand name element command
     * @return deleted name element
     */
    public static NameElement assertDelete(NameElementCommandConfirm nameElementCommand) {
        NameElement[] deletedNameElements = assertDelete(AuthorizationChoice.NONE, object2Json(new NameElementCommandConfirm[] {nameElementCommand}), HttpURLConnection.HTTP_OK);
        return deletedNameElements != null
                ? deletedNameElements[0]
                : null;
    }
    /**
     * Utility method to delete a name element and assert result and response code.
     *
     * @param nameElementCommand name element command
     * @param responseCode response code
     * @return deleted name element
     */
    public static NameElement assertDelete(NameElementCommandConfirm nameElementCommand, int responseCode) {
        NameElement[] deletedNameElements = assertDelete(AuthorizationChoice.NONE, object2Json(new NameElementCommandConfirm[] {nameElementCommand}), responseCode);
        return deletedNameElements != null
                ? deletedNameElements[0]
                : null;
    }
    /**
     * Utility method to delete name elements and assert result.
     *
     * @param nameElementCommands name element commands
     * @return deleted name elements
     */
    public static NameElement[] assertDelete(NameElementCommandConfirm[] nameElementCommands) {
        return assertDelete(AuthorizationChoice.NONE, object2Json(nameElementCommands), HttpURLConnection.HTTP_OK);
    }
    /**
     * Utility method to delete json (name elements) and assert result and response code.
     *
     * @param authorizationChoice authorization choice (none, user, admin)
     * @param json json
     * @param responseCode response code
     * @return deleted name elements
     */
    public static NameElement[] assertDelete(AuthorizationChoice authorizationChoice, String json, int responseCode) {
        NameElement[] deletedNameElements = null;

        try {
            String[] response = null;

            response = ITUtil.runShellCommand(ITUtil.curlDeletePathJson(authorizationChoice, EndpointChoice.NAMES, "", json));
            ITUtil.assertResponseLength2Code(response, responseCode);

            if (HttpURLConnection.HTTP_OK == responseCode) {
                deletedNameElements = mapper.readValue(response[1], NameElement[].class);

                for (NameElement deletedNameElement : deletedNameElements) {
                    assertNotNull(deletedNameElement.getUuid());
                    assertEquals(Boolean.TRUE, deletedNameElement.isLatest());
                    assertEquals(Boolean.TRUE, deletedNameElement.isDeleted());
                    assertNotNull(deletedNameElement.getWhen());
                }

            }
        } catch (IOException e) {
            fail();
        } catch (InterruptedException e) {
            fail();
        } catch (Exception e) {
            fail();
        }

        return deletedNameElements;
    }

}
