/*
 * Copyright (C) 2022 European Spallation Source ERIC.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.openepics.names.docker;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.fail;

import java.io.File;
import java.io.IOException;
import java.net.HttpURLConnection;
import java.util.UUID;

import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;
import org.openepics.names.rest.beans.Type;
import org.openepics.names.rest.beans.element.NameElement;
import org.openepics.names.rest.beans.element.NameElementCommand;
import org.openepics.names.rest.beans.element.NameElementCommandConfirm;
import org.openepics.names.rest.beans.element.NameElementCommandCreate;
import org.openepics.names.rest.beans.element.NameElementCommandUpdate;
import org.openepics.names.rest.beans.element.StructureElement;
import org.openepics.names.rest.beans.element.StructureElementCommandConfirm;
import org.openepics.names.rest.beans.element.StructureElementCommandCreate;
import org.openepics.names.util.NameCommand;
import org.openepics.names.util.NameElementUtil;
import org.testcontainers.containers.DockerComposeContainer;
import org.testcontainers.containers.wait.strategy.Wait;
import org.testcontainers.junit.jupiter.Container;
import org.testcontainers.junit.jupiter.Testcontainers;

import com.fasterxml.jackson.databind.ObjectMapper;

/**
 * Integration tests for Naming and PostgreSQL that make use of existing dockerization
 * with docker-compose.yml / Dockerfile.
 *
 * <p>
 * Focus of this class is names endpoint.
 * </p>
 *
 * @author Lars Johansson
 */
@Testcontainers
class NamesIT {

    // note
    //     NameElement extends NameElementCommand, StructureElement extends StructureElementCommand
    //     -->
    //     for such operations that send Command objects to backend, it is also possible to send
    //     NameElement, StructureElement
    //
    //     NameElementCommand
    //	       create -       parentSystemStructure, parentDeviceStructure, index, description, comment
    //	       update - uuid, parentSystemStructure, parentDeviceStructure, index, description, comment
    //	       delete - uuid,                                                                   comment
    //
    //     response status codes

    @Container
    public static final DockerComposeContainer<?> ENVIRONMENT =
        new DockerComposeContainer<>(new File("docker-compose-integrationtest.yml"))
            .waitingFor(ITUtil.NAMING, Wait.forLogMessage(".*Started NamingApplication.*", 1));

    private static UUID systemGroupAcc  = null;
    private static UUID systemRFQ       = null;
    private static UUID subsystem010PRL = null;
    private static UUID subsystem010    = null;
    private static UUID subsystemN1U1   = null;

    private static UUID disciplineEMR   = null;
    private static UUID deviceGroupEMR  = null;
    private static UUID deviceTypeFS    = null;
    private static UUID deviceTypeRFA   = null;
    private static UUID deviceTypeTT    = null;

    @BeforeAll
    public static void initAll() {
        // init system group, system, subsystem, discipline, device group, device type

        StructureElementCommandCreate structureElementCommandCreate = null;
        StructureElement approvedStructureElement = null;

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "Accelerator", "Acc",
                "The ESS Linear Accelerator", "approved by alfio");
        approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        systemGroupAcc = approvedStructureElement.getUuid();

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEM, systemGroupAcc,
                "Radio Frequency Quadrupole", "RFQ",
                "empty", "empty");
        approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        systemRFQ = approvedStructureElement.getUuid();

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SUBSYSTEM, systemRFQ,
                "01 Phase Reference Line", "010PRL",
                "empty", "Approved by Daniel Piso");
        approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        subsystem010PRL = approvedStructureElement.getUuid();

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SUBSYSTEM, systemRFQ,
                "RFQ-010", "010",
                "empty", "Approved by Daniel Piso");
        approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        subsystem010 = approvedStructureElement.getUuid();

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SUBSYSTEM, systemRFQ,
                "Power switch board 01", "N1U1",
                "Electrical power cabinets", "Approved by Daniel Piso");
        approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        subsystemN1U1 = approvedStructureElement.getUuid();

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DISCIPLINE, null,
                "Electromagnetic Resonators", "EMR",
                "empty", "empty");
        approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        disciplineEMR = approvedStructureElement.getUuid();

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICEGROUP, disciplineEMR,
                "Control", null,
                "empty", "These names are needed now, so I am approving, but please add a description to these later.");
        approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        deviceGroupEMR = approvedStructureElement.getUuid();

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupEMR,
                "Flow Switch", "FS",
                "empty", "Approve names added from misc device group");
        approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        deviceTypeFS = approvedStructureElement.getUuid();

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupEMR,
                "RF Antenna", "RFA",
                "empty", "Approve names added from misc device group");
        approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        deviceTypeRFA = approvedStructureElement.getUuid();

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupEMR,
                "Temperature Transmitter", "TT",
                "empty", "Approve names added from misc device group");
        approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        deviceTypeTT = approvedStructureElement.getUuid();
    }

    @Test
    void equivalenceName() {
        // purpose
        //     test name equivalence
        //
        // what
        //     read        equivalence name

        ITUtilNameElement.assertEquivalence("A2T-010PRL:RFS-PRLTap-071", "A2T-10PR1:RFS-PR1TAP-71");
    }

    @Test
    void existsName() {
        // purpose
        //     test if name exists
        //
        // what
        //     read        exists name

        ITUtilNameElement.assertExists("RFQ-010PRL:EMR-RFA-071", Boolean.FALSE);
    }

    @Test
    void legacyName() {
        // purpose
        //     test if name is legacy name
        //
        // what - combination of
        //     read        exists name
        //     read        is legacy name
        //     read        is valid to create name
        //     create      create names
        //     delete      delete structures
        //     patch       approve structures

        try {
            NameElementCommandCreate nameElementCommandCreate = null;
            NameElementCommandUpdate nameElementCommandUpdate = null;
            NameElementCommandConfirm nameElementCommandConfirm = null;
            NameElementCommand nameElementCommand = null;
            NameElement createdNameElement = null;
            StructureElementCommandConfirm structureElementCommandConfirm = null;

            nameElementCommand = new NameElementCommand(
                    null, subsystemN1U1, deviceTypeRFA, "071",
                    "description", "comment");
            nameElementCommandCreate = NameElementUtil.convertCommand2CommandCreate(nameElementCommand);
            nameElementCommandUpdate = NameElementUtil.convertCommand2CommandUpdate(nameElementCommand);
            nameElementCommandConfirm = NameElementUtil.convertCommand2CommandConfirm(nameElementCommand);

            Long totalCount = ITUtilNameElement.assertRead("").getTotalCount();
            assertNotNull(totalCount);

            ITUtilNameElement.assertExists("RFQ-N1U1:EMR-RFA-071", Boolean.FALSE);
            ITUtilNameElement.assertIsValidToCreate("RFQ-N1U1:EMR-RFA-071", Boolean.TRUE);
            ITUtilNameElement.assertValidate(nameElementCommandCreate,  NameCommand.CREATE, Boolean.TRUE);
            ITUtilNameElement.assertValidate(nameElementCommandUpdate,  NameCommand.UPDATE, Boolean.FALSE);
            ITUtilNameElement.assertValidate(nameElementCommandConfirm, NameCommand.DELETE, Boolean.FALSE);

            ITUtilNameElement.assertIsLegacy("RFQ",                    Boolean.FALSE);
            ITUtilNameElement.assertIsLegacy("RFQ-N1U1",               Boolean.FALSE);
            ITUtilNameElement.assertIsLegacy("RFQ-N1U1:EMR-RFA-071",   Boolean.FALSE);
            ITUtilNameElement.assertIsLegacy("RFQ-010PRL:EMR-RFA-071", Boolean.FALSE);
            ITUtilNameElement.assertRead("/legacy",             0, 0);
            ITUtilNameElement.assertRead("/legacy?name=RFQ%",   0, 0);
            ITUtilNameElement.assertRead("/legacy?name=%N1U1",  0, 0);
            ITUtilNameElement.assertRead("/legacy?name=%N1U1%", 0, 0);
            ITUtilNameElement.assertRead("/legacy?name=%EMR%",  0, 0);
            ITUtilNameElement.assertRead("/legacy?name=%EMR%",  0, 0);
            ITUtilNameElement.assertRead("/legacy?name=%FS%",   0, 0);

            // create
            createdNameElement = ITUtilNameElement.assertCreate(nameElementCommandCreate);
            nameElementCommandUpdate.setUuid(createdNameElement.getUuid());
            nameElementCommandConfirm.setUuid(createdNameElement.getUuid());

            Long totalCount2 = ITUtilNameElement.assertRead("").getTotalCount();
            assertNotNull(totalCount2);
            assertEquals(totalCount + 1, totalCount2);

            ITUtilNameElement.assertExists("RFQ-N1U1:EMR-RFA-071", Boolean.TRUE);
            ITUtilNameElement.assertIsValidToCreate("RFQ-N1U1:EMR-RFA-071", Boolean.FALSE);
            ITUtilNameElement.assertValidate(nameElementCommandCreate,  NameCommand.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertValidate(nameElementCommandUpdate,  NameCommand.UPDATE, Boolean.TRUE);
            ITUtilNameElement.assertValidate(nameElementCommandConfirm, NameCommand.DELETE, Boolean.TRUE);

            ITUtilNameElement.assertIsLegacy("RFQ",                    Boolean.FALSE);
            ITUtilNameElement.assertIsLegacy("RFQ-N1U1",               Boolean.FALSE);
            ITUtilNameElement.assertIsLegacy("RFQ-N1U1:EMR-RFA-071",   Boolean.FALSE);
            ITUtilNameElement.assertIsLegacy("RFQ-010PRL:EMR-RFA-071", Boolean.FALSE);
            ITUtilNameElement.assertRead("/legacy",             0, 0);
            ITUtilNameElement.assertRead("/legacy?name=RFQ%",   0, 0);
            ITUtilNameElement.assertRead("/legacy?name=%N1U1",  0, 0);
            ITUtilNameElement.assertRead("/legacy?name=%N1U1%", 0, 0);
            ITUtilNameElement.assertRead("/legacy?name=%EMR%",  0, 0);
            ITUtilNameElement.assertRead("/legacy?name=%EMR%",  0, 0);
            ITUtilNameElement.assertRead("/legacy?name=%FS%",   0, 0);

            // delete, approve
            structureElementCommandConfirm = new StructureElementCommandConfirm(
                    subsystemN1U1, Type.SUBSYSTEM, "comment");

            ITUtilStructureElement.assertDelete(structureElementCommandConfirm);

            ITUtilNameElement.assertExists("RFQ-N1U1:EMR-RFA-071", Boolean.TRUE);
            ITUtilNameElement.assertIsValidToCreate("RFQ-N1U1:EMR-RFA-071", Boolean.FALSE);
            ITUtilNameElement.assertValidate(nameElementCommandCreate,  NameCommand.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertValidate(nameElementCommandUpdate,  NameCommand.UPDATE, Boolean.TRUE);
            ITUtilNameElement.assertValidate(nameElementCommandConfirm, NameCommand.DELETE, Boolean.TRUE);

            ITUtilNameElement.assertIsLegacy("RFQ",                    Boolean.FALSE);
            ITUtilNameElement.assertIsLegacy("RFQ-N1U1",               Boolean.FALSE);
            ITUtilNameElement.assertIsLegacy("RFQ-N1U1:EMR-RFA-071",   Boolean.FALSE);
            ITUtilNameElement.assertIsLegacy("RFQ-010PRL:EMR-RFA-071", Boolean.FALSE);
            ITUtilNameElement.assertRead("/legacy",             0, 0);
            ITUtilNameElement.assertRead("/legacy?name=RFQ%",   0, 0);
            ITUtilNameElement.assertRead("/legacy?name=%N1U1",  0, 0);
            ITUtilNameElement.assertRead("/legacy?name=%N1U1%", 0, 0);
            ITUtilNameElement.assertRead("/legacy?name=%EMR%",  0, 0);
            ITUtilNameElement.assertRead("/legacy?name=%EMR%",  0, 0);
            ITUtilNameElement.assertRead("/legacy?name=%FS%",   0, 0);

            ITUtilStructureElement.assertApprove(structureElementCommandConfirm);

            totalCount = ITUtilNameElement.assertRead("").getTotalCount();
            assertNotNull(totalCount);
            assertEquals(totalCount2, totalCount);

            ITUtilNameElement.assertExists("RFQ-N1U1:EMR-RFA-071", Boolean.TRUE);
            ITUtilNameElement.assertIsValidToCreate("RFQ-N1U1:EMR-RFA-071", Boolean.FALSE);
            ITUtilNameElement.assertValidate(nameElementCommandCreate,  NameCommand.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertValidate(nameElementCommandUpdate,  NameCommand.UPDATE, Boolean.FALSE);
            ITUtilNameElement.assertValidate(nameElementCommandConfirm, NameCommand.DELETE, Boolean.TRUE);

            ITUtilNameElement.assertIsLegacy("RFQ",                    Boolean.FALSE);
            ITUtilNameElement.assertIsLegacy("RFQ-N1U1",               Boolean.TRUE);
            ITUtilNameElement.assertIsLegacy("RFQ-N1U1:EMR-RFA-071",   Boolean.TRUE);
            ITUtilNameElement.assertIsLegacy("RFQ-010PRL:EMR-RFA-071", Boolean.FALSE);
            ITUtilNameElement.assertRead("/legacy",             2, 2);
            ITUtilNameElement.assertRead("/legacy?name=RFQ%",   2, 2);
            ITUtilNameElement.assertRead("/legacy?name=%N1U1",  1, 1);
            ITUtilNameElement.assertRead("/legacy?name=%N1U1%", 2, 2);
            ITUtilNameElement.assertRead("/legacy?name=%EMR%",  1, 1);
            ITUtilNameElement.assertRead("/legacy?name=%EMR%",  1, 1);
            ITUtilNameElement.assertRead("/legacy?name=%FS%",   0, 0);

            ITUtilNameElement.assertDelete(nameElementCommandConfirm);

            ITUtilNameElement.assertExists("RFQ-N1U1:EMR-RFA-071", Boolean.FALSE);
            ITUtilNameElement.assertIsValidToCreate("RFQ-N1U1:EMR-RFA-071", Boolean.FALSE);
            ITUtilNameElement.assertValidate(nameElementCommandCreate,  NameCommand.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertValidate(nameElementCommandUpdate,  NameCommand.UPDATE, Boolean.FALSE);
            ITUtilNameElement.assertValidate(nameElementCommandConfirm, NameCommand.DELETE, Boolean.FALSE);

            ITUtilNameElement.assertIsLegacy("RFQ",                    Boolean.FALSE);
            ITUtilNameElement.assertIsLegacy("RFQ-N1U1",               Boolean.TRUE);
            ITUtilNameElement.assertIsLegacy("RFQ-N1U1:EMR-RFA-071",   Boolean.FALSE);
            ITUtilNameElement.assertIsLegacy("RFQ-010PRL:EMR-RFA-071", Boolean.FALSE);
            ITUtilNameElement.assertRead("/legacy",             1, 1);
            ITUtilNameElement.assertRead("/legacy?name=RFQ%",   1, 1);
            ITUtilNameElement.assertRead("/legacy?name=%N1U1",  1, 1);
            ITUtilNameElement.assertRead("/legacy?name=%N1U1%", 1, 1);
            ITUtilNameElement.assertRead("/legacy?name=%EMR%",  0, 0);
            ITUtilNameElement.assertRead("/legacy?name=%EMR%",  0, 0);
            ITUtilNameElement.assertRead("/legacy?name=%FS%",   0, 0);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void isValidToCreateName() {
        // purpose
        //     test if name is valid to create
        //
        // what
        //     read        is valid to create name

        ITUtilNameElement.assertIsValidToCreate("RFQ-010PRL:EMR-RFA-071", Boolean.TRUE);
    }

    @Test
    void checkCreate() {
        // purpose
        //     test conditions for create name
        //         not create itself
        //
        // what - combination of
        //     read        exists name
        //     read        is legacy name
        //     read        is valid to create
        //     read        validate create
        //
        // validate create
        //         system structure
        //     or  system structure + device structure + index
        //     +
        //     description
        //     comment

        try {
            NameElementCommandCreate nameElementCommandCreate = new NameElementCommandCreate();

            ITUtilNameElement.assertExists("RFQ-010PRL:EMR-RFA-051", Boolean.FALSE);
            ITUtilNameElement.assertIsLegacy("RFQ-010PRL:EMR-RFA-051", Boolean.FALSE);
            ITUtilNameElement.assertIsValidToCreate("RFQ-010PRL:EMR-RFA-051", Boolean.TRUE);

            // validate create

            ITUtilNameElement.assertValidate("[{asdf]", NameCommand.CREATE, HttpURLConnection.HTTP_BAD_REQUEST);
            ITUtilNameElement.assertCreate("[{asdf]", HttpURLConnection.HTTP_BAD_REQUEST);

            ITUtilNameElement.assertValidate(nameElementCommandCreate, NameCommand.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertCreate(nameElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            // ----------
            // system structure
            //     subsystem
            // description
            // comment

            nameElementCommandCreate.setDescription("description");
            ITUtilNameElement.assertValidate(nameElementCommandCreate, NameCommand.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertCreate(nameElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            nameElementCommandCreate.setComment("comment");
            ITUtilNameElement.assertValidate(nameElementCommandCreate, NameCommand.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertCreate(nameElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            // convention name exists
            // already exists since create subsystem was approved
            nameElementCommandCreate.setParentSystemStructure(subsystem010PRL);
            ITUtilNameElement.assertValidate(nameElementCommandCreate, NameCommand.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertCreate(nameElementCommandCreate, HttpURLConnection.HTTP_CONFLICT);

            // ----------
            // system structure + device structure + index
            //     subsystem, device type
            // description
            // comment

            nameElementCommandCreate.setIndex("051");
            ITUtilNameElement.assertValidate(nameElementCommandCreate, NameCommand.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertCreate(nameElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            nameElementCommandCreate.setParentDeviceStructure(deviceTypeRFA);
            ITUtilNameElement.assertValidate(nameElementCommandCreate, NameCommand.CREATE, Boolean.TRUE);

            // ----------

            nameElementCommandCreate.setParentSystemStructure(null);
            nameElementCommandCreate.setIndex(null);
            nameElementCommandCreate.setParentDeviceStructure(null);

            ITUtilNameElement.assertValidate(nameElementCommandCreate, NameCommand.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertCreate(nameElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            // ----------
            // system structure
            //     system
            // description
            // comment

            // convention name exists
            // already exists since create system was approved
            nameElementCommandCreate.setParentSystemStructure(systemRFQ);
            ITUtilNameElement.assertValidate(nameElementCommandCreate, NameCommand.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertCreate(nameElementCommandCreate, HttpURLConnection.HTTP_CONFLICT);

            // ----------
            // system structure + device structure + index
            //     system, device type
            // description
            // comment

            nameElementCommandCreate.setIndex("051");
            ITUtilNameElement.assertValidate(nameElementCommandCreate, NameCommand.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertCreate(nameElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            nameElementCommandCreate.setParentDeviceStructure(deviceTypeRFA);
            ITUtilNameElement.assertValidate(nameElementCommandCreate, NameCommand.CREATE, Boolean.TRUE);

            // ----------

            nameElementCommandCreate.setParentSystemStructure(null);
            nameElementCommandCreate.setIndex(null);
            nameElementCommandCreate.setParentDeviceStructure(null);

            ITUtilNameElement.assertValidate(nameElementCommandCreate, NameCommand.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertCreate(nameElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            // ----------
            // system structure
            //     system group
            // description
            // comment

            // convention name exists
            // already exists since create system group was approved
            nameElementCommandCreate.setParentSystemStructure(systemGroupAcc);
            ITUtilNameElement.assertValidate(nameElementCommandCreate, NameCommand.CREATE, Boolean.FALSE);

            // ----------
            // system structure + device structure + index
            //     system group, device type
            // description
            // comment

            nameElementCommandCreate.setIndex("051");
            ITUtilNameElement.assertValidate(nameElementCommandCreate, NameCommand.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertCreate(nameElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            nameElementCommandCreate.setParentDeviceStructure(deviceTypeTT);
            ITUtilNameElement.assertValidate(nameElementCommandCreate, NameCommand.CREATE, Boolean.TRUE);

            nameElementCommandCreate.setParentDeviceStructure(deviceTypeRFA);
            ITUtilNameElement.assertValidate(nameElementCommandCreate, NameCommand.CREATE, Boolean.TRUE);

            // ----------

            nameElementCommandCreate.setParentSystemStructure(null);
            nameElementCommandCreate.setIndex(null);
            nameElementCommandCreate.setParentDeviceStructure(null);

            ITUtilNameElement.assertValidate(nameElementCommandCreate, NameCommand.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertCreate(nameElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            // ----------
            // system structure + device structure
            //     system, device type
            // description
            // comment

            // convention name exists
            // already exists since create system was approved
            nameElementCommandCreate.setParentSystemStructure(systemRFQ);
            ITUtilNameElement.assertValidate(nameElementCommandCreate, NameCommand.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertCreate(nameElementCommandCreate, HttpURLConnection.HTTP_CONFLICT);

            nameElementCommandCreate.setParentDeviceStructure(deviceTypeRFA);
            ITUtilNameElement.assertValidate(nameElementCommandCreate, NameCommand.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertCreate(nameElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            nameElementCommandCreate.setIndex("051");
            ITUtilNameElement.assertValidate(nameElementCommandCreate, NameCommand.CREATE, Boolean.TRUE);

            // ----------
            // random uuid

            nameElementCommandCreate.setParentSystemStructure(UUID.randomUUID());
            ITUtilNameElement.assertValidate(nameElementCommandCreate, NameCommand.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertCreate(nameElementCommandCreate, HttpURLConnection.HTTP_NOT_FOUND);

            nameElementCommandCreate.setParentSystemStructure(systemRFQ);
            nameElementCommandCreate.setParentDeviceStructure(UUID.randomUUID());
            ITUtilNameElement.assertValidate(nameElementCommandCreate, NameCommand.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertCreate(nameElementCommandCreate, HttpURLConnection.HTTP_NOT_FOUND);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void create() {
        // purpose
        //     test create name
        //
        // what - combination of
        //     create      create names
        //     read        exists name
        //     read        is legacy name
        //     read        is valid to create
        //     read        validate create
        //
        // note
        //     with and without index

        try {
            ObjectMapper mapper = new ObjectMapper();

            NameElementCommandCreate nameElementCommandCreate  = null;
            NameElementCommandCreate nameElementCommandCreate2 = null;

            nameElementCommandCreate = new NameElementCommandCreate(
                    subsystem010PRL, deviceTypeRFA, "052",
                    "description", "comment");

            nameElementCommandCreate2 = new NameElementCommandCreate(
                    subsystem010PRL, deviceTypeRFA, "061",
                    "description", "comment");

            ITUtilNameElement.assertExists("RFQ-010PRL:EMR-RFA-052", Boolean.FALSE);
            ITUtilNameElement.assertIsLegacy("RFQ-010PRL:EMR-RFA-052", Boolean.FALSE);
            ITUtilNameElement.assertIsValidToCreate("RFQ-010PRL:EMR-RFA-052", Boolean.TRUE);

            ITUtilNameElement.assertValidate(
                    "[" + mapper.writeValueAsString(nameElementCommandCreate) + "," + mapper.writeValueAsString(nameElementCommandCreate2) + "]",
                    NameCommand.CREATE, HttpURLConnection.HTTP_OK);

            ITUtilNameElement.assertValidate(nameElementCommandCreate,  NameCommand.CREATE, Boolean.TRUE);
            ITUtilNameElement.assertValidate(nameElementCommandCreate2, NameCommand.CREATE, Boolean.TRUE);

            // create
            ITUtilNameElement.assertCreate(nameElementCommandCreate);

            ITUtilNameElement.assertExists("RFQ-010PRL:EMR-RFA-052", Boolean.TRUE);
            ITUtilNameElement.assertIsLegacy("RFQ-010PRL:EMR-RFA-052", Boolean.FALSE);
            ITUtilNameElement.assertIsValidToCreate("RFQ-010PRL:EMR-RFA-052", Boolean.FALSE);

            ITUtilNameElement.assertValidate(nameElementCommandCreate, NameCommand.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertCreate(nameElementCommandCreate, HttpURLConnection.HTTP_CONFLICT);
        } catch (IOException e) {
            fail();
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void checkUpdate() {
        // purpose
        //     test conditions for update name
        //         not update
        //
        // what - combination of
        //     create      create names
        //     read        validate create
        //     read        validate update
        //
        // note
        //     create in order to update
        //
        // validate update
        //     uuid
        //     +
        //         system structure
        //     or  system structure + device structure + index
        //     +
        //     description
        //     comment

        try {
            NameElementCommandCreate nameElementCommandCreate = null;
            NameElementCommandUpdate nameElementCommandUpdate = null;
            NameElement createdNameElement = null;

            nameElementCommandCreate = new NameElementCommandCreate(
                    subsystem010PRL, deviceTypeRFA, "053",
                    "description", "comment");

            // create
            createdNameElement = ITUtilNameElement.assertCreate(nameElementCommandCreate);
            nameElementCommandUpdate = NameElementUtil.convertElement2CommandUpdate(createdNameElement);

            // validate update

            nameElementCommandUpdate.setUuid(null);
            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.FALSE);
            ITUtilNameElement.assertUpdate(nameElementCommandUpdate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            nameElementCommandUpdate.setUuid(createdNameElement.getUuid());
            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.TRUE);

            nameElementCommandUpdate.setDescription(null);
            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.FALSE);
            ITUtilNameElement.assertUpdate(nameElementCommandUpdate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            nameElementCommandUpdate.setDescription("checkUpdate");
            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.TRUE);

            nameElementCommandUpdate.setComment(null);
            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.FALSE);
            ITUtilNameElement.assertUpdate(nameElementCommandUpdate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            nameElementCommandUpdate.setComment("checkUpdate");
            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.TRUE);

            // ----------

            nameElementCommandUpdate.setParentSystemStructure(null);
            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.FALSE);
            ITUtilNameElement.assertUpdate(nameElementCommandUpdate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            nameElementCommandUpdate.setParentSystemStructure(systemGroupAcc);
            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.TRUE);

            nameElementCommandUpdate.setParentSystemStructure(null);
            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.FALSE);
            ITUtilNameElement.assertUpdate(nameElementCommandUpdate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            nameElementCommandUpdate.setParentSystemStructure(systemRFQ);
            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.TRUE);

            nameElementCommandUpdate.setParentSystemStructure(null);
            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.FALSE);
            ITUtilNameElement.assertUpdate(nameElementCommandUpdate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            // subsystemN1U1 used in isLegacyName but not here
            // order of tests not guaranteed

            nameElementCommandUpdate.setParentSystemStructure(subsystem010PRL);
            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.TRUE);

            // ----------

            nameElementCommandUpdate.setParentDeviceStructure(null);
            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.FALSE);
            ITUtilNameElement.assertUpdate(nameElementCommandUpdate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            // convention name exists
            // already exists since create subsystem was approved
            nameElementCommandUpdate.setIndex(null);
            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.FALSE);
            ITUtilNameElement.assertUpdate(nameElementCommandUpdate, HttpURLConnection.HTTP_CONFLICT);

            nameElementCommandUpdate.setIndex("053");
            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.FALSE);
            ITUtilNameElement.assertUpdate(nameElementCommandUpdate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            nameElementCommandUpdate.setParentDeviceStructure(deviceGroupEMR);
            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.FALSE);
            ITUtilNameElement.assertUpdate(nameElementCommandUpdate, HttpURLConnection.HTTP_NOT_FOUND);

            nameElementCommandUpdate.setParentDeviceStructure(deviceTypeFS);
            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.TRUE);

            nameElementCommandUpdate.setParentDeviceStructure(deviceTypeTT);
            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.TRUE);

            nameElementCommandUpdate.setParentDeviceStructure(deviceTypeRFA);
            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.TRUE);

            nameElementCommandUpdate.setIndex(null);
            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.FALSE);
            ITUtilNameElement.assertUpdate(nameElementCommandUpdate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            nameElementCommandUpdate.setIndex("053");
            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.TRUE);

            // system structure, device structure not used for validation

            // ----------
            // random uuid

            nameElementCommandUpdate.setUuid(UUID.randomUUID());
            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.FALSE);
            ITUtilNameElement.assertUpdate(nameElementCommandUpdate, HttpURLConnection.HTTP_NOT_FOUND);
            nameElementCommandUpdate.setUuid(createdNameElement.getUuid());

            nameElementCommandUpdate.setParentSystemStructure(UUID.randomUUID());
            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.FALSE);
            ITUtilNameElement.assertUpdate(nameElementCommandUpdate, HttpURLConnection.HTTP_NOT_FOUND);
            nameElementCommandUpdate.setParentSystemStructure(subsystem010PRL);

            nameElementCommandUpdate.setParentDeviceStructure(UUID.randomUUID());
            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.FALSE);
            ITUtilNameElement.assertUpdate(nameElementCommandUpdate, HttpURLConnection.HTTP_NOT_FOUND);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void update() {
        // purpose
        //     test update name
        //
        // what - combination of
        //     create      create names
        //     read        validate create
        //     read        validate update
        //     update      update names
        //
        // note
        //     create in order to update

        try {
            NameElementCommandCreate nameElementCommandCreate = null;
            NameElementCommandUpdate nameElementCommandUpdate = null;
            NameElementCommand nameElementCommand = null;
            NameElement createdNameElement = null;

            nameElementCommand = new NameElementCommand(
                    null, subsystem010PRL, deviceTypeRFA, "054",
                    "description", "comment");
            nameElementCommandCreate = NameElementUtil.convertCommand2CommandCreate(nameElementCommand);
            nameElementCommandUpdate = NameElementUtil.convertCommand2CommandUpdate(nameElementCommand);

            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.FALSE);

            // create
            createdNameElement = ITUtilNameElement.assertCreate(nameElementCommandCreate);

            // status code
            nameElementCommandUpdate.setUuid(UUID.randomUUID());
            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.FALSE);
            ITUtilNameElement.assertUpdate(nameElementCommandUpdate, HttpURLConnection.HTTP_NOT_FOUND);
            nameElementCommandUpdate.setUuid(createdNameElement.getUuid());

            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.TRUE);

            nameElementCommandUpdate.setDescription("updated description");
            nameElementCommandUpdate.setComment("updated comment");

            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.TRUE);

            // update
            ITUtilNameElement.assertUpdate(nameElementCommandUpdate);
            nameElementCommandUpdate.setDescription("another description");
            nameElementCommandUpdate.setComment("another comment");

            ITUtilNameElement.assertValidate(nameElementCommandUpdate, NameCommand.UPDATE, Boolean.TRUE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void checkDelete() {
        // purpose
        //     test conditions for delete name
        //         not delete
        //
        // what - combination of
        //     create      create names
        //     read        validate create
        //     read        validate delete
        //
        // note
        //     create in order to delete
        //
        // validate delete
        //     uuid
        //     comment

        try {
            NameElementCommandCreate nameElementCommandCreate = null;
            NameElementCommandConfirm nameElementCommandConfirm = null;
            NameElement createdNameElement = null;

            nameElementCommandCreate = new NameElementCommandCreate(
                    subsystem010PRL, deviceTypeRFA, "055",
                    "description", "comment");

            // create
            createdNameElement = ITUtilNameElement.assertCreate(nameElementCommandCreate);
            nameElementCommandConfirm = NameElementUtil.convertElement2CommandConfirm(createdNameElement);

            // validate delete

            ITUtilNameElement.assertValidate(nameElementCommandConfirm, NameCommand.DELETE, Boolean.TRUE);

            nameElementCommandConfirm.setUuid(null);
            ITUtilNameElement.assertValidate(nameElementCommandConfirm, NameCommand.DELETE, Boolean.FALSE);
            ITUtilNameElement.assertDelete(nameElementCommandConfirm, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            nameElementCommandConfirm.setUuid(UUID.randomUUID());
            ITUtilNameElement.assertValidate(nameElementCommandConfirm, NameCommand.DELETE, Boolean.FALSE);
            ITUtilNameElement.assertDelete(nameElementCommandConfirm, HttpURLConnection.HTTP_NOT_FOUND);

            nameElementCommandConfirm.setUuid(createdNameElement.getUuid());
            ITUtilNameElement.assertValidate(nameElementCommandConfirm, NameCommand.DELETE, Boolean.TRUE);

            nameElementCommandConfirm.setComment(null);
            ITUtilNameElement.assertValidate(nameElementCommandConfirm, NameCommand.DELETE, Boolean.FALSE);
            ITUtilNameElement.assertDelete(nameElementCommandConfirm, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            nameElementCommandConfirm.setComment("checkDelete");
            ITUtilNameElement.assertValidate(nameElementCommandConfirm, NameCommand.DELETE, Boolean.TRUE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void delete() {
        // purpose
        //     test delete name
        //
        // what - combination of
        //     create      create names
        //     read        validate create
        //     read        validate update
        //     delete      delete names
        //
        // note
        //     create in order to delete

        try {
            NameElementCommandCreate nameElementCommandCreate = null;
            NameElementCommandConfirm nameElementCommandConfirm = null;
            NameElement createdNameElement = null;

            nameElementCommandCreate = new NameElementCommandCreate(
                    subsystem010PRL, deviceTypeRFA, "056",
                    "description", "comment");

            ITUtilNameElement.assertValidate(nameElementCommandCreate, NameCommand.DELETE, Boolean.FALSE);

            // create
            createdNameElement = ITUtilNameElement.assertCreate(nameElementCommandCreate);
            nameElementCommandConfirm = NameElementUtil.convertElement2CommandConfirm(createdNameElement);

            ITUtilNameElement.assertValidate(nameElementCommandConfirm, NameCommand.DELETE, Boolean.TRUE);

            nameElementCommandConfirm.setComment("deleted comment");

            ITUtilNameElement.assertValidate(nameElementCommandConfirm, NameCommand.DELETE, Boolean.TRUE);

            // delete
            ITUtilNameElement.assertDelete(nameElementCommandConfirm);

            ITUtilNameElement.assertValidate(nameElementCommandConfirm, NameCommand.DELETE, Boolean.FALSE);
            ITUtilNameElement.assertDelete(nameElementCommandConfirm, HttpURLConnection.HTTP_NOT_FOUND);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void readSearchHistoryDeleted() {
        // purpose
        //     test read names in various ways for create
        //         search
        //       ( latest )
        //         history
        //         deleted
        //
        // note
        //     create (and more) to read (with content)

        NameElementCommandCreate nameElementCommandCreate = null;
        NameElementCommandUpdate nameElementCommandUpdate = null;
        NameElementCommandConfirm nameElementCommandConfirm = null;
        NameElement responseNameElement = null;
        UUID uuid, uuid2 = null;

        NameElement nameElement1, nameElement7, nameElement8 = null;

        nameElementCommandCreate = new NameElementCommandCreate(
                subsystem010, deviceTypeFS, "001",
                "description", "comment");

        // create
        nameElement1 = ITUtilNameElement.assertCreate(nameElementCommandCreate);
        uuid = nameElement1.getUuid();

        nameElementCommandCreate = new NameElementCommandCreate(
                subsystem010, deviceTypeFS, "002",
                "description", "comment");
        responseNameElement = ITUtilNameElement.assertCreate(nameElementCommandCreate);
        uuid2 = responseNameElement.getUuid();

        nameElementCommandCreate = new NameElementCommandCreate(
                subsystem010, deviceTypeFS, "003",
                "description", "comment");
        ITUtilNameElement.assertCreate(nameElementCommandCreate);

        nameElementCommandCreate = new NameElementCommandCreate(
                subsystem010, deviceTypeFS, "004",
                "description", "comment");
        ITUtilNameElement.assertCreate(nameElementCommandCreate);

        nameElementCommandCreate = new NameElementCommandCreate(
                subsystem010, deviceTypeFS, "005",
                "description", "comment");
        ITUtilNameElement.assertCreate(nameElementCommandCreate);

        nameElementCommandCreate = new NameElementCommandCreate(
                subsystem010, deviceTypeFS, "006",
                "description", "comment");
        ITUtilNameElement.assertCreate(nameElementCommandCreate);

        nameElementCommandCreate = new NameElementCommandCreate(
                subsystem010, deviceTypeFS, "007",
                "description", "comment");
        nameElement7 = ITUtilNameElement.assertCreate(nameElementCommandCreate);

        nameElementCommandCreate = new NameElementCommandCreate(
                subsystem010, deviceTypeFS, "008",
                "description", "comment");
        nameElement8 = ITUtilNameElement.assertCreate(nameElementCommandCreate);

        // update element 1 twice
        nameElementCommandUpdate = NameElementUtil.convertElement2CommandUpdate(nameElement1);
        nameElementCommandUpdate.setDescription("updated description");
        nameElementCommandUpdate.setComment("updated comment");

        ITUtilNameElement.assertUpdate(nameElementCommandUpdate);

        nameElementCommandUpdate.setDescription("updated description again");
        nameElementCommandUpdate.setComment("updated comment again");

        ITUtilNameElement.assertUpdate(nameElementCommandUpdate);

        // delete element 7, 8
        nameElementCommandConfirm = NameElementUtil.convertElement2CommandConfirm(nameElement7);
        ITUtilNameElement.assertDelete(nameElementCommandConfirm);
        nameElementCommandConfirm = NameElementUtil.convertElement2CommandConfirm(nameElement8);
        ITUtilNameElement.assertDelete(nameElementCommandConfirm);

        try {
            // read & search
            //     deleted, uuid, name, nameEquivalence, systemStructure, deviceStructure, index, description
            //     combination
            ITUtilNameElement.assertRead("", 8, -1);

            ITUtilNameElement.assertRead("?deleted=false", 6, -1);
            ITUtilNameElement.assertRead("?deleted=true",  2, -1);

            ITUtilNameElement.assertRead("?uuid=" + deviceTypeFS.toString(), 0);
            ITUtilNameElement.assertRead("?uuid=" + uuid.toString(),         1);

            ITUtilNameElement.assertRead("?name=RFQ-010:EMR-FS-005",  1);
            ITUtilNameElement.assertRead("?name=RFQ-010%",            8, -1);
            ITUtilNameElement.assertRead("?name=RFQ-10%",             0);

            ITUtilNameElement.assertRead("?nameEquivalence=RFQ-10%",  8, -1);

            ITUtilNameElement.assertRead("?systemStructure=RFQ-010",  6, -1);
            ITUtilNameElement.assertRead("?systemStructure=RFQ-0",    0);
            ITUtilNameElement.assertRead("?systemStructure=RFQ-0__",  6, -1);
            ITUtilNameElement.assertRead("?systemStructure=RFQ-N1U1", 0, -1);

            ITUtilNameElement.assertRead("?deviceStructure=EMR-FS",   6, -1);
            ITUtilNameElement.assertRead("?deviceStructure=EMR-F",    0);
            ITUtilNameElement.assertRead("?deviceStructure=EMR-F_",   6, -1);
            ITUtilNameElement.assertRead("?deviceStructure=EMR-TT",   0);

            ITUtilNameElement.assertRead("?index=003", 1);

            ITUtilNameElement.assertRead("?description=description",               7, -1);
            ITUtilNameElement.assertRead("?description=%description%",             8, -1);
            ITUtilNameElement.assertRead("?description=updated description",       0, -1);
            ITUtilNameElement.assertRead("?description=updated description%",      1, -1);
            ITUtilNameElement.assertRead("?description=updated description again", 1);

            ITUtilNameElement.assertRead("?nameEquivalence=RFQ-10%&deviceStructure=EMR-FS", 6, -1);

            // name or uuid
            //     /{name}
            ITUtilNameElement.assertRead("/" + systemGroupAcc.toString(), 0);
            ITUtilNameElement.assertRead("/" + systemRFQ.toString(),      0);
            ITUtilNameElement.assertRead("/" + subsystem010.toString(),   0);
            ITUtilNameElement.assertRead("/" + disciplineEMR.toString(),  0);
            ITUtilNameElement.assertRead("/" + deviceGroupEMR.toString(), 0);
            ITUtilNameElement.assertRead("/" + deviceTypeFS.toString(),   0);
            ITUtilNameElement.assertRead("/" + uuid.toString(),           1);
            ITUtilNameElement.assertRead("/RFQ-010:EMR-FS-001",           1);
            ITUtilNameElement.assertRead("/RFQ-010:EMR-FS-0",             0);
            ITUtilNameElement.assertRead("/RFQ-010:EMR-FS-0__",           6, -1);
            ITUtilNameElement.assertRead("/RFQ-010:EMR-FS-0%",            6, -1);

            // system structure mnemonic path
            //     /systemStructure/{mnemonicPath}
            ITUtilNameElement.assertRead("/systemStructure/RFQ-010",      6, -1);
            ITUtilNameElement.assertRead("/systemStructure/RFQ-0",        0);
            ITUtilNameElement.assertRead("/systemStructure/RFQ-0__",      6, -1);
            ITUtilNameElement.assertRead("/systemStructure/RFQ-N1U1",     0, -1);

            // device structure mnemonic path
            //     /deviceStructure/{mnemonicPath}
            ITUtilNameElement.assertRead("/deviceStructure/EMR-FS",       6, -1);
            ITUtilNameElement.assertRead("/deviceStructure/EMR-F",        0);
            ITUtilNameElement.assertRead("/deviceStructure/EMR-F_",       6, -1);
            ITUtilNameElement.assertRead("/deviceStructure/EMR-TT",       0);

            // history
            //     /history/{uuid}
            ITUtilNameElement.assertRead("/history/" + systemGroupAcc.toString(), 0);
            ITUtilNameElement.assertRead("/history/" + systemRFQ.toString(),      0);
            ITUtilNameElement.assertRead("/history/" + subsystem010.toString(),   0);
            ITUtilNameElement.assertRead("/history/" + disciplineEMR.toString(),  0);
            ITUtilNameElement.assertRead("/history/" + deviceGroupEMR.toString(), 0);
            ITUtilNameElement.assertRead("/history/" + deviceTypeFS.toString(),   0);
            ITUtilNameElement.assertRead("/history/" + uuid.toString(),           3);
            ITUtilNameElement.assertRead("/history/" + uuid2.toString(),          1);
        } catch (Exception e) {
            fail();
        }
    }

}
