/*
 * Copyright (C) 2022 European Spallation Source ERIC.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.openepics.names.docker;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.fail;

import java.io.File;
import java.net.HttpURLConnection;
import java.util.UUID;

import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;
import org.openepics.names.rest.beans.Type;
import org.openepics.names.rest.beans.element.StructureElement;
import org.openepics.names.rest.beans.element.StructureElementCommand;
import org.openepics.names.rest.beans.element.StructureElementCommandConfirm;
import org.openepics.names.rest.beans.element.StructureElementCommandCreate;
import org.openepics.names.rest.beans.element.StructureElementCommandUpdate;
import org.openepics.names.util.StructureCommand;
import org.openepics.names.util.StructureElementUtil;
import org.testcontainers.containers.DockerComposeContainer;
import org.testcontainers.containers.wait.strategy.Wait;
import org.testcontainers.junit.jupiter.Container;
import org.testcontainers.junit.jupiter.Testcontainers;

/**
 * Integration tests for Naming and PostgreSQL that make use of existing dockerization
 * with docker-compose.yml / Dockerfile.
 *
 * <p>
 * Focus of this class is structures endpoint and device type.
 * </p>
 *
 * @author Lars Johansson
 */
@Testcontainers
class StructuresDeviceTypeIT {

    // note
    //     StructureElement - when, who, comment - depend on requested or processed
    //         requested, requested by, requested comment
    //         processed, processed by, processed comment
    //     if less than a second between requested and processed, then considered one entry with processed
    //     history
    //         mnemonic path does not make same sense for history
    //         (very) tricky to find mnemonic path for uuid at proper time (history)
    //         therefore empty mnemonic path for history for structure
    //         one history entry if less than one second between requested and processed, otherwise two history entries
    //     attributes for entry for operations - create, update, delete, approve, cancel, reject
    //         some set client side, others set server side
    //         client side
    //             type, uuid, parent uuid, name, mnemonic, description, comment
    //         may be set client side for test purposes
    //     order of methods in test not known
    //         therefore mnemonic values chosen to not interfere with mnemonic values in other methods
    //
    //     StructureElement extends StructureElementCommand
    //     -->
    //     for such operations that send Command objects to backend, it is also possible to send StructureElement
    //
    //     StructureElementCommand
    //	       create  -       type, parent, name, mnemonic, description, comment
    //	       update  - uuid, type, parent, name, mnemonic, description, comment
    //	       delete  - uuid, type,                                      comment
    //	       approve - uuid, type,                                      comment
    //	       cancel  - uuid, type,                                      comment
    //	       reject  - uuid, type,                                      comment
    //
    //     response status codes

    @Container
    public static final DockerComposeContainer<?> ENVIRONMENT =
        new DockerComposeContainer<>(new File("docker-compose-integrationtest.yml"))
            .waitingFor(ITUtil.NAMING, Wait.forLogMessage(".*Started NamingApplication.*", 1));

    private static UUID disciplineUuid = null;
    private static UUID deviceGroupUuid = null;

    @BeforeAll
    public static void initAll() {
        // init discipline, device group

        StructureElementCommandCreate structureElementCommandCreate = null;
        StructureElement approvedStructureElement = null;

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DISCIPLINE, null,
                "name", "Di",
                "description", "comment");
        approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        disciplineUuid = approvedStructureElement.getUuid();

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICEGROUP, disciplineUuid,
                "name", null,
                "description", "comment");
        approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        deviceGroupUuid = approvedStructureElement.getUuid();
    }

    @Test
    void equivalenceMnemonic() {
        // purpose
        //     test mnemonic equivalence
        //
        // what
        //     read        equivalence mnemonic

        ITUtilStructureElement.assertEquivalence("Dt", "DT");
    }

    @Test
    void checkCreate() {
        // purpose
        //     test conditions for create device type
        //         not create itself
        //
        // what - combination of
        //     read        exists in structure
        //     read        is valid to create
        //     read        validate create
        //
        // validate create
        //     type
        //     parent
        //     name
        //     mnemonic
        //     description
        //     comment

        try {
            StructureElementCommandCreate structureElementCommandCreate = new StructureElementCommandCreate();

            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "CC", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Cc", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DEVICETYPE, "Db-Cc", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DEVICETYPE, "Db-CC", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DEVICETYPE, "DB-Cc", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DEVICETYPE, "DB-CC", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DEVICETYPE, "Di-Cc", Boolean.TRUE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DEVICETYPE, "Di-CC", Boolean.TRUE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DEVICETYPE, "DI-Cc", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DEVICETYPE, "DI-CC", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DEVICETYPE, "Di-Di", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DEVICETYPE, "Di-DI", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DEVICETYPE, "DI-Di", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DEVICETYPE, "DI-DI", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DEVICETYPE, "Cc",    Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DEVICETYPE, "CC",    Boolean.FALSE);

            ITUtilStructureElement.assertValidate("[{asdf]", StructureCommand.CREATE, HttpURLConnection.HTTP_BAD_REQUEST);
            ITUtilStructureElement.assertCreate("[{asdf]", HttpURLConnection.HTTP_BAD_REQUEST);

            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setType(Type.DEVICETYPE);
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setParent(deviceGroupUuid);
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setName("name");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setMnemonic("Cc");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setDescription("description");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setComment("comment");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setType(null);
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void checkCreateMnemonic() {
        // purpose
        //     test conditions for create device type
        //         not create itself
        //
        // what - combination of
        //     read        validate create
        //
        // note
        //     mnemonic
        try {
            StructureElementCommandCreate structureElementCommandCreate = new StructureElementCommandCreate();

            structureElementCommandCreate.setType(Type.DEVICETYPE);
            structureElementCommandCreate.setParent(deviceGroupUuid);
            structureElementCommandCreate.setName("name");
            structureElementCommandCreate.setDescription("description");
            structureElementCommandCreate.setComment("comment");

            // mnemonic rules

            structureElementCommandCreate.setMnemonic(null);
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setMnemonic("");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setMnemonic("C");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("Cc");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("Ccc");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("Cccc");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("Ccccc");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("Cccccc");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("Ccccccc");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setMnemonic("Cccccccc");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setMnemonic("Ccccccccc");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            // mnemonic rules (2)

            structureElementCommandCreate.setMnemonic(" ");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setMnemonic("Dev ");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setMnemonic("Dev");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("000");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("Dev0");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic(":");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setMnemonic("Dev:");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setMnemonic("Dev:   ");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setMnemonic("1");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("12");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("123");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("1234");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("12345");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("123456");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("1234567");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setMnemonic("12345678");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setMnemonic("123456789");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void createApprove() {
        // purpose
        //     test create and approve system
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate approve
        //     patch       approve structures
        //
        // note
        //     create in order to approve

        try {
            StructureElementCommandCreate structureElementCommandCreate = null;
            StructureElementCommandConfirm structureElementCommandConfirm = null;
            StructureElementCommand structureElementCommand = null;
            StructureElement createdStructureElement = null;

            structureElementCommand = new StructureElementCommand(
                    null, Type.DEVICETYPE, deviceGroupUuid,
                    "name", "Ca",
                    "description", "comment");
            structureElementCommandCreate = StructureElementUtil.convertCommand2CommandCreate(structureElementCommand);
            structureElementCommandConfirm = StructureElementUtil.convertCommand2CommandConfirm(structureElementCommand);

            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-Ca", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DEVICETYPE, "Di-Ca", Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandCreate,  StructureCommand.CREATE,  Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.APPROVE, Boolean.FALSE);

            // create
            createdStructureElement = ITUtilStructureElement.assertCreate(structureElementCommandCreate);
            structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);

            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-Ca", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DEVICETYPE, "Di-Ca", Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandCreate,  StructureCommand.CREATE,  Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.APPROVE, Boolean.TRUE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, HttpURLConnection.HTTP_CONFLICT);

            // approve
            ITUtilStructureElement.assertApprove(structureElementCommandConfirm);

            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-Ca", Boolean.TRUE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DEVICETYPE, "Di-Ca", Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElementCommandCreate,  StructureCommand.CREATE,  Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.APPROVE, Boolean.FALSE);

            structureElementCommandConfirm.setUuid(null);
            ITUtilStructureElement.assertApprove(structureElementCommandConfirm, ITUtil.HTTP_UNPROCESSABLE_ENTITY);
            structureElementCommandConfirm.setUuid(UUID.randomUUID());
            ITUtilStructureElement.assertApprove(structureElementCommandConfirm, HttpURLConnection.HTTP_NOT_FOUND);
            structureElementCommandConfirm.setUuid(createdStructureElement.getUuid());
            ITUtilStructureElement.assertApprove(structureElementCommandConfirm, HttpURLConnection.HTTP_CONFLICT);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void createCancel() {
        // purpose
        //     test create and cancel system
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate cancel
        //     patch       cancel structures
        //
        // note
        //     create in order to cancel

        try {
            StructureElementCommandCreate structureElementCommandCreate = null;
            StructureElementCommandConfirm structureElementCommandConfirm = null;
            StructureElementCommand structureElementCommand = null;
            StructureElement createdStructureElement = null;

            structureElementCommand = new StructureElementCommand(
                    null, Type.DEVICETYPE, deviceGroupUuid,
                    "name", "Cc",
                    "description", "comment");
            structureElementCommandCreate = StructureElementUtil.convertCommand2CommandCreate(structureElementCommand);
            structureElementCommandConfirm = StructureElementUtil.convertCommand2CommandConfirm(structureElementCommand);

            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-Cc", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DEVICETYPE, "Di-Cc", Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandCreate,  StructureCommand.CREATE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.CANCEL, Boolean.FALSE);

            // create
            createdStructureElement = ITUtilStructureElement.assertCreate(structureElementCommandCreate);
            structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
            UUID uuid = structureElementCommandConfirm.getUuid();

            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-Cc", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DEVICETYPE, "Di-Cc", Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandCreate,  StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.CANCEL, Boolean.TRUE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, HttpURLConnection.HTTP_CONFLICT);

            structureElementCommandConfirm.setUuid(null);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.CANCEL, Boolean.FALSE);
            structureElementCommandConfirm.setUuid(uuid);

            // cancel
            ITUtilStructureElement.assertCancel(structureElementCommandConfirm);

            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-Cc", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DEVICETYPE, "Di-Cc", Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandCreate,  StructureCommand.CREATE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.CANCEL, Boolean.FALSE);

            structureElementCommandConfirm.setUuid(null);
            ITUtilStructureElement.assertCancel(structureElementCommandConfirm, ITUtil.HTTP_UNPROCESSABLE_ENTITY);
            structureElementCommandConfirm.setUuid(UUID.randomUUID());
            ITUtilStructureElement.assertCancel(structureElementCommandConfirm, HttpURLConnection.HTTP_NOT_FOUND);
            structureElementCommandConfirm.setUuid(uuid);
            ITUtilStructureElement.assertCancel(structureElementCommandConfirm, HttpURLConnection.HTTP_CONFLICT);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void createReject() {
        // purpose
        //     test create and reject system
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate reject
        //     patch       reject structures
        //
        // note
        //     create in order to reject

        try {
            StructureElementCommandCreate structureElementCommandCreate = null;
            StructureElementCommandConfirm structureElementCommandConfirm = null;
            StructureElementCommand structureElementCommand = null;
            StructureElement createdStructureElement = null;

            structureElementCommand = new StructureElementCommand(
                    null, Type.DEVICETYPE, deviceGroupUuid,
                    "name", "Cr",
                    "description", "comment");
            structureElementCommandCreate = StructureElementUtil.convertCommand2CommandCreate(structureElementCommand);
            structureElementCommandConfirm = StructureElementUtil.convertCommand2CommandConfirm(structureElementCommand);

            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-Cr", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DEVICETYPE, "Di-Cr", Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandCreate,  StructureCommand.CREATE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.REJECT, Boolean.FALSE);

            // create
            createdStructureElement = ITUtilStructureElement.assertCreate(structureElementCommandCreate);
            structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
            UUID uuid = structureElementCommandConfirm.getUuid();

            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-Cr", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DEVICETYPE, "Di-Cr", Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandCreate,  StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.REJECT, Boolean.TRUE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, HttpURLConnection.HTTP_CONFLICT);

            structureElementCommandConfirm.setUuid(null);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.REJECT, Boolean.FALSE);
            structureElementCommandConfirm.setUuid(uuid);

            // reject
            ITUtilStructureElement.assertReject(structureElementCommandConfirm);

            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-Cr", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DEVICETYPE, "Di-Cr", Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandCreate,  StructureCommand.CREATE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.REJECT, Boolean.FALSE);

            structureElementCommandConfirm.setUuid(null);
            ITUtilStructureElement.assertReject(structureElementCommandConfirm, ITUtil.HTTP_UNPROCESSABLE_ENTITY);
            structureElementCommandConfirm.setUuid(UUID.randomUUID());
            ITUtilStructureElement.assertReject(structureElementCommandConfirm, HttpURLConnection.HTTP_NOT_FOUND);
            structureElementCommandConfirm.setUuid(uuid);
            ITUtilStructureElement.assertReject(structureElementCommandConfirm, HttpURLConnection.HTTP_CONFLICT);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void checkUpdate() {
        // purpose
        //     test conditions for update device type
        //         not update
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate update
        //     read        validate approve
        //     patch       approve structures
        //
        // note
        //     create, approve in order to update
        //
        // validate update
        //     uuid
        //     type
        //     parent
        //     name
        //     mnemonic
        //     description
        //     comment

        try {
            UUID uuid = null;

            StructureElementCommandCreate structureElementCommandCreate = null;
            StructureElementCommandUpdate structureElementCommandUpdate = null;
            StructureElement approvedStructureElement = null;

            // create, approve
            structureElementCommandCreate = new StructureElementCommandCreate(
                    Type.DEVICETYPE, deviceGroupUuid,
                    "name", "Cu",
                    "description", "comment");
            approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
            structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(approvedStructureElement);
            uuid = approvedStructureElement.getUuid();

            // validate update

            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.TRUE);

            structureElementCommandUpdate.setUuid(null);
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.FALSE);
            ITUtilStructureElement.assertUpdate(structureElementCommandUpdate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandUpdate.setUuid(UUID.randomUUID());
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.FALSE);
            ITUtilStructureElement.assertUpdate(structureElementCommandUpdate, HttpURLConnection.HTTP_NOT_FOUND);

            structureElementCommandUpdate.setUuid(uuid);
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.TRUE);

            structureElementCommandUpdate.setType(null);
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.FALSE);
            ITUtilStructureElement.assertUpdate(structureElementCommandUpdate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandUpdate.setType(Type.SUBSYSTEM);
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.FALSE);
            ITUtilStructureElement.assertUpdate(structureElementCommandUpdate, HttpURLConnection.HTTP_NOT_FOUND);

            structureElementCommandUpdate.setType(Type.DEVICETYPE);
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.TRUE);

            structureElementCommandUpdate.setParent(null);
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.FALSE);
            ITUtilStructureElement.assertUpdate(structureElementCommandUpdate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandUpdate.setParent(deviceGroupUuid);
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.TRUE);

            structureElementCommandUpdate.setName(null);
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.FALSE);
            ITUtilStructureElement.assertUpdate(structureElementCommandUpdate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandUpdate.setName("name");
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.TRUE);

            structureElementCommandUpdate.setMnemonic(null);
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.FALSE);
            ITUtilStructureElement.assertUpdate(structureElementCommandUpdate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandUpdate.setMnemonic("Cu");
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.TRUE);

            structureElementCommandUpdate.setDescription(null);
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.FALSE);
            ITUtilStructureElement.assertUpdate(structureElementCommandUpdate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandUpdate.setDescription("description");
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.TRUE);

            structureElementCommandUpdate.setComment(null);
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.FALSE);
            ITUtilStructureElement.assertUpdate(structureElementCommandUpdate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandUpdate.setComment("comment");
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.TRUE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void updateApprove() {
        // purpose
        //     test update and approve device type
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate update
        //     read        validate approve
        //     update      update structures
        //     patch       approve structures
        //
        // note
        //     create, approve in order to update, approve

        try {
            StructureElementCommandCreate structureElementCommandCreate = null;
            StructureElementCommandUpdate structureElementCommandUpdate = null;
            StructureElementCommandConfirm structureElementCommandConfirm = null;
            StructureElement approvedStructureElement = null;

            // create, approve
            structureElementCommandCreate = new StructureElementCommandCreate(
                    Type.DEVICETYPE, deviceGroupUuid,
                    "name", "Ua",
                    "description", "comment");
            approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
            structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(approvedStructureElement);
            structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(approvedStructureElement);

            structureElementCommandUpdate.setComment("comment update approve check");

            ITUtilStructureElement.assertValidate(structureElementCommandUpdate,  StructureCommand.UPDATE,  Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.APPROVE, Boolean.FALSE);

            // update
            ITUtilStructureElement.assertUpdate(structureElementCommandUpdate);

            ITUtilStructureElement.assertValidate(structureElementCommandUpdate,  StructureCommand.UPDATE,  Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.APPROVE, Boolean.TRUE);
            ITUtilStructureElement.assertUpdate(structureElementCommandUpdate, HttpURLConnection.HTTP_CONFLICT);

            // approve
            ITUtilStructureElement.assertApprove(structureElementCommandConfirm);

            ITUtilStructureElement.assertValidate(structureElementCommandUpdate,  StructureCommand.UPDATE,  Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.APPROVE, Boolean.FALSE);

            structureElementCommandConfirm.setUuid(null);
            ITUtilStructureElement.assertApprove(structureElementCommandConfirm, ITUtil.HTTP_UNPROCESSABLE_ENTITY);
            structureElementCommandConfirm.setUuid(UUID.randomUUID());
            ITUtilStructureElement.assertApprove(structureElementCommandConfirm, HttpURLConnection.HTTP_NOT_FOUND);
            structureElementCommandConfirm.setUuid(approvedStructureElement.getUuid());
            ITUtilStructureElement.assertApprove(structureElementCommandConfirm, HttpURLConnection.HTTP_CONFLICT);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void updateCancel() {
        // purpose
        //     test update and cancel device type
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate update
        //     read        validate approve
        //     read        validate cancel
        //     update      update structures
        //     patch       approve structures
        //     patch       cancel structures
        //
        // note
        //     create, approve in order to update, cancel

        try {
            StructureElementCommandCreate structureElementCommandCreate = null;
            StructureElementCommandUpdate structureElementCommandUpdate = null;
            StructureElementCommandConfirm structureElementCommandConfirm = null;
            StructureElement approvedStructureElement = null;

            // create, approve
            structureElementCommandCreate = new StructureElementCommandCreate(
                    Type.DEVICETYPE, deviceGroupUuid,
                    "name", "Uc",
                    "description", "comment");
            approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
            structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(approvedStructureElement);
            structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(approvedStructureElement);

            structureElementCommandUpdate.setComment("comment update cancel check");

            ITUtilStructureElement.assertValidate(structureElementCommandUpdate,  StructureCommand.UPDATE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.CANCEL, Boolean.FALSE);

            // update
            ITUtilStructureElement.assertUpdate(structureElementCommandUpdate);

            ITUtilStructureElement.assertValidate(structureElementCommandUpdate,  StructureCommand.UPDATE, Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.CANCEL, Boolean.TRUE);
            ITUtilStructureElement.assertUpdate(structureElementCommandUpdate, HttpURLConnection.HTTP_CONFLICT);

            // cancel
            ITUtilStructureElement.assertCancel(structureElementCommandConfirm);

            ITUtilStructureElement.assertValidate(structureElementCommandUpdate,  StructureCommand.UPDATE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.CANCEL, Boolean.FALSE);

            structureElementCommandConfirm.setUuid(null);
            ITUtilStructureElement.assertCancel(structureElementCommandConfirm, ITUtil.HTTP_UNPROCESSABLE_ENTITY);
            structureElementCommandConfirm.setUuid(UUID.randomUUID());
            ITUtilStructureElement.assertCancel(structureElementCommandConfirm, HttpURLConnection.HTTP_NOT_FOUND);
            structureElementCommandConfirm.setUuid(approvedStructureElement.getUuid());
            ITUtilStructureElement.assertCancel(structureElementCommandConfirm, HttpURLConnection.HTTP_CONFLICT);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void updateReject() {
        // purpose
        //     test update and reject device type
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate update
        //     read        validate approve
        //     read        validate reject
        //     update      update structures
        //     patch       approve structures
        //     patch       reject structures
        //
        // note
        //     create, approve in order to update, reject

        try {
            StructureElementCommandCreate structureElementCommandCreate = null;
            StructureElementCommandUpdate structureElementCommandUpdate = null;
            StructureElementCommandConfirm structureElementCommandConfirm = null;
            StructureElement approvedStructureElement = null;

            // create, approve
            structureElementCommandCreate = new StructureElementCommandCreate(
                    Type.DEVICETYPE, deviceGroupUuid,
                    "name", "Ur",
                    "description", "comment");
            approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
            structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(approvedStructureElement);
            structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(approvedStructureElement);

            structureElementCommandUpdate.setComment("comment update reject check");

            ITUtilStructureElement.assertValidate(structureElementCommandUpdate,  StructureCommand.UPDATE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.REJECT, Boolean.FALSE);

            // update
            ITUtilStructureElement.assertUpdate(structureElementCommandUpdate);

            ITUtilStructureElement.assertValidate(structureElementCommandUpdate,  StructureCommand.UPDATE, Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.REJECT, Boolean.TRUE);
            ITUtilStructureElement.assertUpdate(structureElementCommandUpdate, HttpURLConnection.HTTP_CONFLICT);

            // reject
            ITUtilStructureElement.assertReject(structureElementCommandConfirm);

            ITUtilStructureElement.assertValidate(structureElementCommandUpdate,  StructureCommand.UPDATE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.REJECT, Boolean.FALSE);

            structureElementCommandConfirm.setUuid(null);
            ITUtilStructureElement.assertReject(structureElementCommandConfirm, ITUtil.HTTP_UNPROCESSABLE_ENTITY);
            structureElementCommandConfirm.setUuid(UUID.randomUUID());
            ITUtilStructureElement.assertReject(structureElementCommandConfirm, HttpURLConnection.HTTP_NOT_FOUND);
            structureElementCommandConfirm.setUuid(approvedStructureElement.getUuid());
            ITUtilStructureElement.assertReject(structureElementCommandConfirm, HttpURLConnection.HTTP_CONFLICT);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void checkDelete() {
        // purpose
        //     test conditions for delete device type
        //         not delete
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate delete
        //     read        validate approve
        //     patch       approve structures
        //
        // note
        //     create, approve in order to delete
        //
        // validate delete
        //     uuid
        //     type
        //     comment

        try {
            UUID uuid = null;

            StructureElementCommandCreate structureElementCommandCreate = null;
            StructureElementCommandConfirm structureElementCommandConfirm = null;
            StructureElement approvedStructureElement = null;

            // create, approve
            structureElementCommandCreate = new StructureElementCommandCreate(
                    Type.DEVICETYPE, deviceGroupUuid,
                    "name", "Cd",
                    "description", "comment");
            approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
            structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(approvedStructureElement);
            uuid = approvedStructureElement.getUuid();

            // validate delete

            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE, Boolean.TRUE);

            structureElementCommandConfirm.setUuid(null);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE, Boolean.FALSE);
            ITUtilStructureElement.assertDelete(structureElementCommandConfirm, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandConfirm.setUuid(UUID.randomUUID());
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE, Boolean.FALSE);
            ITUtilStructureElement.assertDelete(structureElementCommandConfirm, HttpURLConnection.HTTP_NOT_FOUND);

            structureElementCommandConfirm.setUuid(uuid);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE, Boolean.TRUE);

            structureElementCommandConfirm.setType(null);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE, Boolean.FALSE);
            ITUtilStructureElement.assertDelete(structureElementCommandConfirm, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandConfirm.setType(Type.SUBSYSTEM);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.UPDATE, Boolean.FALSE);
            ITUtilStructureElement.assertDelete(structureElementCommandConfirm, HttpURLConnection.HTTP_NOT_FOUND);

            structureElementCommandConfirm.setType(Type.DEVICETYPE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE, Boolean.TRUE);

            structureElementCommandConfirm.setComment(null);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE, Boolean.FALSE);
            ITUtilStructureElement.assertDelete(structureElementCommandConfirm, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandConfirm.setComment("comment");
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE, Boolean.TRUE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void deleteApprove() {
        // purpose
        //     test delete and approve device type
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate delete
        //     read        validate approve
        //     delete      delete structures
        //     patch       approve structures
        //
        // note
        //     create, approve in order to delete, approve

        try {
            StructureElementCommandCreate structureElementCommandCreate = null;
            StructureElementCommandConfirm structureElementCommandConfirm = null;
            StructureElement approvedStructureElement = null;

            // create, approve
            structureElementCommandCreate = new StructureElementCommandCreate(
                    Type.DEVICETYPE, deviceGroupUuid,
                    "name", "Da",
                    "description", "comment");
            approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
            structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(approvedStructureElement);

            structureElementCommandConfirm.setComment("comment delete approve check");

            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE,  Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.APPROVE, Boolean.FALSE);

            // delete
            ITUtilStructureElement.assertDelete(structureElementCommandConfirm);

            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE,  Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.APPROVE, Boolean.TRUE);
            ITUtilStructureElement.assertDelete(structureElementCommandConfirm, HttpURLConnection.HTTP_CONFLICT);

            // approve
            ITUtilStructureElement.assertApprove(structureElementCommandConfirm);

            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE,  Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.APPROVE, Boolean.FALSE);

            structureElementCommandConfirm.setUuid(null);
            ITUtilStructureElement.assertApprove(structureElementCommandConfirm, ITUtil.HTTP_UNPROCESSABLE_ENTITY);
            structureElementCommandConfirm.setUuid(UUID.randomUUID());
            ITUtilStructureElement.assertApprove(structureElementCommandConfirm, HttpURLConnection.HTTP_NOT_FOUND);
            structureElementCommandConfirm.setUuid(approvedStructureElement.getUuid());
            ITUtilStructureElement.assertApprove(structureElementCommandConfirm, HttpURLConnection.HTTP_CONFLICT);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void deleteCancel() {
        // purpose
        //     test delete and cancel device type
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate delete
        //     read        validate approve
        //     read        validate cancel
        //     update      delete structures
        //     patch       approve structures
        //     patch       cancel structures
        //
        // note
        //     create, approve in order to delete, cancel

        try {
            StructureElementCommandCreate structureElementCommandCreate = null;
            StructureElementCommandConfirm structureElementCommandConfirm = null;
            StructureElement approvedStructureElement = null;

            // create, approve
            structureElementCommandCreate = new StructureElementCommandCreate(
                    Type.DEVICETYPE, deviceGroupUuid,
                    "name", "Dc",
                    "description", "comment");
            approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
            structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(approvedStructureElement);

            structureElementCommandConfirm.setComment("comment update delete check");

            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.CANCEL, Boolean.FALSE);

            // delete
            ITUtilStructureElement.assertDelete(structureElementCommandConfirm);

            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE, Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.CANCEL, Boolean.TRUE);
            ITUtilStructureElement.assertDelete(structureElementCommandConfirm, HttpURLConnection.HTTP_CONFLICT);

            // cancel
            ITUtilStructureElement.assertCancel(structureElementCommandConfirm);

            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.CANCEL, Boolean.FALSE);

            structureElementCommandConfirm.setUuid(null);
            ITUtilStructureElement.assertCancel(structureElementCommandConfirm, ITUtil.HTTP_UNPROCESSABLE_ENTITY);
            structureElementCommandConfirm.setUuid(UUID.randomUUID());
            ITUtilStructureElement.assertCancel(structureElementCommandConfirm, HttpURLConnection.HTTP_NOT_FOUND);
            structureElementCommandConfirm.setUuid(approvedStructureElement.getUuid());
            ITUtilStructureElement.assertCancel(structureElementCommandConfirm, HttpURLConnection.HTTP_CONFLICT);

            structureElementCommandConfirm.setUuid(null);
            ITUtilStructureElement.assertReject(structureElementCommandConfirm, ITUtil.HTTP_UNPROCESSABLE_ENTITY);
            structureElementCommandConfirm.setUuid(UUID.randomUUID());
            ITUtilStructureElement.assertReject(structureElementCommandConfirm, HttpURLConnection.HTTP_NOT_FOUND);
            structureElementCommandConfirm.setUuid(approvedStructureElement.getUuid());
            ITUtilStructureElement.assertReject(structureElementCommandConfirm, HttpURLConnection.HTTP_CONFLICT);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void deleteReject() {
        // purpose
        //     test delete and reject device type
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate delete
        //     read        validate approve
        //     read        validate reject
        //     update      delete structures
        //     patch       approve structures
        //     patch       reject structures
        //
        // note
        //     create, approve in order to delete, reject

        try {
            StructureElementCommandCreate structureElementCommandCreate = null;
            StructureElementCommandConfirm structureElementCommandConfirm = null;
            StructureElement approvedStructureElement = null;

            // create, approve
            structureElementCommandCreate = new StructureElementCommandCreate(
                    Type.DEVICETYPE, deviceGroupUuid,
                    "name", "Dr",
                    "description", "comment");
            approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
            structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(approvedStructureElement);

            structureElementCommandConfirm.setComment("comment delete reject check");

            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.REJECT, Boolean.FALSE);

            // delete
            ITUtilStructureElement.assertDelete(structureElementCommandConfirm);

            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE, Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.REJECT, Boolean.TRUE);
            ITUtilStructureElement.assertDelete(structureElementCommandConfirm, HttpURLConnection.HTTP_CONFLICT);

            // reject
            ITUtilStructureElement.assertReject(structureElementCommandConfirm);

            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.REJECT, Boolean.FALSE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void readSearchHistoryApprove() {
        // purpose
        //     test read device type in various ways for create, approve
        //         search
        //         latest
        //         history
        //
        // note
        //     create (and more) to read (with content)

        try {
            StructureElementCommandCreate structureElementCommandCreate = null;
            StructureElementCommandConfirm structureElementCommandConfirm = null;
            StructureElement createdStructureElement  = null;
            StructureElement approvedStructureElement = null;
            int length = 0;

            structureElementCommandCreate = new StructureElementCommandCreate(
                    Type.DEVICETYPE, deviceGroupUuid,
                    "name", "Rsha",
                    "description", "comment");

            // create
            createdStructureElement = ITUtilStructureElement.assertCreate(structureElementCommandCreate);
            structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);

            // read (1)
            length = ITUtilStructureElement.assertRead("/DEVICETYPE", 1, -1).getListSize();

            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=PENDING",   1, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=APPROVED",  0, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=CANCELLED", 0, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=REJECTED",  0, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=PENDING&mnemonic=Rsha",   1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=APPROVED&mnemonic=Rsha",  0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=CANCELLED&mnemonic=Rsha", 0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=REJECTED&mnemonic=Rsha",  0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?uuid=" + createdStructureElement.getUuid().toString(), 1);

            ITUtilStructureElement.assertRead("/mnemonic/Di",          1);
            ITUtilStructureElement.assertRead("/mnemonic/Rsha",        0);
            ITUtilStructureElement.assertRead("/mnemonic/Di-Rsha",     0);

            ITUtilStructureElement.assertRead("/mnemonicPath/Di",      1, -1);
            ITUtilStructureElement.assertRead("/mnemonicPath/Rsha",    0);
            ITUtilStructureElement.assertRead("/mnemonicPath/Di-Rsha", 0);

            ITUtilStructureElement.assertRead("/history/" + createdStructureElement.getUuid().toString(), 1);

            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Rsha",    Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-Rsha", Boolean.FALSE);

            // approve
            approvedStructureElement = ITUtilStructureElement.assertApprove(structureElementCommandConfirm);

            // read (2)
            assertEquals(length, ITUtilStructureElement.assertRead("/DEVICETYPE", 1, -1).getListSize());

            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=PENDING",   0, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=APPROVED",  1, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=CANCELLED", 0, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=REJECTED",  0, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=PENDING&mnemonic=Rsha",   0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=APPROVED&mnemonic=Rsha",  1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=CANCELLED&mnemonic=Rsha", 0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=REJECTED&mnemonic=Rsha",  0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?uuid=" + approvedStructureElement.getUuid().toString(), 1);

            ITUtilStructureElement.assertRead("/mnemonic/Di",          1);
            ITUtilStructureElement.assertRead("/mnemonic/Rsha",        1);
            ITUtilStructureElement.assertRead("/mnemonic/Di-Rsha",     0);

            ITUtilStructureElement.assertRead("/mnemonicPath/Di",      1, -1);
            ITUtilStructureElement.assertRead("/mnemonicPath/Rsha",    0);
            ITUtilStructureElement.assertRead("/mnemonicPath/Di-Rsha", 1);

            ITUtilStructureElement.assertRead("/history/" + approvedStructureElement.getUuid().toString(), 2);

            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Rsha",    Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-Rsha", Boolean.TRUE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void readSearchHistoryCancel() {
        // purpose
        //     test read device type in various ways for create, cancel
        //         search
        //         latest
        //         history
        //
        // note
        //     create (and more) to read (with content)

        try {
            StructureElementCommandCreate structureElementCommandCreate = null;
            StructureElementCommandConfirm structureElementCommandConfirm = null;
            StructureElement createdStructureElement   = null;
            StructureElement cancelledStructureElement = null;
            int length = 0;

            structureElementCommandCreate = new StructureElementCommandCreate(
                    Type.DEVICETYPE, deviceGroupUuid,
                    "name", "Rshc",
                    "description", "comment");

            // create
            createdStructureElement = ITUtilStructureElement.assertCreate(structureElementCommandCreate);
            structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);

            // read (1)
            length = ITUtilStructureElement.assertRead("/DEVICETYPE", 1, -1).getListSize();

            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=PENDING",   1, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=APPROVED",  0, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=CANCELLED", 0, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=REJECTED",  0, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=PENDING&mnemonic=Rshc",   1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=APPROVED&mnemonic=Rshc",  0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=CANCELLED&mnemonic=Rshc", 0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=REJECTED&mnemonic=Rshc",  0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?uuid=" + createdStructureElement.getUuid().toString(), 1);

            ITUtilStructureElement.assertRead("/mnemonic/Di",          1);
            ITUtilStructureElement.assertRead("/mnemonic/Rshc",        0);
            ITUtilStructureElement.assertRead("/mnemonic/Di-Rshc",     0);

            ITUtilStructureElement.assertRead("/mnemonicPath/Di",      1, -1);
            ITUtilStructureElement.assertRead("/mnemonicPath/Rshc",    0);
            ITUtilStructureElement.assertRead("/mnemonicPath/Di-Rshc", 0);

            ITUtilStructureElement.assertRead("/history/" + createdStructureElement.getUuid().toString(), 1);

            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Rshc",    Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-Rshc", Boolean.FALSE);

            // cancel
            cancelledStructureElement = ITUtilStructureElement.assertCancel(structureElementCommandConfirm);

            // read (2)
            assertEquals(length, ITUtilStructureElement.assertRead("/DEVICETYPE", 1, -1).getListSize());

            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=PENDING",   0, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=APPROVED",  0, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=CANCELLED", 1, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=REJECTED",  0, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=PENDING&mnemonic=Rshc",   0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=APPROVED&mnemonic=Rshc",  0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=CANCELLED&mnemonic=Rshc", 1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=REJECTED&mnemonic=Rshc",  0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?uuid=" + cancelledStructureElement.getUuid().toString(), 1);

            ITUtilStructureElement.assertRead("/mnemonic/Di",          1);
            ITUtilStructureElement.assertRead("/mnemonic/Rshc",        0);
            ITUtilStructureElement.assertRead("/mnemonic/Di-Rshc",     0);

            ITUtilStructureElement.assertRead("/mnemonicPath/Di",      1, -1);
            ITUtilStructureElement.assertRead("/mnemonicPath/Rshc",    0);
            ITUtilStructureElement.assertRead("/mnemonicPath/Di-Rshc", 0);

            ITUtilStructureElement.assertRead("/history/" + cancelledStructureElement.getUuid().toString(), 2);

            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Rshc",    Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-Rshc", Boolean.FALSE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void readSearchHistoryReject() {
        // purpose
        //     test read device type in various ways for create, reject
        //         search
        //         latest
        //         history
        //
        // note
        //     create (and more) to read (with content)

        try {
            StructureElementCommandCreate structureElementCommandCreate = null;
            StructureElementCommandConfirm structureElementCommandConfirm = null;
            StructureElement createdStructureElement  = null;
            StructureElement rejectedStructureElement = null;
            int length = 0;

            structureElementCommandCreate = new StructureElementCommandCreate(
                    Type.DEVICETYPE, deviceGroupUuid,
                    "name", "Rshr",
                    "description", "comment");

            // create
            createdStructureElement = ITUtilStructureElement.assertCreate(structureElementCommandCreate);
            structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);

            // read (1)
            length = ITUtilStructureElement.assertRead("/DEVICETYPE", 1, -1).getListSize();

            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=PENDING",   1, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=APPROVED",  0, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=CANCELLED", 0, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=REJECTED",  0, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=PENDING&mnemonic=Rshr",   1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=APPROVED&mnemonic=Rshr",  0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=CANCELLED&mnemonic=Rshr", 0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=REJECTED&mnemonic=Rshr",  0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?uuid=" + createdStructureElement.getUuid().toString(), 1);

            ITUtilStructureElement.assertRead("/mnemonic/Di",          1);
            ITUtilStructureElement.assertRead("/mnemonic/Rshr",        0);
            ITUtilStructureElement.assertRead("/mnemonic/Di-Rshr",     0);

            ITUtilStructureElement.assertRead("/mnemonicPath/Di",      1, -1);
            ITUtilStructureElement.assertRead("/mnemonicPath/Rshr",    0);
            ITUtilStructureElement.assertRead("/mnemonicPath/Di-Rshr", 0);

            ITUtilStructureElement.assertRead("/history/" + createdStructureElement.getUuid().toString(), 1);

            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Rshr",    Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-Rshr", Boolean.FALSE);

            // reject
            rejectedStructureElement = ITUtilStructureElement.assertReject(structureElementCommandConfirm);

            // read (2)
            assertEquals(length, ITUtilStructureElement.assertRead("/DEVICETYPE", 1, -1).getListSize());

            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=PENDING",   0, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=APPROVED",  0, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=CANCELLED", 0, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=REJECTED",  1, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=PENDING&mnemonic=Rshr",   0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=APPROVED&mnemonic=Rshr",  0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=CANCELLED&mnemonic=Rshr", 0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=REJECTED&mnemonic=Rshr",  1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?uuid=" + rejectedStructureElement.getUuid().toString(), 1);

            ITUtilStructureElement.assertRead("/mnemonic/Di",          1);
            ITUtilStructureElement.assertRead("/mnemonic/Rshr",        0);
            ITUtilStructureElement.assertRead("/mnemonic/Di-Rshr",     0);

            ITUtilStructureElement.assertRead("/mnemonicPath/Di",      1, -1);
            ITUtilStructureElement.assertRead("/mnemonicPath/Rshr",    0);
            ITUtilStructureElement.assertRead("/mnemonicPath/Di-Rshr", 0);

            ITUtilStructureElement.assertRead("/history/" + rejectedStructureElement.getUuid().toString(), 2);

            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Rshr",    Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-Rshr", Boolean.FALSE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void readSearchHistoryStatusLatest1() {
        // purpose
        //     test read device type in various ways
        //         status
        //         latest
        //       ( deleted )
        //         history
        //
        // what
        //     entries with different statuses
        //
        // note
        //     create (and more) to read (with content)
        //     4 different lines  of uuid with combinations of status and latest
        //         1 entry for each line of uuid
        //         in particular different statuses for last entry in lines of uuid
        //         mnemonic - P, C, R, A (pending, cancelled, rejected, approved)
        //
        // note!
        //     not only
        //         exclude APPROVED and not latest
        //     not only
        //         exclude earlier than APPROVED and latest (considered history)
        //         include PENDING and not latest           (considered future)
        //     instead
        //         exclude earlier than APPROVED and latest (considered history)
        //       ( include if latest )
        //         include if latest not exists             (considered present + future)

        StructureElementCommandCreate structureElementCommandCreate = null;
        StructureElement createdStructureElement = null;
        UUID uuid, uuid2, uuid3, uuid4 = null;

        // a number of entries for which the last entry has status PENDING
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "P1",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreate(structureElementCommandCreate);
        uuid = createdStructureElement.getUuid();

        // a number of entries for which the last entry has status CANCELLED
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "C1",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateCancel(structureElementCommandCreate);
        uuid2 = createdStructureElement.getUuid();

        // a number of entries for which the last entry has status REJECTED
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "R1",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateReject(structureElementCommandCreate);
        uuid3 = createdStructureElement.getUuid();

        // a number of entries for which the last entry has status APPROVED
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "A1",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        uuid4 = createdStructureElement.getUuid();

        try {
            // from first structure element
            assertNotNull(uuid);

            // note
            //     exclude earlier than APPROVED and latest (considered history)
            //   ( include if latest )
            //     include if latest not exists             (considered present + future)

            ITUtilStructureElement.assertRead("/DEVICETYPE?uuid=" + uuid.toString(),  1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?uuid=" + uuid2.toString(), 1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?uuid=" + uuid3.toString(), 1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?uuid=" + uuid4.toString(), 1);

            ITUtilStructureElement.assertRead("/mnemonic/P1", 0);
            ITUtilStructureElement.assertRead("/mnemonic/C1", 0);
            ITUtilStructureElement.assertRead("/mnemonic/R1", 0);
            ITUtilStructureElement.assertRead("/mnemonic/A1", 1);

            ITUtilStructureElement.assertRead("/history/" + uuid.toString(),  1);
            ITUtilStructureElement.assertRead("/history/" + uuid2.toString(), 2);
            ITUtilStructureElement.assertRead("/history/" + uuid3.toString(), 2);
            ITUtilStructureElement.assertRead("/history/" + uuid4.toString(), 2);

            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "P1",    Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "C1",    Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "R1",    Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "A1",    Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-P1", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-C1", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-R1", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-A1", Boolean.TRUE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void readSearchHistoryStatusLatest9() {
        // purpose
        //     test read device type in various ways
        //         status
        //         latest
        //       ( deleted )
        //         history
        //
        // what
        //     entries with different statuses
        //
        // note
        //     create (and more) to read (with content)
        //     4 different lines  of uuid with combinations of status and latest
        //         9 entries for each line of uuid
        //         in particular different statuses for last entry in lines of uuid
        //         mnemonic - P, C, R, A (pending, cancelled, rejected, approved)
        //
        // note!
        //     not only
        //         exclude APPROVED and not latest
        //     not only
        //         exclude earlier than APPROVED and latest (considered history)
        //         include PENDING and not latest           (considered future)
        //     instead
        //         exclude earlier than APPROVED and latest (considered history)
        //       ( include if latest )
        //         include if latest not exists             (considered present + future)

        StructureElementCommandCreate structureElementCommandCreate = null;
        StructureElementCommandUpdate structureElementCommandUpdate = null;
        StructureElement createdStructureElement = null;
        UUID uuid, uuid2, uuid3, uuid4 = null;

        // a number of entries for which the last entry has status PENDING
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "P9",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(createdStructureElement);
        uuid = createdStructureElement.getUuid();

        structureElementCommandUpdate.setComment("comment 2");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 3");
        ITUtilStructureElement.assertUpdateCancel(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 4");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 5");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 6");
        ITUtilStructureElement.assertUpdateReject(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 7");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 8");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 9");
        ITUtilStructureElement.assertUpdate(structureElementCommandUpdate);

        // a number of entries for which the last entry has status CANCELLED
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "C9",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(createdStructureElement);
        uuid2 = createdStructureElement.getUuid();

        structureElementCommandUpdate.setComment("comment 2");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 3");
        ITUtilStructureElement.assertUpdateCancel(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 4");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 5");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 6");
        ITUtilStructureElement.assertUpdateReject(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 7");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 8");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 9");
        ITUtilStructureElement.assertUpdateCancel(structureElementCommandUpdate);

        // a number of entries for which the last entry has status REJECTED
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "R9",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(createdStructureElement);
        uuid3 = createdStructureElement.getUuid();

        structureElementCommandUpdate.setComment("comment 2");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 3");
        ITUtilStructureElement.assertUpdateCancel(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 4");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 5");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 6");
        ITUtilStructureElement.assertUpdateReject(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 7");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 8");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 9");
        ITUtilStructureElement.assertUpdateReject(structureElementCommandUpdate);

        // a number of entries for which the last entry has status APPROVED
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "A9",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(createdStructureElement);
        uuid4 = createdStructureElement.getUuid();

        structureElementCommandUpdate.setComment("comment 2");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 3");
        ITUtilStructureElement.assertUpdateCancel(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 4");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 5");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 6");
        ITUtilStructureElement.assertUpdateReject(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 7");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 8");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 9");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);

        try {
            // from first structure element
            assertNotNull(uuid);

            // note
            //     exclude earlier than APPROVED and latest (considered history)
            //   ( include if latest )
            //     include if latest not exists             (considered present + future)

            ITUtilStructureElement.assertRead("/DEVICETYPE?uuid=" + uuid.toString(),  2);
            ITUtilStructureElement.assertRead("/DEVICETYPE?uuid=" + uuid2.toString(), 2);
            ITUtilStructureElement.assertRead("/DEVICETYPE?uuid=" + uuid3.toString(), 2);
            ITUtilStructureElement.assertRead("/DEVICETYPE?uuid=" + uuid4.toString(), 1);

            ITUtilStructureElement.assertRead("/mnemonic/P9", 1);
            ITUtilStructureElement.assertRead("/mnemonic/C9", 1);
            ITUtilStructureElement.assertRead("/mnemonic/R9", 1);
            ITUtilStructureElement.assertRead("/mnemonic/A9", 1);

            ITUtilStructureElement.assertRead("/history/" + uuid.toString(),  17);
            ITUtilStructureElement.assertRead("/history/" + uuid2.toString(), 18);
            ITUtilStructureElement.assertRead("/history/" + uuid3.toString(), 18);
            ITUtilStructureElement.assertRead("/history/" + uuid4.toString(), 18);

            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "P9",    Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "C9",    Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "R9",    Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "A9",    Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-P9", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-C9", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-R9", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-A9", Boolean.TRUE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void readSearchStatusDeletedChildren() {
        // purpose
        //     test read device type in various ways
        //         status
        //       ( latest )
        //         deleted
        //         children
        //
        // what
        //     entries with different statuses
        //
        // note
        //     create (and more) to read (with content)
        //     querying for Status.APPROVED means latest approved

        StructureElementCommandCreate structureElementCommandCreate = null;
        StructureElementCommandUpdate structureElementCommandUpdate = null;
        StructureElementCommandConfirm structureElementCommandConfirm = null;
        StructureElement createdStructureElement = null;
        UUID uuid = null;

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AA1",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        uuid = createdStructureElement.getUuid();

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AA2",
                "description", "comment");
        ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AA3",
                "description", "comment");
        ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AA4",
                "description", "comment");
        ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AA5",
                "description", "comment");
        ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AB1",
                "description", "comment");
        ITUtilStructureElement.assertCreateCancel(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AB2",
                "description", "comment");
        ITUtilStructureElement.assertCreateCancel(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AB3",
                "description", "comment");
        ITUtilStructureElement.assertCreateCancel(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AB4",
                "description", "comment");
        ITUtilStructureElement.assertCreateCancel(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AB5",
                "description", "comment");
        ITUtilStructureElement.assertCreateCancel(structureElementCommandCreate);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AC1",
                "description", "comment");
        ITUtilStructureElement.assertCreateReject(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AC2",
                "description", "comment");
        ITUtilStructureElement.assertCreateReject(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AC3",
                "description", "comment");
        ITUtilStructureElement.assertCreateReject(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AC4",
                "description", "comment");
        ITUtilStructureElement.assertCreateReject(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AC5",
                "description", "comment");
        ITUtilStructureElement.assertCreateReject(structureElementCommandCreate);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AD1",
                "description", "comment");
        ITUtilStructureElement.assertCreate(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AD2",
                "description", "comment");
        ITUtilStructureElement.assertCreate(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AD3",
                "description", "comment");
        ITUtilStructureElement.assertCreate(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AD4",
                "description", "comment");
        ITUtilStructureElement.assertCreate(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AD5",
                "description", "comment");
        ITUtilStructureElement.assertCreate(structureElementCommandCreate);

        String description2 = "some other description";
        String comment2 = "some other comment";
        String description3 = "more description";
        String comment3 = "more comment";

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AE1",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(createdStructureElement);
        structureElementCommandUpdate.setDescription(description2);
        structureElementCommandUpdate.setComment(comment2);
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(createdStructureElement);
        structureElementCommandUpdate.setDescription(description3);
        structureElementCommandUpdate.setComment(comment3);
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDeleteReject(structureElementCommandConfirm);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AE2",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(createdStructureElement);
        structureElementCommandUpdate.setDescription(description2);
        structureElementCommandUpdate.setComment(comment2);
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(createdStructureElement);
        structureElementCommandUpdate.setDescription(description3);
        structureElementCommandUpdate.setComment(comment3);
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDeleteReject(structureElementCommandConfirm);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AE3",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(createdStructureElement);
        structureElementCommandUpdate.setDescription(description2);
        structureElementCommandUpdate.setComment(comment2);
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(createdStructureElement);
        structureElementCommandUpdate.setDescription(description3);
        structureElementCommandUpdate.setComment(comment3);
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDeleteReject(structureElementCommandConfirm);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AE4",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(createdStructureElement);
        structureElementCommandUpdate.setDescription(description2);
        structureElementCommandUpdate.setComment(comment2);
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(createdStructureElement);
        structureElementCommandUpdate.setDescription(description3);
        structureElementCommandUpdate.setComment(comment3);
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDeleteReject(structureElementCommandConfirm);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AE5",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(createdStructureElement);
        structureElementCommandUpdate.setDescription(description2);
        structureElementCommandUpdate.setComment(comment2);
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(createdStructureElement);
        structureElementCommandUpdate.setDescription(description3);
        structureElementCommandUpdate.setComment(comment3);
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDeleteReject(structureElementCommandConfirm);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AF1",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDeleteApprove(structureElementCommandConfirm);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AF1",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDeleteApprove(structureElementCommandConfirm);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AF3",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDeleteApprove(structureElementCommandConfirm);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AF4",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDeleteApprove(structureElementCommandConfirm);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AF5",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDeleteApprove(structureElementCommandConfirm);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AG1",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDelete(structureElementCommandConfirm);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AG2",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDelete(structureElementCommandConfirm);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AG3",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDelete(structureElementCommandConfirm);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AG4",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDelete(structureElementCommandConfirm);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.DEVICETYPE, deviceGroupUuid,
                "name", "AG5",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDelete(structureElementCommandConfirm);

        // 60 device type entries

        try {
            // from first structure element
            assertNotNull(uuid);

            // read & search
            //     /{type}
            //     type, statuses, deleted, uuid, parent, name, mnemonic, mnemonicEquivalence, mnemonicPath, description
            //     combination
            ITUtilStructureElement.assertRead("/DEVICETYPE?mnemonic=A__",                                                  45);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=PENDING&mnemonic=A__",                                 10);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=APPROVED&mnemonic=A__",                                20);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=CANCELLED&mnemonic=A__",                                5);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=REJECTED&mnemonic=A__",                                10);
            ITUtilStructureElement.assertRead("/DEVICETYPE?statuses=PENDING&statuses=APPROVED&mnemonic=A__",               30);

            ITUtilStructureElement.assertRead("/DEVICETYPE?deleted=false&mnemonic=A__",                                    30);
            ITUtilStructureElement.assertRead("/DEVICETYPE?deleted=false&statuses=PENDING&mnemonic=A__",                    5);
            ITUtilStructureElement.assertRead("/DEVICETYPE?deleted=false&statuses=APPROVED&mnemonic=A__",                  15);
            ITUtilStructureElement.assertRead("/DEVICETYPE?deleted=false&statuses=CANCELLED&mnemonic=A__",                  5);
            ITUtilStructureElement.assertRead("/DEVICETYPE?deleted=false&statuses=REJECTED&mnemonic=A__",                   5);
            ITUtilStructureElement.assertRead("/DEVICETYPE?deleted=false&statuses=PENDING&statuses=APPROVED&mnemonic=A__", 20);

            ITUtilStructureElement.assertRead("/DEVICETYPE?deleted=true&mnemonic=A__",                                     15);
            ITUtilStructureElement.assertRead("/DEVICETYPE?deleted=true&statuses=PENDING&mnemonic=A__",                     5);
            ITUtilStructureElement.assertRead("/DEVICETYPE?deleted=true&statuses=APPROVED&mnemonic=A__",                    5);
            ITUtilStructureElement.assertRead("/DEVICETYPE?deleted=true&statuses=CANCELLED&mnemonic=A__",                   0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?deleted=true&statuses=REJECTED&mnemonic=A__",                    5);
            ITUtilStructureElement.assertRead("/DEVICETYPE?deleted=true&statuses=PENDING&statuses=APPROVED&mnemonic=A__",  10);

            ITUtilStructureElement.assertRead("/DEVICETYPE?uuid=" + disciplineUuid.toString(),     0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?uuid=" + deviceGroupUuid.toString(),    0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?uuid=" + uuid.toString(),               1);

            ITUtilStructureElement.assertRead("/DEVICETYPE?parent=" + disciplineUuid.toString(),   0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?parent=" + deviceGroupUuid.toString(), 45, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?parent=" + uuid.toString(),             0);

            ITUtilStructureElement.assertRead("/DEVICETYPE?name=na",                  0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?name=na_",                 0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?name=na__",               45, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?name=na___",               0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?name=na%",                45, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?name=name",               45, -1);

            ITUtilStructureElement.assertRead("/DEVICETYPE?mnemonicEquivalence=A__", 45, -1);

            ITUtilStructureElement.assertRead("/DEVICETYPE?mnemonicPath=A__",         0);

            ITUtilStructureElement.assertRead("/DEVICETYPE?description=desc",         0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?description=desc%",       35, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?description=sc",           0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?description=sc%",          0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?description=%sc",          0);
            ITUtilStructureElement.assertRead("/DEVICETYPE?description=%sc%",        45, -1);
            ITUtilStructureElement.assertRead("/DEVICETYPE?description=description", 35, -1);

            // children
            //     /children/{uuid}
            //     uuid, type, statuses, deleted, name, mnemonic, mnemonicEquivalence, mnemonicPath, description
            ITUtilStructureElement.assertRead("/children/" + uuid.toString(),                                   0);
            ITUtilStructureElement.assertRead("/children/" + deviceGroupUuid.toString(),                       15, -1);
            ITUtilStructureElement.assertRead("/children/" + disciplineUuid.toString(),                         1);
            ITUtilStructureElement.assertRead("/children/" + uuid.toString()            + "?type=DEVICETYPE",   0);
            ITUtilStructureElement.assertRead("/children/" + deviceGroupUuid.toString() + "?type=DEVICETYPE",   0);
            ITUtilStructureElement.assertRead("/children/" + disciplineUuid.toString()  + "?type=DEVICETYPE",   0);
            ITUtilStructureElement.assertRead("/children/" + uuid.toString()            + "?type=DEVICEGROUP",  0);
            ITUtilStructureElement.assertRead("/children/" + deviceGroupUuid.toString() + "?type=DEVICEGROUP", 15, -1);
            ITUtilStructureElement.assertRead("/children/" + disciplineUuid.toString()  + "?type=DEVICEGROUP",  0);
            ITUtilStructureElement.assertRead("/children/" + uuid.toString()            + "?type=DISCIPLINE",   0);
            ITUtilStructureElement.assertRead("/children/" + deviceGroupUuid.toString() + "?type=DISCIPLINE",   0);
            ITUtilStructureElement.assertRead("/children/" + disciplineUuid.toString()  + "?type=DISCIPLINE",   1);

            // mnemonic
            //     /mnemonic/{mnemonic}
            ITUtilStructureElement.assertRead("/mnemonic/A",           0);
            ITUtilStructureElement.assertRead("/mnemonic/A__",        15, -1);
            ITUtilStructureElement.assertRead("/mnemonic/AG_",         5);
            ITUtilStructureElement.assertRead("/mnemonic/AG1",         1);
            ITUtilStructureElement.assertRead("/mnemonic/Di",          1);

            // mnemonic path
            //     /mnemonicPath/{mnemonicPath}
            ITUtilStructureElement.assertRead("/mnemonicPath/A",       0);
            ITUtilStructureElement.assertRead("/mnemonicPath/A__",     0);
            ITUtilStructureElement.assertRead("/mnemonicPath/AG_",     0);
            ITUtilStructureElement.assertRead("/mnemonicPath/AG1",     0);
            ITUtilStructureElement.assertRead("/mnemonicPath/Di-A%",  15, -1);
            ITUtilStructureElement.assertRead("/mnemonicPath/Di-A",    0);
            ITUtilStructureElement.assertRead("/mnemonicPath/Di-A__", 15, -1);
            ITUtilStructureElement.assertRead("/mnemonicPath/Di-AG_",  5);
            ITUtilStructureElement.assertRead("/mnemonicPath/Di-AG1",  1);

            // history
            //     /history/{uuid}
            ITUtilStructureElement.assertRead("/history/" + uuid.toString(),            2);
            ITUtilStructureElement.assertRead("/history/" + deviceGroupUuid.toString(), 2);
            ITUtilStructureElement.assertRead("/history/" + disciplineUuid.toString(),  2);

            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AA1", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AA2", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AA3", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AA4", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AA5", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AB1", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AB2", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AB3", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AB4", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AB5", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AC1", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AC2", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AC3", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AC4", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AC5", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AD1", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AD2", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AD3", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AD4", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AD5", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AE1", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AE2", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AE3", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AE4", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AE5", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AF1", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AF2", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AF3", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AF4", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AF5", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AG1", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AG2", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AG3", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AG4", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DEVICETYPE, "Di-AG5", Boolean.TRUE);
        } catch (Exception e) {
            fail();
        }
    }

}
