/*
 * Copyright (C) 2022 European Spallation Source ERIC.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.openepics.names.docker;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.fail;

import java.io.File;
import java.net.HttpURLConnection;
import java.util.UUID;

import org.junit.jupiter.api.Test;
import org.openepics.names.rest.beans.Type;
import org.openepics.names.rest.beans.element.StructureElement;
import org.openepics.names.rest.beans.element.StructureElementCommand;
import org.openepics.names.rest.beans.element.StructureElementCommandConfirm;
import org.openepics.names.rest.beans.element.StructureElementCommandCreate;
import org.openepics.names.rest.beans.element.StructureElementCommandUpdate;
import org.openepics.names.util.StructureCommand;
import org.openepics.names.util.StructureElementUtil;
import org.testcontainers.containers.DockerComposeContainer;
import org.testcontainers.containers.wait.strategy.Wait;
import org.testcontainers.junit.jupiter.Container;
import org.testcontainers.junit.jupiter.Testcontainers;

/**
 * Integration tests for Naming and PostgreSQL that make use of existing dockerization
 * with docker-compose.yml / Dockerfile.
 *
 * <p>
 * Focus of this class is structures endpoint and system group.
 * </p>
 *
 * @author Lars Johansson
 */
@Testcontainers
class StructuresSystemGroupIT {

    // note
    //     StructureElement - when, who, comment - depend on requested or processed
    //         requested, requested by, requested comment
    //         processed, processed by, processed comment
    //     if less than a second between requested and processed, then considered one entry with processed
    //     history
    //         mnemonic path does not make same sense for history
    //         (very) tricky to find mnemonic path for uuid at proper time (history)
    //         therefore empty mnemonic path for history for structure
    //         one history entry if less than one second between requested and processed, otherwise two history entries
    //     attributes for entry for operations - create, update, delete, approve, cancel, reject
    //         some set client side, others set server side
    //         client side
    //             type, uuid, parent uuid, name, mnemonic, description, comment
    //         may be set client side for test purposes
    //     order of methods in test not known
    //         therefore mnemonic values chosen to not interfere with mnemonic values in other methods
    //
    //     StructureElement extends StructureElementCommand
    //     -->
    //     for such operations that send Command objects to backend, it is also possible to send StructureElement
    //
    //     StructureElementCommand
    //	       create  -       type, parent, name, mnemonic, description, comment
    //	       update  - uuid, type, parent, name, mnemonic, description, comment
    //	       delete  - uuid, type,                                      comment
    //	       approve - uuid, type,                                      comment
    //	       cancel  - uuid, type,                                      comment
    //	       reject  - uuid, type,                                      comment

    @Container
    public static final DockerComposeContainer<?> ENVIRONMENT =
        new DockerComposeContainer<>(new File("docker-compose-integrationtest.yml"))
            .waitingFor(ITUtil.NAMING, Wait.forLogMessage(".*Started NamingApplication.*", 1));

    @Test
    void equivalenceMnemonic() {
        // purpose
        //     test mnemonic equivalence
        //
        // what
        //     read        equivalence mnemonic

        ITUtilStructureElement.assertEquivalence("Sg", "SG");
    }

    @Test
    void checkCreate() {
        // purpose
        //     test conditions for create system group
        //         not create itself
        //
        // what - combination of
        //     read        exists in structure
        //     read        is valid to create
        //     read        validate create
        //
        // note
        //     with and without mnemonic
        //
        // validate create
        //     type
        //     name
        //     mnemonic
        //     description
        //     comment

        try {
            StructureElementCommandCreate structureElementCommandCreate = new StructureElementCommandCreate();

            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "Cc", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "CC", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.SYSTEMGROUP, "Cc", Boolean.TRUE);
            ITUtilStructureElement.assertIsValidToCreate(Type.SYSTEMGROUP, "CC", Boolean.TRUE);

            // validate create

            ITUtilStructureElement.assertValidate("[{asdf]", StructureCommand.CREATE, HttpURLConnection.HTTP_BAD_REQUEST);
            ITUtilStructureElement.assertCreate("[{asdf]", HttpURLConnection.HTTP_BAD_REQUEST);

            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setType(Type.SYSTEMGROUP);
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setName("name");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setMnemonic("Cc");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setDescription("description");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setComment("comment");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic(null);
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setType(null);
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void checkCreateMnemonic() {
        // purpose
        //     test conditions for create system group
        //         not create itself
        //
        // what - combination of
        //     read        validate create
        //
        // note
        //     mnemonic
        try {
            StructureElementCommandCreate structureElementCommandCreate = new StructureElementCommandCreate();

            structureElementCommandCreate.setType(Type.SYSTEMGROUP);
            structureElementCommandCreate.setName("name");
            structureElementCommandCreate.setDescription("description");
            structureElementCommandCreate.setComment("comment");

            // mnemonic rules

            structureElementCommandCreate.setMnemonic(null);
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("C");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("Cc");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("Ccc");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("Cccc");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("Ccccc");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("Cccccc");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("Ccccccc");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setMnemonic("Cccccccc");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setMnemonic("Ccccccccc");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            // mnemonic rules (2)

            structureElementCommandCreate.setMnemonic(" ");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setMnemonic("Sys ");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setMnemonic("Sys");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("000");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("Sys0");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic(":");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setMnemonic("Sys:");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setMnemonic("Sys:   ");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setMnemonic("1");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("12");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("123");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("1234");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("12345");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("123456");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("1234567");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setMnemonic("12345678");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setMnemonic("123456789");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandCreate.setMnemonic("Ac1");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);

            structureElementCommandCreate.setMnemonic("Acc");
            ITUtilStructureElement.assertValidate(structureElementCommandCreate, StructureCommand.CREATE, Boolean.TRUE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void createApprove() {
        // purpose
        //     test create and approve system group
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate approve
        //     patch       approve structures
        //
        // note
        //     create in order to approve
        //     with and without mnemonic

        try {
            StructureElementCommandCreate structureElementCommandCreate = null;
            StructureElementCommandConfirm structureElementCommandConfirm = null;
            StructureElementCommand structureElementCommand = null;
            StructureElement createdStructureElement = null;
            UUID uuid, uuid2, uuid3 = null;

            structureElementCommand = new StructureElementCommand(
                    null, Type.SYSTEMGROUP, null,
                    "name", "Ca",
                    "description", "comment");
            structureElementCommandCreate = StructureElementUtil.convertCommand2CommandCreate(structureElementCommand);
            structureElementCommandConfirm = StructureElementUtil.convertCommand2CommandConfirm(structureElementCommand);

            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "Ca", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.SYSTEMGROUP, "Ca", Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandCreate,  StructureCommand.CREATE,  Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.APPROVE, Boolean.FALSE);

            // create
            createdStructureElement = ITUtilStructureElement.assertCreate(structureElementCommandCreate);
            structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
            uuid = createdStructureElement.getUuid();

            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "Ca", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.SYSTEMGROUP, "Ca", Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandCreate,  StructureCommand.CREATE,  Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.APPROVE, Boolean.TRUE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, HttpURLConnection.HTTP_CONFLICT);

            // approve
            ITUtilStructureElement.assertApprove(structureElementCommandConfirm);

            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "Ca", Boolean.TRUE);
            ITUtilStructureElement.assertIsValidToCreate(Type.SYSTEMGROUP, "Ca", Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElementCommandCreate,  StructureCommand.CREATE,  Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.APPROVE, Boolean.FALSE);

            structureElementCommandConfirm.setUuid(null);
            ITUtilStructureElement.assertApprove(structureElementCommandConfirm, ITUtil.HTTP_UNPROCESSABLE_ENTITY);
            structureElementCommandConfirm.setUuid(UUID.randomUUID());
            ITUtilStructureElement.assertApprove(structureElementCommandConfirm, HttpURLConnection.HTTP_NOT_FOUND);
            structureElementCommandConfirm.setUuid(uuid);
            ITUtilStructureElement.assertApprove(structureElementCommandConfirm, HttpURLConnection.HTTP_CONFLICT);

            // create
            structureElementCommandCreate = new StructureElementCommandCreate(
                    Type.SYSTEMGROUP, null,
                    "name2", null,
                    "description2", "comment2");
            createdStructureElement = ITUtilStructureElement.assertCreate(structureElementCommandCreate);
            uuid2 = createdStructureElement.getUuid();

            // create
            structureElementCommandCreate = new StructureElementCommandCreate(
                    Type.SYSTEMGROUP, null,
                    "name3", null,
                    "description3", "comment3");
            createdStructureElement = ITUtilStructureElement.assertCreate(structureElementCommandCreate);
            uuid3 = createdStructureElement.getUuid();

            assertNotEquals(uuid, uuid2);
            assertNotEquals(uuid, uuid3);
            assertNotEquals(uuid2, uuid3);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void createCancel() {
        // purpose
        //     test create and cancel system group
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate cancel
        //     patch       cancel structures
        //
        // note
        //     create in order to cancel

        try {
            StructureElementCommandCreate structureElementCommandCreate = null;
            StructureElementCommandConfirm structureElementCommandConfirm = null;
            StructureElementCommand structureElementCommand = null;
            StructureElement createdStructureElement = null;

            structureElementCommand = new StructureElementCommand(
                    null, Type.SYSTEMGROUP, null,
                    "name", "Cc",
                    "description", "comment");
            structureElementCommandCreate = StructureElementUtil.convertCommand2CommandCreate(structureElementCommand);
            structureElementCommandConfirm = StructureElementUtil.convertCommand2CommandConfirm(structureElementCommand);

            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "Cc", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.SYSTEMGROUP, "Cc", Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandCreate,  StructureCommand.CREATE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.CANCEL, Boolean.FALSE);

            // create
            createdStructureElement = ITUtilStructureElement.assertCreate(structureElementCommandCreate);
            structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
            UUID uuid = structureElementCommandConfirm.getUuid();

            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "Cc", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.SYSTEMGROUP, "Cc", Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandCreate,  StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.CANCEL, Boolean.TRUE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, HttpURLConnection.HTTP_CONFLICT);

            structureElementCommandConfirm.setUuid(null);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.CANCEL, Boolean.FALSE);
            structureElementCommandConfirm.setUuid(uuid);

            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.CANCEL, Boolean.TRUE);
            structureElementCommandConfirm.setType(null);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.CANCEL, Boolean.FALSE);
            structureElementCommandConfirm.setType(Type.SYSTEMGROUP);

            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.CANCEL, Boolean.TRUE);
            structureElementCommandConfirm.setComment(null);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.CANCEL, Boolean.FALSE);
            structureElementCommandConfirm.setComment("comment");

            // cancel
            ITUtilStructureElement.assertCancel(structureElementCommandConfirm);

            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "Cc", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.SYSTEMGROUP, "Cc", Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandCreate,  StructureCommand.CREATE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.CANCEL, Boolean.FALSE);

            structureElementCommandConfirm.setUuid(null);
            ITUtilStructureElement.assertCancel(structureElementCommandConfirm, ITUtil.HTTP_UNPROCESSABLE_ENTITY);
            structureElementCommandConfirm.setUuid(UUID.randomUUID());
            ITUtilStructureElement.assertCancel(structureElementCommandConfirm, HttpURLConnection.HTTP_NOT_FOUND);
            structureElementCommandConfirm.setUuid(uuid);
            ITUtilStructureElement.assertCancel(structureElementCommandConfirm, HttpURLConnection.HTTP_CONFLICT);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void createReject() {
        // purpose
        //     test create and reject system group
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate reject
        //     patch       reject structures
        //
        // note
        //     create in order to reject

        try {
            StructureElementCommandCreate structureElementCommandCreate = null;
            StructureElementCommandConfirm structureElementCommandConfirm = null;
            StructureElementCommand structureElementCommand = null;
            StructureElement createdStructureElement = null;

            structureElementCommand = new StructureElementCommand(
                    null, Type.SYSTEMGROUP, null,
                    "name", "Cr",
                    "description", "comment");
            structureElementCommandCreate = StructureElementUtil.convertCommand2CommandCreate(structureElementCommand);
            structureElementCommandConfirm = StructureElementUtil.convertCommand2CommandConfirm(structureElementCommand);

            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "Cr", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.SYSTEMGROUP, "Cr", Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandCreate,  StructureCommand.CREATE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.REJECT, Boolean.FALSE);

            // create
            createdStructureElement = ITUtilStructureElement.assertCreate(structureElementCommandCreate);
            structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
            UUID uuid = structureElementCommandConfirm.getUuid();

            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "Cr", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.SYSTEMGROUP, "Cr", Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandCreate,  StructureCommand.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.REJECT, Boolean.TRUE);
            ITUtilStructureElement.assertCreate(structureElementCommandCreate, HttpURLConnection.HTTP_CONFLICT);

            structureElementCommandConfirm.setUuid(null);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.REJECT, Boolean.FALSE);
            structureElementCommandConfirm.setUuid(uuid);

            // reject
            ITUtilStructureElement.assertReject(structureElementCommandConfirm);

            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "Cr", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.SYSTEMGROUP, "Cr", Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandCreate,  StructureCommand.CREATE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.REJECT, Boolean.FALSE);

            structureElementCommandConfirm.setUuid(null);
            ITUtilStructureElement.assertReject(structureElementCommandConfirm, ITUtil.HTTP_UNPROCESSABLE_ENTITY);
            structureElementCommandConfirm.setUuid(UUID.randomUUID());
            ITUtilStructureElement.assertReject(structureElementCommandConfirm, HttpURLConnection.HTTP_NOT_FOUND);
            structureElementCommandConfirm.setUuid(uuid);
            ITUtilStructureElement.assertReject(structureElementCommandConfirm, HttpURLConnection.HTTP_CONFLICT);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void checkUpdate() {
        // purpose
        //     test conditions for update system group
        //         not update
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate update
        //     read        validate approve
        //     patch       approve structures
        //
        // note
        //     create, approve in order to update
        //
        // validate update
        //     uuid
        //     type
        //     name
        //     mnemonic
        //     description
        //     comment

        try {
            UUID uuid = null;

            StructureElementCommandCreate structureElementCommandCreate = null;
            StructureElementCommandUpdate structureElementCommandUpdate = null;
            StructureElement approvedStructureElement = null;

            // create, approve
            structureElementCommandCreate = new StructureElementCommandCreate(
                    Type.SYSTEMGROUP, null,
                    "name", "Cu",
                    "description", "comment");
            approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
            structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(approvedStructureElement);
            uuid = approvedStructureElement.getUuid();

            // validate update

            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.TRUE);

            structureElementCommandUpdate.setUuid(null);
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.FALSE);
            ITUtilStructureElement.assertUpdate(structureElementCommandUpdate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandUpdate.setUuid(UUID.randomUUID());
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.FALSE);
            ITUtilStructureElement.assertUpdate(structureElementCommandUpdate, HttpURLConnection.HTTP_NOT_FOUND);

            structureElementCommandUpdate.setUuid(uuid);
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.TRUE);

            structureElementCommandUpdate.setType(null);
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.FALSE);
            ITUtilStructureElement.assertUpdate(structureElementCommandUpdate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandUpdate.setType(Type.DISCIPLINE);
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.FALSE);
            ITUtilStructureElement.assertUpdate(structureElementCommandUpdate, HttpURLConnection.HTTP_NOT_FOUND);

            structureElementCommandUpdate.setType(Type.SYSTEMGROUP);
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.TRUE);

            structureElementCommandUpdate.setName(null);
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.FALSE);
            ITUtilStructureElement.assertUpdate(structureElementCommandUpdate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandUpdate.setName("name");
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.TRUE);

            structureElementCommandUpdate.setMnemonic(null);
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.TRUE);

            structureElementCommandUpdate.setMnemonic("Cu");
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.TRUE);

            structureElementCommandUpdate.setDescription(null);
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.FALSE);
            ITUtilStructureElement.assertUpdate(structureElementCommandUpdate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandUpdate.setDescription("description");
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.TRUE);

            structureElementCommandUpdate.setComment(null);
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.FALSE);
            ITUtilStructureElement.assertUpdate(structureElementCommandUpdate, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandUpdate.setComment("comment");
            ITUtilStructureElement.assertValidate(structureElementCommandUpdate, StructureCommand.UPDATE, Boolean.TRUE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void updateApprove() {
        // purpose
        //     test update and approve system group
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate update
        //     read        validate approve
        //     update      update structures
        //     patch       approve structures
        //
        // note
        //     create, approve in order to update, approve

        try {
            StructureElementCommandCreate structureElementCommandCreate = null;
            StructureElementCommandUpdate structureElementCommandUpdate = null;
            StructureElementCommandConfirm structureElementCommandConfirm = null;
            StructureElement approvedStructureElement = null;

            // create, approve
            structureElementCommandCreate = new StructureElementCommandCreate(
                    Type.SYSTEMGROUP, null,
                    "name", "Ua",
                    "description", "comment");
            approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
            structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(approvedStructureElement);
            structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(approvedStructureElement);

            structureElementCommandUpdate.setDescription("description update approve check");
            structureElementCommandUpdate.setComment("comment update approve check");

            ITUtilStructureElement.assertValidate(structureElementCommandUpdate,  StructureCommand.UPDATE,  Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.APPROVE, Boolean.FALSE);

            // update
            ITUtilStructureElement.assertUpdate(structureElementCommandUpdate);

            ITUtilStructureElement.assertValidate(structureElementCommandUpdate,  StructureCommand.UPDATE,  Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.APPROVE, Boolean.TRUE);
            ITUtilStructureElement.assertUpdate(structureElementCommandUpdate, HttpURLConnection.HTTP_CONFLICT);

            // approve
            ITUtilStructureElement.assertApprove(structureElementCommandConfirm);

            ITUtilStructureElement.assertValidate(structureElementCommandUpdate,  StructureCommand.UPDATE,  Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.APPROVE, Boolean.FALSE);

            structureElementCommandConfirm.setUuid(null);
            ITUtilStructureElement.assertApprove(structureElementCommandConfirm, ITUtil.HTTP_UNPROCESSABLE_ENTITY);
            structureElementCommandConfirm.setUuid(UUID.randomUUID());
            ITUtilStructureElement.assertApprove(structureElementCommandConfirm, HttpURLConnection.HTTP_NOT_FOUND);
            structureElementCommandConfirm.setUuid(approvedStructureElement.getUuid());
            ITUtilStructureElement.assertApprove(structureElementCommandConfirm, HttpURLConnection.HTTP_CONFLICT);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void updateCancel() {
        // purpose
        //     test update and cancel system group
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate update
        //     read        validate approve
        //     read        validate cancel
        //     update      update structures
        //     patch       approve structures
        //     patch       cancel structures
        //
        // note
        //     create, approve in order to update, cancel

        try {
            StructureElementCommandCreate structureElementCommandCreate = null;
            StructureElementCommandUpdate structureElementCommandUpdate = null;
            StructureElementCommandConfirm structureElementCommandConfirm = null;
            StructureElement approvedStructureElement = null;

            // create, approve
            structureElementCommandCreate = new StructureElementCommandCreate(
                    Type.SYSTEMGROUP, null,
                    "name", "Uc",
                    "description", "comment");
            approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
            structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(approvedStructureElement);
            structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(approvedStructureElement);

            structureElementCommandUpdate.setDescription("description update cancel check");
            structureElementCommandUpdate.setComment("comment update cancel check");

            ITUtilStructureElement.assertValidate(structureElementCommandUpdate,  StructureCommand.UPDATE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.CANCEL, Boolean.FALSE);

            // update
            ITUtilStructureElement.assertUpdate(structureElementCommandUpdate);

            ITUtilStructureElement.assertValidate(structureElementCommandUpdate,  StructureCommand.UPDATE, Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.CANCEL, Boolean.TRUE);
            ITUtilStructureElement.assertUpdate(structureElementCommandUpdate, HttpURLConnection.HTTP_CONFLICT);

            // cancel
            ITUtilStructureElement.assertCancel(structureElementCommandConfirm);

            ITUtilStructureElement.assertValidate(structureElementCommandUpdate,  StructureCommand.UPDATE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.CANCEL, Boolean.FALSE);

            structureElementCommandConfirm.setUuid(null);
            ITUtilStructureElement.assertCancel(structureElementCommandConfirm, ITUtil.HTTP_UNPROCESSABLE_ENTITY);
            structureElementCommandConfirm.setUuid(UUID.randomUUID());
            ITUtilStructureElement.assertCancel(structureElementCommandConfirm, HttpURLConnection.HTTP_NOT_FOUND);
            structureElementCommandConfirm.setUuid(approvedStructureElement.getUuid());
            ITUtilStructureElement.assertCancel(structureElementCommandConfirm, HttpURLConnection.HTTP_CONFLICT);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void updateReject() {
        // purpose
        //     test update and reject system group
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate update
        //     read        validate approve
        //     read        validate reject
        //     update      update structures
        //     patch       approve structures
        //     patch       reject structures
        //
        // note
        //     create, approve in order to update, reject

        try {
            StructureElementCommandCreate structureElementCommandCreate = null;
            StructureElementCommandUpdate structureElementCommandUpdate = null;
            StructureElementCommandConfirm structureElementCommandConfirm = null;
            StructureElement approvedStructureElement = null;

            // create, approve
            structureElementCommandCreate = new StructureElementCommandCreate(
                    Type.SYSTEMGROUP, null,
                    "name", "Ur",
                    "description", "comment");
            approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
            structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(approvedStructureElement);
            structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(approvedStructureElement);

            structureElementCommandUpdate.setDescription("description update reject check");
            structureElementCommandUpdate.setComment("comment update reject check");

            ITUtilStructureElement.assertValidate(structureElementCommandUpdate,  StructureCommand.UPDATE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.REJECT, Boolean.FALSE);

            // update
            ITUtilStructureElement.assertUpdate(structureElementCommandUpdate);

            ITUtilStructureElement.assertValidate(structureElementCommandUpdate,  StructureCommand.UPDATE, Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.REJECT, Boolean.TRUE);
            ITUtilStructureElement.assertUpdate(structureElementCommandUpdate, HttpURLConnection.HTTP_CONFLICT);

            // reject
            ITUtilStructureElement.assertReject(structureElementCommandConfirm);

            ITUtilStructureElement.assertValidate(structureElementCommandUpdate,  StructureCommand.UPDATE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.REJECT, Boolean.FALSE);

            structureElementCommandConfirm.setUuid(null);
            ITUtilStructureElement.assertReject(structureElementCommandConfirm, ITUtil.HTTP_UNPROCESSABLE_ENTITY);
            structureElementCommandConfirm.setUuid(UUID.randomUUID());
            ITUtilStructureElement.assertReject(structureElementCommandConfirm, HttpURLConnection.HTTP_NOT_FOUND);
            structureElementCommandConfirm.setUuid(approvedStructureElement.getUuid());
            ITUtilStructureElement.assertReject(structureElementCommandConfirm, HttpURLConnection.HTTP_CONFLICT);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void checkDelete() {
        // purpose
        //     test conditions for delete system group
        //         not delete
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate delete
        //     read        validate approve
        //     patch       approve structures
        //
        // note
        //     create, approve in order to delete
        //
        // validate delete
        //     uuid
        //     type
        //     comment

        try {
            UUID uuid = null;

            StructureElementCommandCreate structureElementCommandCreate = null;
            StructureElementCommandConfirm structureElementCommandConfirm = null;
            StructureElement approvedStructureElement = null;

            // create, approve
            structureElementCommandCreate = new StructureElementCommandCreate(
                    Type.SYSTEMGROUP, null,
                    "name", "Cd",
                    "description", "comment");
            approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
            structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(approvedStructureElement);
            uuid = approvedStructureElement.getUuid();

            // validate delete

            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE, Boolean.TRUE);

            structureElementCommandConfirm.setUuid(null);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE, Boolean.FALSE);
            ITUtilStructureElement.assertDelete(structureElementCommandConfirm, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandConfirm.setUuid(UUID.randomUUID());
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE, Boolean.FALSE);
            ITUtilStructureElement.assertDelete(structureElementCommandConfirm, HttpURLConnection.HTTP_NOT_FOUND);

            structureElementCommandConfirm.setUuid(uuid);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE, Boolean.TRUE);

            structureElementCommandConfirm.setType(null);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE, Boolean.FALSE);
            ITUtilStructureElement.assertDelete(structureElementCommandConfirm, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandConfirm.setType(Type.DISCIPLINE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.UPDATE, Boolean.FALSE);
            ITUtilStructureElement.assertDelete(structureElementCommandConfirm, HttpURLConnection.HTTP_NOT_FOUND);

            structureElementCommandConfirm.setType(Type.SYSTEMGROUP);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE, Boolean.TRUE);

            structureElementCommandConfirm.setComment(null);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE, Boolean.FALSE);
            ITUtilStructureElement.assertDelete(structureElementCommandConfirm, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElementCommandConfirm.setComment("comment");
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE, Boolean.TRUE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void deleteApprove() {
        // purpose
        //     test delete and approve system group
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate delete
        //     read        validate approve
        //     delete      delete structures
        //     patch       approve structures
        //
        // note
        //     create, approve in order to delete, approve

        try {
            StructureElementCommandCreate structureElementCommandCreate = null;
            StructureElementCommandConfirm structureElementCommandConfirm = null;
            StructureElement approvedStructureElement = null;

            // create, approve
            structureElementCommandCreate = new StructureElementCommandCreate(
                    Type.SYSTEMGROUP, null,
                    "name", "Da",
                    "description", "comment");
            approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
            structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(approvedStructureElement);

            structureElementCommandConfirm.setComment("comment delete approve check");

            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE,  Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.APPROVE, Boolean.FALSE);

            // delete
            ITUtilStructureElement.assertDelete(structureElementCommandConfirm);

            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE,  Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.APPROVE, Boolean.TRUE);
            ITUtilStructureElement.assertDelete(structureElementCommandConfirm, HttpURLConnection.HTTP_CONFLICT);

            // approve
            ITUtilStructureElement.assertApprove(structureElementCommandConfirm);

            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE,  Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.APPROVE, Boolean.FALSE);

            structureElementCommandConfirm.setUuid(null);
            ITUtilStructureElement.assertApprove(structureElementCommandConfirm, ITUtil.HTTP_UNPROCESSABLE_ENTITY);
            structureElementCommandConfirm.setUuid(UUID.randomUUID());
            ITUtilStructureElement.assertApprove(structureElementCommandConfirm, HttpURLConnection.HTTP_NOT_FOUND);
            structureElementCommandConfirm.setUuid(approvedStructureElement.getUuid());
            ITUtilStructureElement.assertApprove(structureElementCommandConfirm, HttpURLConnection.HTTP_CONFLICT);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void deleteCancel() {
        // purpose
        //     test delete and cancel system group
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate delete
        //     read        validate approve
        //     read        validate cancel
        //     update      delete structures
        //     patch       approve structures
        //     patch       cancel structures
        //
        // note
        //     create, approve in order to delete, cancel

        try {
            StructureElementCommandCreate structureElementCommandCreate = null;
            StructureElementCommandConfirm structureElementCommandConfirm = null;
            StructureElement approvedStructureElement = null;

            // create, approve
            structureElementCommandCreate = new StructureElementCommandCreate(
                    Type.SYSTEMGROUP, null,
                    "name", "Dc",
                    "description", "comment");
            approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
            structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(approvedStructureElement);

            structureElementCommandConfirm.setComment("comment update delete check");

            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.CANCEL, Boolean.FALSE);

            // delete
            ITUtilStructureElement.assertDelete(structureElementCommandConfirm);

            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE, Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.CANCEL, Boolean.TRUE);
            ITUtilStructureElement.assertDelete(structureElementCommandConfirm, HttpURLConnection.HTTP_CONFLICT);

            // cancel
            ITUtilStructureElement.assertCancel(structureElementCommandConfirm);

            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.CANCEL, Boolean.FALSE);

            structureElementCommandConfirm.setUuid(null);
            ITUtilStructureElement.assertCancel(structureElementCommandConfirm, ITUtil.HTTP_UNPROCESSABLE_ENTITY);
            structureElementCommandConfirm.setUuid(UUID.randomUUID());
            ITUtilStructureElement.assertCancel(structureElementCommandConfirm, HttpURLConnection.HTTP_NOT_FOUND);
            structureElementCommandConfirm.setUuid(approvedStructureElement.getUuid());
            ITUtilStructureElement.assertCancel(structureElementCommandConfirm, HttpURLConnection.HTTP_CONFLICT);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void deleteReject() {
        // purpose
        //     test delete and reject system group
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate delete
        //     read        validate approve
        //     read        validate reject
        //     update      delete structures
        //     patch       approve structures
        //     patch       reject structures
        //
        // note
        //     create, approve in order to delete, reject

        try {
            StructureElementCommandCreate structureElementCommandCreate = null;
            StructureElementCommandConfirm structureElementCommandConfirm = null;
            StructureElement approvedStructureElement = null;

            // create, approve
            structureElementCommandCreate = new StructureElementCommandCreate(
                    Type.SYSTEMGROUP, null,
                    "name", "Dr",
                    "description", "comment");
            approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
            structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(approvedStructureElement);

            structureElementCommandConfirm.setComment("comment delete reject check");

            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.REJECT, Boolean.FALSE);

            // delete
            ITUtilStructureElement.assertDelete(structureElementCommandConfirm);

            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE, Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.REJECT, Boolean.TRUE);
            ITUtilStructureElement.assertDelete(structureElementCommandConfirm, HttpURLConnection.HTTP_CONFLICT);

            // reject
            ITUtilStructureElement.assertReject(structureElementCommandConfirm);

            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.DELETE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElementCommandConfirm, StructureCommand.REJECT, Boolean.FALSE);

            structureElementCommandConfirm.setUuid(null);
            ITUtilStructureElement.assertReject(structureElementCommandConfirm, ITUtil.HTTP_UNPROCESSABLE_ENTITY);
            structureElementCommandConfirm.setUuid(UUID.randomUUID());
            ITUtilStructureElement.assertReject(structureElementCommandConfirm, HttpURLConnection.HTTP_NOT_FOUND);
            structureElementCommandConfirm.setUuid(approvedStructureElement.getUuid());
            ITUtilStructureElement.assertReject(structureElementCommandConfirm, HttpURLConnection.HTTP_CONFLICT);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void readSearchHistoryApprove() {
        // purpose
        //     test read system group in various ways for create, approve
        //         search
        //         latest
        //         history
        //
        // note
        //     create (and more) to read (with content)

        try {
            StructureElementCommandCreate structureElementCommandCreate = null;
            StructureElementCommandConfirm structureElementCommandConfirm = null;
            StructureElement createdStructureElement  = null;
            StructureElement approvedStructureElement = null;
            int length = 0;

            structureElementCommandCreate = new StructureElementCommandCreate(
                    Type.SYSTEMGROUP, null,
                    "name", "Rsha",
                    "description", "comment");

            // create
            createdStructureElement = ITUtilStructureElement.assertCreate(structureElementCommandCreate);
            structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);

            // read (1)
            length = ITUtilStructureElement.assertRead("/SYSTEMGROUP", 1, -1).getListSize();

            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=PENDING",   1, -1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=APPROVED",  0, -1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=CANCELLED", 0, -1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=REJECTED",  0, -1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=PENDING&mnemonic=Rsha",   1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=APPROVED&mnemonic=Rsha",  0);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=CANCELLED&mnemonic=Rsha", 0);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=REJECTED&mnemonic=Rsha",  0);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?uuid=" + createdStructureElement.getUuid().toString(), 1);

            ITUtilStructureElement.assertRead("/mnemonic/Rsha",     0);

            ITUtilStructureElement.assertRead("/mnemonicPath/Rsha", 0);

            ITUtilStructureElement.assertRead("/history/" + createdStructureElement.getUuid().toString(), 1);

            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "Rsha", Boolean.FALSE);

            // approve
            approvedStructureElement = ITUtilStructureElement.assertApprove(structureElementCommandConfirm);

            // read (2)
            assertEquals(length, ITUtilStructureElement.assertRead("/SYSTEMGROUP", 1, -1).getListSize());

            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=PENDING",   0, -1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=APPROVED",  1, -1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=CANCELLED", 0, -1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=REJECTED",  0, -1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=PENDING&mnemonic=Rsha",   0);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=APPROVED&mnemonic=Rsha",  1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=CANCELLED&mnemonic=Rsha", 0);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=REJECTED&mnemonic=Rsha",  0);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?uuid=" + approvedStructureElement.getUuid().toString(), 1);

            ITUtilStructureElement.assertRead("/mnemonic/Rsha",     1);

            ITUtilStructureElement.assertRead("/mnemonicPath/Rsha", 1);

            ITUtilStructureElement.assertRead("/history/" + approvedStructureElement.getUuid().toString(), 2);

            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "Rsha", Boolean.TRUE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void readSearchHistoryCancel() {
        // purpose
        //     test read system group in various ways for create, cancel
        //         search
        //         latest
        //         history
        //
        // note
        //     create (and more) to read (with content)

        try {
            StructureElementCommandCreate structureElementCommandCreate = null;
            StructureElementCommandConfirm structureElementCommandConfirm = null;
            StructureElement createdStructureElement   = null;
            StructureElement cancelledStructureElement = null;
            int length = 0;

            structureElementCommandCreate = new StructureElementCommandCreate(
                    Type.SYSTEMGROUP, null,
                    "name", "Rshc",
                    "description", "comment");

            // create
            createdStructureElement = ITUtilStructureElement.assertCreate(structureElementCommandCreate);
            structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);

            // read (1)
            length = ITUtilStructureElement.assertRead("/SYSTEMGROUP", 1, -1).getListSize();

            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=PENDING",   1, -1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=APPROVED",  0, -1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=CANCELLED", 0, -1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=REJECTED",  0, -1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=PENDING&mnemonic=Rshc",   1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=APPROVED&mnemonic=Rshc",  0);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=CANCELLED&mnemonic=Rshc", 0);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=REJECTED&mnemonic=Rshc",  0);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?uuid=" + createdStructureElement.getUuid().toString(), 1);

            ITUtilStructureElement.assertRead("/mnemonic/Rshc",     0);

            ITUtilStructureElement.assertRead("/mnemonicPath/Rshc", 0);

            ITUtilStructureElement.assertRead("/history/" + createdStructureElement.getUuid().toString(), 1);

            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "Rshc", Boolean.FALSE);

            // cancel
            cancelledStructureElement = ITUtilStructureElement.assertCancel(structureElementCommandConfirm);

            // read (2)
            assertEquals(length, ITUtilStructureElement.assertRead("/SYSTEMGROUP", 1, -1).getListSize());

            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=PENDING",   0, -1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=APPROVED",  0, -1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=CANCELLED", 1, -1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=REJECTED",  0, -1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=PENDING&mnemonic=Rshc",   0);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=APPROVED&mnemonic=Rshc",  0);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=CANCELLED&mnemonic=Rshc", 1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=REJECTED&mnemonic=Rshc",  0);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?uuid=" + cancelledStructureElement.getUuid().toString(), 1);

            ITUtilStructureElement.assertRead("/mnemonic/Rshc",     0);

            ITUtilStructureElement.assertRead("/mnemonicPath/Rshc", 0);

            ITUtilStructureElement.assertRead("/history/" + cancelledStructureElement.getUuid().toString(), 2);

            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "Rshc", Boolean.FALSE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void readSearchHistoryReject() {
        // purpose
        //     test read system group in various ways for create, reject
        //         search
        //         latest
        //         history
        //
        // note
        //     create (and more) to read (with content)

        try {
            StructureElementCommandCreate structureElementCommandCreate = null;
            StructureElementCommandConfirm structureElementCommandConfirm = null;
            StructureElement createdStructureElement  = null;
            StructureElement rejectedStructureElement = null;
            int length = 0;

            structureElementCommandCreate = new StructureElementCommandCreate(
                    Type.SYSTEMGROUP, null,
                    "name", "Rshr",
                    "description", "comment");

            // create
            createdStructureElement = ITUtilStructureElement.assertCreate(structureElementCommandCreate);
            structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);

            // read (1)
            length = ITUtilStructureElement.assertRead("/SYSTEMGROUP", 1, -1).getListSize();

            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=PENDING",   1, -1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=APPROVED",  0, -1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=CANCELLED", 0, -1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=REJECTED",  0, -1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=PENDING&mnemonic=Rshr",   1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=APPROVED&mnemonic=Rshr",  0);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=CANCELLED&mnemonic=Rshr", 0);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=REJECTED&mnemonic=Rshr",  0);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?uuid=" + createdStructureElement.getUuid().toString(), 1);

            ITUtilStructureElement.assertRead("/mnemonic/Rshr",     0);

            ITUtilStructureElement.assertRead("/mnemonicPath/Rshr", 0);

            ITUtilStructureElement.assertRead("/history/" + createdStructureElement.getUuid().toString(), 1);

            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "Rshr", Boolean.FALSE);

            // reject
            rejectedStructureElement = ITUtilStructureElement.assertReject(structureElementCommandConfirm);

            // read (2)
            assertEquals(length, ITUtilStructureElement.assertRead("/SYSTEMGROUP", 1, -1).getListSize());

            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=PENDING",   0, -1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=APPROVED",  0, -1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=CANCELLED", 0, -1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=REJECTED",  1, -1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=PENDING&mnemonic=Rshr",   0);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=APPROVED&mnemonic=Rshr",  0);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=CANCELLED&mnemonic=Rshr", 0);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=REJECTED&mnemonic=Rshr",  1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?uuid=" + rejectedStructureElement.getUuid().toString(), 1);

            ITUtilStructureElement.assertRead("/mnemonic/Rshr",     0);

            ITUtilStructureElement.assertRead("/mnemonicPath/Rshr", 0);

            ITUtilStructureElement.assertRead("/history/" + rejectedStructureElement.getUuid().toString(), 2);

            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "Rshr", Boolean.FALSE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void readSearchHistoryStatusLatest1() {
        // purpose
        //     test read system group in various ways
        //         status
        //         latest
        //       ( deleted )
        //         history
        //
        // what
        //     entries with different statuses
        //
        // note
        //     create (and more) to read (with content)
        //     4 different lines  of uuid with combinations of status and latest
        //         1 entry for each line of uuid
        //         in particular different statuses for last entry in lines of uuid
        //         mnemonic - P, C, R, A (pending, cancelled, rejected, approved)
        //
        // note!
        //     not only
        //         exclude APPROVED and not latest
        //     not only
        //         exclude earlier than APPROVED and latest (considered history)
        //         include PENDING and not latest           (considered future)
        //     instead
        //         exclude earlier than APPROVED and latest (considered history)
        //       ( include if latest )
        //         include if latest not exists             (considered present + future)

        StructureElementCommandCreate structureElementCommandCreate = null;
        StructureElement createdStructureElement = null;
        UUID uuid, uuid2, uuid3, uuid4 = null;

        // a number of entries for which the last entry has status PENDING
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "P1",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreate(structureElementCommandCreate);
        uuid = createdStructureElement.getUuid();

        // a number of entries for which the last entry has status CANCELLED
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "C1",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateCancel(structureElementCommandCreate);
        uuid2 = createdStructureElement.getUuid();

        // a number of entries for which the last entry has status REJECTED
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "R1",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateReject(structureElementCommandCreate);
        uuid3 = createdStructureElement.getUuid();

        // a number of entries for which the last entry has status APPROVED
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "A1",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        uuid4 = createdStructureElement.getUuid();

        try {
            // from first structure element
            assertNotNull(uuid);

            // note
            //     exclude earlier than APPROVED and latest (considered history)
            //   ( include if latest )
            //     include if latest not exists             (considered present + future)

            ITUtilStructureElement.assertRead("/SYSTEMGROUP?uuid=" + uuid.toString(),  1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?uuid=" + uuid2.toString(), 1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?uuid=" + uuid3.toString(), 1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?uuid=" + uuid4.toString(), 1);

            ITUtilStructureElement.assertRead("/mnemonic/P1", 0);
            ITUtilStructureElement.assertRead("/mnemonic/C1", 0);
            ITUtilStructureElement.assertRead("/mnemonic/R1", 0);
            ITUtilStructureElement.assertRead("/mnemonic/A1", 1);

            ITUtilStructureElement.assertRead("/history/" + uuid.toString(),  1);
            ITUtilStructureElement.assertRead("/history/" + uuid2.toString(), 2);
            ITUtilStructureElement.assertRead("/history/" + uuid3.toString(), 2);
            ITUtilStructureElement.assertRead("/history/" + uuid4.toString(), 2);

            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "P1", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "C1", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "R1", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "A1", Boolean.TRUE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void readSearchHistoryStatusLatest9() {
        // purpose
        //     test read system group in various ways
        //         status
        //         latest
        //       ( deleted )
        //         history
        //
        // what
        //     entries with different statuses
        //
        // note
        //     create (and more) to read (with content)
        //     4 different lines  of uuid with combinations of status and latest
        //         9 entries for each line of uuid
        //         in particular different statuses for last entry in lines of uuid
        //         mnemonic - P, C, R, A (pending, cancelled, rejected, approved)
        //
        // note!
        //     not only
        //         exclude APPROVED and not latest
        //     not only
        //         exclude earlier than APPROVED and latest (considered history)
        //         include PENDING and not latest           (considered future)
        //     instead
        //         exclude earlier than APPROVED and latest (considered history)
        //       ( include if latest )
        //         include if latest not exists             (considered present + future)

        StructureElementCommandCreate structureElementCommandCreate = null;
        StructureElementCommandUpdate structureElementCommandUpdate = null;
        StructureElement createdStructureElement = null;
        UUID uuid, uuid2, uuid3, uuid4 = null;

        // a number of entries for which the last entry has status PENDING
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "P9",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(createdStructureElement);
        uuid = createdStructureElement.getUuid();

        structureElementCommandUpdate.setComment("comment 2");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 3");
        ITUtilStructureElement.assertUpdateCancel(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 4");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 5");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 6");
        ITUtilStructureElement.assertUpdateReject(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 7");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 8");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 9");
        ITUtilStructureElement.assertUpdate(structureElementCommandUpdate);

        // a number of entries for which the last entry has status CANCELLED
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "C9",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(createdStructureElement);
        uuid2 = createdStructureElement.getUuid();

        structureElementCommandUpdate.setComment("comment 2");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 3");
        ITUtilStructureElement.assertUpdateCancel(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 4");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 5");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 6");
        ITUtilStructureElement.assertUpdateReject(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 7");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 8");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 9");
        ITUtilStructureElement.assertUpdateCancel(structureElementCommandUpdate);

        // a number of entries for which the last entry has status REJECTED
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "R9",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(createdStructureElement);
        uuid3 = createdStructureElement.getUuid();

        structureElementCommandUpdate.setComment("comment 2");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 3");
        ITUtilStructureElement.assertUpdateCancel(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 4");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 5");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 6");
        ITUtilStructureElement.assertUpdateReject(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 7");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 8");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 9");
        ITUtilStructureElement.assertUpdateReject(structureElementCommandUpdate);

        // a number of entries for which the last entry has status APPROVED
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "A9",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(createdStructureElement);
        uuid4 = createdStructureElement.getUuid();

        structureElementCommandUpdate.setComment("comment 2");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 3");
        ITUtilStructureElement.assertUpdateCancel(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 4");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 5");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 6");
        ITUtilStructureElement.assertUpdateReject(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 7");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 8");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate.setComment("comment 9");
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);

        try {
            // from first structure element
            assertNotNull(uuid);

            // note
            //     exclude earlier than APPROVED and latest (considered history)
            //   ( include if latest )
            //     include if latest not exists             (considered present + future)

            ITUtilStructureElement.assertRead("/SYSTEMGROUP?uuid=" + uuid.toString(),  2);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?uuid=" + uuid2.toString(), 2);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?uuid=" + uuid3.toString(), 2);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?uuid=" + uuid4.toString(), 1);

            ITUtilStructureElement.assertRead("/mnemonic/P9", 1);
            ITUtilStructureElement.assertRead("/mnemonic/C9", 1);
            ITUtilStructureElement.assertRead("/mnemonic/R9", 1);
            ITUtilStructureElement.assertRead("/mnemonic/A9", 1);

            ITUtilStructureElement.assertRead("/history/" + uuid.toString(),  17);
            ITUtilStructureElement.assertRead("/history/" + uuid2.toString(), 18);
            ITUtilStructureElement.assertRead("/history/" + uuid3.toString(), 18);
            ITUtilStructureElement.assertRead("/history/" + uuid4.toString(), 18);

            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "P9", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "C9", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "R9", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "A9", Boolean.TRUE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void readSearchStatusDeletedChildren() {
        // purpose
        //     test read system group in various ways
        //         status
        //       ( latest )
        //         deleted
        //         children
        //
        // what
        //     entries with different statuses
        //
        // note
        //     create (and more) to read (with content)
        //     querying for Status.APPROVED means latest approved

        StructureElementCommandCreate structureElementCommandCreate = null;
        StructureElementCommandUpdate structureElementCommandUpdate = null;
        StructureElementCommandConfirm structureElementCommandConfirm = null;
        StructureElement createdStructureElement = null;
        UUID uuid = null;

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AA1",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        uuid = createdStructureElement.getUuid();

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AA2",
                "description", "comment");
        ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AA3",
                "description", "comment");
        ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AA4",
                "description", "comment");
        ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AA5",
                "description", "comment");
        ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AB1",
                "description", "comment");
        ITUtilStructureElement.assertCreateCancel(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AB2",
                "description", "comment");
        ITUtilStructureElement.assertCreateCancel(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AB3",
                "description", "comment");
        ITUtilStructureElement.assertCreateCancel(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AB4",
                "description", "comment");
        ITUtilStructureElement.assertCreateCancel(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AB5",
                "description", "comment");
        ITUtilStructureElement.assertCreateCancel(structureElementCommandCreate);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AC1",
                "description", "comment");
        ITUtilStructureElement.assertCreateReject(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AC2",
                "description", "comment");
        ITUtilStructureElement.assertCreateReject(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AC3",
                "description", "comment");
        ITUtilStructureElement.assertCreateReject(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AC4",
                "description", "comment");
        ITUtilStructureElement.assertCreateReject(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AC5",
                "description", "comment");
        ITUtilStructureElement.assertCreateReject(structureElementCommandCreate);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AD1",
                "description", "comment");
        ITUtilStructureElement.assertCreate(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AD2",
                "description", "comment");
        ITUtilStructureElement.assertCreate(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AD3",
                "description", "comment");
        ITUtilStructureElement.assertCreate(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AD4",
                "description", "comment");
        ITUtilStructureElement.assertCreate(structureElementCommandCreate);
        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AD5",
                "description", "comment");
        ITUtilStructureElement.assertCreate(structureElementCommandCreate);

        String description2 = "some other description";
        String comment2 = "some other comment";
        String description3 = "more description";
        String comment3 = "more comment";

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AE1",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(createdStructureElement);
        structureElementCommandUpdate.setDescription(description2);
        structureElementCommandUpdate.setComment(comment2);
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(createdStructureElement);
        structureElementCommandUpdate.setDescription(description3);
        structureElementCommandUpdate.setComment(comment3);
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDeleteReject(structureElementCommandConfirm);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AE2",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(createdStructureElement);
        structureElementCommandUpdate.setDescription(description2);
        structureElementCommandUpdate.setComment(comment2);
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(createdStructureElement);
        structureElementCommandUpdate.setDescription(description3);
        structureElementCommandUpdate.setComment(comment3);
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDeleteReject(structureElementCommandConfirm);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AE3",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(createdStructureElement);
        structureElementCommandUpdate.setDescription(description2);
        structureElementCommandUpdate.setComment(comment2);
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(createdStructureElement);
        structureElementCommandUpdate.setDescription(description3);
        structureElementCommandUpdate.setComment(comment3);
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDeleteReject(structureElementCommandConfirm);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AE4",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(createdStructureElement);
        structureElementCommandUpdate.setDescription(description2);
        structureElementCommandUpdate.setComment(comment2);
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(createdStructureElement);
        structureElementCommandUpdate.setDescription(description3);
        structureElementCommandUpdate.setComment(comment3);
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDeleteReject(structureElementCommandConfirm);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AE5",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(createdStructureElement);
        structureElementCommandUpdate.setDescription(description2);
        structureElementCommandUpdate.setComment(comment2);
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandUpdate = StructureElementUtil.convertElement2CommandUpdate(createdStructureElement);
        structureElementCommandUpdate.setDescription(description3);
        structureElementCommandUpdate.setComment(comment3);
        ITUtilStructureElement.assertUpdateApprove(structureElementCommandUpdate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDeleteReject(structureElementCommandConfirm);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AF1",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDeleteApprove(structureElementCommandConfirm);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AF2",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDeleteApprove(structureElementCommandConfirm);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AF3",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDeleteApprove(structureElementCommandConfirm);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AF4",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDeleteApprove(structureElementCommandConfirm);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AF5",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDeleteApprove(structureElementCommandConfirm);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AG1",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDelete(structureElementCommandConfirm);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AG2",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDelete(structureElementCommandConfirm);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AG3",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDelete(structureElementCommandConfirm);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AG4",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDelete(structureElementCommandConfirm);

        structureElementCommandCreate = new StructureElementCommandCreate(
                Type.SYSTEMGROUP, null,
                "name", "AG5",
                "description", "comment");
        createdStructureElement = ITUtilStructureElement.assertCreateApprove(structureElementCommandCreate);
        structureElementCommandConfirm = StructureElementUtil.convertElement2CommandConfirm(createdStructureElement);
        ITUtilStructureElement.assertDelete(structureElementCommandConfirm);

        // 60 system group entries

        try {
            // from first structure element
            assertNotNull(uuid);

            // read & search
            //     /{type}
            //     type, statuses, deleted, uuid, parent, name, mnemonic, mnemonicEquivalence, mnemonicPath, description
            //     combination
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?mnemonic=A__",                                                  45);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=PENDING&mnemonic=A__",                                 10);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=APPROVED&mnemonic=A__",                                20);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=CANCELLED&mnemonic=A__",                                5);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=REJECTED&mnemonic=A__",                                10);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?statuses=PENDING&statuses=APPROVED&mnemonic=A__",               30);

            ITUtilStructureElement.assertRead("/SYSTEMGROUP?deleted=false&mnemonic=A__",                                    30);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?deleted=false&statuses=PENDING&mnemonic=A__",                    5);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?deleted=false&statuses=APPROVED&mnemonic=A__",                  15);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?deleted=false&statuses=CANCELLED&mnemonic=A__",                  5);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?deleted=false&statuses=REJECTED&mnemonic=A__",                   5);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?deleted=false&statuses=PENDING&statuses=APPROVED&mnemonic=A__", 20);

            ITUtilStructureElement.assertRead("/SYSTEMGROUP?deleted=true&mnemonic=A__",                                     15);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?deleted=true&statuses=PENDING&mnemonic=A__",                     5);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?deleted=true&statuses=APPROVED&mnemonic=A__",                    5);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?deleted=true&statuses=CANCELLED&mnemonic=A__",                   0);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?deleted=true&statuses=REJECTED&mnemonic=A__",                    5);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?deleted=true&statuses=PENDING&statuses=APPROVED&mnemonic=A__",  10);

            ITUtilStructureElement.assertRead("/SYSTEMGROUP?uuid=" + uuid.toString(),    1);

            ITUtilStructureElement.assertRead("/SYSTEMGROUP?parent=" + uuid.toString(),  0);

            ITUtilStructureElement.assertRead("/SYSTEMGROUP?name=na",                        0);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?name=na_",                       0);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?name=na__",                     45, -1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?name=na___",                     0);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?name=na%",                      45, -1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?name=name",                     45, -1);

            ITUtilStructureElement.assertRead("/SYSTEMGROUP?mnemonicEquivalence=A__",       45, -1);

            ITUtilStructureElement.assertRead("/SYSTEMGROUP?mnemonicPath=A__",              45, -1);

            ITUtilStructureElement.assertRead("/SYSTEMGROUP?description=desc",               0);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?description=desc%",             35, -1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?description=sc",                 0);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?description=sc%",                0);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?description=%sc",                0);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?description=%sc%",              45, -1);
            ITUtilStructureElement.assertRead("/SYSTEMGROUP?description=description",       35, -1);

            // children
            //     /children/{uuid}
            //     uuid, type, statuses, deleted, name, mnemonic, mnemonicEquivalence, mnemonicPath, description
            ITUtilStructureElement.assertRead("/children/" + uuid.toString(),                       0);
            ITUtilStructureElement.assertRead("/children/" + uuid.toString() + "?type=SYSTEMGROUP", 0);

            // mnemonic
            //     /mnemonic/{mnemonic}
            ITUtilStructureElement.assertRead("/mnemonic/A",        0);
            ITUtilStructureElement.assertRead("/mnemonic/A__",     15, -1);
            ITUtilStructureElement.assertRead("/mnemonic/AG_",      5);
            ITUtilStructureElement.assertRead("/mnemonic/AG1",      1);

            // mnemonic path
            //     /mnemonicPath/{mnemonicPath}
            ITUtilStructureElement.assertRead("/mnemonicPath/A",    0);
            ITUtilStructureElement.assertRead("/mnemonicPath/A__", 15, -1);
            ITUtilStructureElement.assertRead("/mnemonicPath/AG_",  5);
            ITUtilStructureElement.assertRead("/mnemonicPath/AG1",  1);

            // history
            //     /history/{uuid}
            ITUtilStructureElement.assertRead("/history/" + uuid.toString(), 2);

            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AA1", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AA2", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AA3", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AA4", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AA5", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AB1", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AB2", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AB3", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AB4", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AB5", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AC1", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AC2", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AC3", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AC4", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AC5", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AD1", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AD2", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AD3", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AD4", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AD5", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AE1", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AE2", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AE3", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AE4", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AE5", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AF1", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AF2", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AF3", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AF4", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AF5", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AG1", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AG2", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AG3", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AG4", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.SYSTEMGROUP, "AG5", Boolean.TRUE);
        } catch (Exception e) {
            fail();
        }
    }

}
